// System includes
#include <cstring>
#include <iostream>
#include <vector>

// Local includes
#include "result_validator.hpp"
#include "validator_client.hpp"

using std::cerr;
using std::cout;
using std::endl;
using std::string;
using std::vector;

namespace prc {

// Initialize the singleton instance to NULL.
ValidatorClient* prc::ValidatorClient::clientInstance = NULL;
const string prc::ValidatorClient::GET_RESULT_RPC_NAME = "server.getResultForValidation";
const string prc::ValidatorClient::MARK_VALID_RPC_NAME = "server.markResultValid";
const string prc::ValidatorClient::MARK_INVALID_RPC_NAME = "server.markResultInvalid";
const string prc::ValidatorClient::SELECT_CANONICAL_RPC_NAME = "server.selectCanonicalResult";
const string prc::ValidatorClient::GET_ASSOCIATED_WU_RPC_NAME = "server.getAssociatedWorkUnit";

/**
 * @return Returns the singleton instance of the ValidatorClient
 */
ValidatorClient* ValidatorClient::instance() {
	if (clientInstance == NULL)
		clientInstance = new ValidatorClient();
	return clientInstance;
}

/**
 * Gets the next result to be validated from the server, then 
 * invokes the appropriate method of the ResultValidator, depending
 * on what type of result needed to be validated.  When the ResultValidator
 * has finished validating the result, this method will inform the project
 * server of the outcome.
 */
void ValidatorClient::validateNextResult() {
	string errorMessage = "Error while requesting result for validation";
	xmlrpc_c::paramList params;

	try {
		// Build the RPC.
		xmlrpc_c::rpcPtr xmlrpc(GET_RESULT_RPC_NAME, params);

		// Execute the RPC.
		xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);
		vector<xmlrpc_c::value> resultVector = ((xmlrpc_c::value_array)result).vectorValueValue();

		ResultValidator validator;

		// Do a different type of validation depending on the enumeration value.
		ValidationType type = static_cast<ValidationType>((int)((xmlrpc_c::value_int)resultVector[0]));

		// Need to validate a single result.
		if (type == SINGLE) {
			Result r;
			r.resultID = (string)((xmlrpc_c::value_string)resultVector[1]);
			r.data = (vector<byte>)((xmlrpc_c::value_bytestring)resultVector[2]).vectorUcharValue();

			// Use the ResultValidator to determine whether the result was valid.
			if (validator.validateSingleResult(r) == true) {
				// If it was valid, mark it valid on the server.
				if (markResultValid(r.resultID) == false) {
					cerr << "Error marking result " << r.resultID << " valid.  ";
					cerr << "The server could not find the specified result ID." << endl;
				}
			}
			else {
				// Otherwise, mark it invalid on the server.
				if (markResultInvalid(r.resultID) == false) {
					cerr << "Error marking result " << r.resultID << " invalid.  ";
					cerr << "The server could not find the specified result ID." << endl;
				}
			}
		}

		// Need to select a canonical result from a set of results.
		else if (type == CANONICAL) {
			int numResults = (int)((xmlrpc_c::value_int)resultVector[1]);
			vector<Result> results;
			for (int ii = 2; ii <= numResults * 2; ii += 2) {
				Result r;
				r.resultID = (string)((xmlrpc_c::value_string)resultVector[ii]);
				r.data = (vector<byte>)((xmlrpc_c::value_bytestring)resultVector[ii+1]).vectorUcharValue();
				results.push_back(r);
			}

			// Select the canonical result with the ResultValidator.
			string canonicalID = validator.selectCanonicalResult(results);

			// Inform the server of the canonical result choice.
			if (selectCanonicalResult(canonicalID) == false) {
				cerr << "Error selecting canonical result " << canonicalID << " invalid.  ";
				cerr << "The server could not find the specified result ID." << endl;
			}
		}

		// Need to validate a spot-check.
		else if (type == SPOTCHECK) {
			SpotCheck check;
			check.resultID = (string)((xmlrpc_c::value_string)resultVector[1]);
			check.clientData = (vector<byte>)((xmlrpc_c::value_bytestring)resultVector[2]).vectorUcharValue();
			check.acceptedData = (vector<byte>)((xmlrpc_c::value_bytestring)resultVector[3]).vectorUcharValue();

			// Use the ResultValidator to determine whether the spot-check was valid.
			if (validator.validateSpotCheck(check) == true) {
				// If it was valid, mark it valid on the server.
				if (markResultValid(check.resultID) == false) {
					cerr << "Error marking spot-check " << check.resultID << " valid.  ";
					cerr << "The server could not find the specified result ID." << endl;
				}
			}
			else {
				// Otherwise, mark it invalid on the server.
				if (markResultInvalid(check.resultID) == false) {
					cerr << "Error marking spot-check " << check.resultID << " invalid.  ";
					cerr << "The server could not find the specified result ID." << endl;
				}
			}
		}
	} catch (string error) {
		// The server did not have any results to validate,
		// or an error occurred such as the given RPC was not found.
		//cout << "The server returned: " << error << endl;
		cout << "The server did not have any results that needed to be validated." << endl;
		// Rethrow the exception to be caught by the main.
		throw error;
	}
}

/**
 * Marks the result with the given ID as valid.
 * @param resultID The ID of the result to mark valid.
 * @return Returns true if the result was successfully marked valid, 
 * false if no result with the given ID was found.
 */
bool ValidatorClient::markResultValid(string& resultID) {
	cout << "Marking valid result " << resultID << endl;
	string errorMessage = "Error while marking result valid";
	xmlrpc_c::paramList params(1);
	params.add(xmlrpc_c::value_string(resultID));

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(MARK_VALID_RPC_NAME, params);

	// Execute the RPC.
	xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);

	// Convert the result into a bool.
	return (bool) ((xmlrpc_c::value_boolean)result);
}

/**
 * Marks the result with the given ID as invalid.
 * @param resultID The ID of the result to mark invalid.
 * @return Returns true if the result was successfully marked invalid, 
 * false if no result with the given ID was found.
 */
bool ValidatorClient::markResultInvalid(string& resultID) {
	cout << "Marking invalid result " << resultID << endl;
	string errorMessage = "Error while marking result invalid";
	xmlrpc_c::paramList params(1);
	params.add(xmlrpc_c::value_string(resultID));

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(MARK_INVALID_RPC_NAME, params);

	// Execute the RPC.
	xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);

	// Convert the result into a bool.
	return (bool) ((xmlrpc_c::value_boolean)result);
}

/**
 * Selects the result with the given ID as the canonical result for a work unit.
 * @param resultID The ID of the result to select.
 * @return Returns true if the result was successfully selected, 
 * false if no result with the given ID was found.
 */
bool ValidatorClient::selectCanonicalResult(string& resultID) {
	cout << "Selecting canonical result " << resultID << endl;
	string errorMessage = "Error while selecting canonical result";
	xmlrpc_c::paramList params(1);
	params.add(xmlrpc_c::value_string(resultID));

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(SELECT_CANONICAL_RPC_NAME, params);

	// Execute the RPC.
	xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);

	// Convert the result into a bool.
	return (bool) ((xmlrpc_c::value_boolean)result);
}

/**
* Gets the associated work unit for the given result.
* @param resultID The result for which to get the associated work unit.
* @return Returns the data contained in the associated work unit.
*/
vector<byte> ValidatorClient::getAssociatedWorkUnit(string& resultID) {
	cout << "Getting associated work unit for result " << resultID << endl;
	string errorMessage = "Error while getting associated work unit";
	xmlrpc_c::paramList params(1);
	params.add(xmlrpc_c::value_string(resultID));

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(GET_ASSOCIATED_WU_RPC_NAME, params);

	// Execute the RPC.
	xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);

	// Convert the result into a byte vector.
	vector<xmlrpc_c::value> resultVector = ((xmlrpc_c::value_array)result).vectorValueValue();
	return (vector<byte>) ((xmlrpc_c::value_bytestring)resultVector[0]).vectorUcharValue();
}

} // end namespace prc
