package edu.wpi.prc.example.java.common;

import org.apache.xmlrpc.WebServer;

/**
 * The Server class is responsible for initializing, starting, and stopping an XML-RPC server.
 * @author James Baldassari
 */
public class Server {
	private static final int EXIT_ERROR = 1;
	private int port;							// The server port to use.
	private WebServer xmlrpcServer = null;		// The XML-RPC server.
	private boolean isRunning = false;			// Is the server running?
	
	/**
	 * Constructor that takes the port number on which to listen.
	 * Initializes the XML-RPC server.
	 * @param port The port number on which to listen for XML-RPC requests.
	 */
	public Server(int port) {
		this.port = port;
		initXmlRpcServer();
	}
	
	/**
	 * Initializes and starts the XML-RPC server.
	 */
	private void initXmlRpcServer() {
		try {
			// Instantiate the web server.
			xmlrpcServer = new WebServer(port);

		} catch (Exception e) {
			System.err.println("Error initializing the XML-RPC server: " + e);
			System.exit(EXIT_ERROR);
		}
	}
	
	/**
	 * Registers the given RPC handler object with the XML-RPC server.
	 * @param handlerName The name of the handler.  All RPCs will be prefixed with this name.
	 * Example: in the call "sciapp.computeResult" "sciapp" is the name of the handler and 
	 * "computeResult" is the name of the method to call inside that handler.
	 * @param rpcHandler The object containing the handler methods.
	 */
	public void registerRPCHandler(String handlerName, Object rpcHandler) {
		// Register RPC handler.
		xmlrpcServer.addHandler(handlerName, rpcHandler);
	}
	
	/**
	 * @return Returns true if the server is running, false otherwise.
	 */
	public boolean isRunning() {
		return isRunning;
	}
	
	/**
	 * Starts the XML-RPC server and begins handling requests.
	 */
	public void start() {
		if (isRunning == false) {
			System.out.println("Starting XML-RPC Server.");
			xmlrpcServer.start();
			isRunning = true;
		}
	}
	
	/**
	 * Performs any tasks necessary to clearnly shut down the server.
	 * @return Returns true.
	 */
	public boolean shutdown() {
		System.out.println("Server: Shutting down...");
		xmlrpcServer.shutdown();
		isRunning = false;
		return true;
	}
}
