package edu.wpi.prc.example.java.science;

import java.util.Vector;

/**
 * This class is the RPC handler for the science application's XML-RPC server.
 * Any of the methods in this class can be called by an XML-RPC client.
 * @author James Baldassari
 */
public class RPCHandler {
	ScienceApplication sciapp = null;
	
	/**
	 * Constructor taking a reference to the science application.
	 * @param sciapp The science application using this RPC handler.
	 */
	public RPCHandler(ScienceApplication sciapp) {
		this.sciapp = sciapp;
	}
	
	/**
	 * This handler is called when the project client sends a new work unit to the science 
	 * application to be computed.  This method will compute the result of the work unit, 
	 * then return that result to the project client.
	 * @param data The work unit to be computed.
	 * @return Returns a Vector representing the computed result.
	 */
	public Vector computeResult(byte[] data) {
		ScienceApplicationClient.instance().setCheckpointing(true);
	
		ComputeThread computeThread = new ComputeThread(data);
		computeThread.start();
		try {
			computeThread.join();
		} catch (InterruptedException e) {
			System.out.println("The compute thread terminated while processing a work unit!");
			shutdown();
		}
		
		// Get the result from the compute thread.
		Vector<byte[]> result = new Vector<byte[]>();
		byte[] resultData = computeThread.getResult();
		result.add(resultData);
		return result;
	}

	/**
	 * This handler is called when the project server sends a spot check work unit to the science 
	 * application to be computed.  This method will compute the result of the work unit, 
	 * then return that result to the project server.  Checkpointing should not be used in this case 
	 * because there is no project client involved in the interaction between the server and science 
	 * application.  Note that only the client implements the checkpointing RPCs.
	 * @param data The work unit to be computed.
	 * @return Returns a Vector representing the computed result.
	 */
	public Vector computeSpotCheckResult(byte[] data) {
		ScienceApplicationClient.instance().setCheckpointing(false);

		ComputeThread computeThread = new ComputeThread(data);
		computeThread.start();
		try {
			computeThread.join();
		} catch (InterruptedException e) {
			System.out.println("The compute thread terminated while processing a work unit!");
			shutdown();
		}
		
		// Get the result from the compute thread.
		Vector<byte[]> result = new Vector<byte[]>();
		byte[] resultData = computeThread.getResult();
		result.add(resultData);
		return result;
	}
	
	/**
	 * Performs any tasks necessary to clearnly shut down the science application.
	 * @return Returns true.
	 */
	public boolean shutdown() {
		return sciapp.shutdown();
	}
}
