package edu.wpi.prc.example.java.science;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.util.ArrayList;

/**
 * The ScienceDataProcessor contains all of the project-specific methods 
 * required to compute the results for work units.  This class is used 
 * by the ComputeThread.
 * @author James Baldassari
 */
public class ScienceDataProcessor {
	private static final int CHECKPOINT_INTERVAL = 100000;		// Save a checkpoint after every 100,000 numbers that are checked.
	private byte[] workUnitData = null;
	
	/**
	 * Constructor that takes a byte array representing the 
	 * work unit data to be processed.
	 * @param workUnitData The work unit data.
	 */
	public ScienceDataProcessor(byte[] workUnitData) {
		this.workUnitData = workUnitData;
	}
	
	/**
	 * This method checks for previously saved checkpoints, then runs the 
	 * science algorithm, findPrimes().  It then collects all of the primes
	 * that were found and produces a byte array representing them, which is 
	 * the result that will be returned to the server.
	 * @return Returns a byte array representing the computed result.
	 */
	public byte[] computeResult() {
		int startRange = 0;
		int endRange = 0;
		
		// Convert the byte array into two integers.
		ByteArrayInputStream bais = new ByteArrayInputStream(workUnitData);
		DataInputStream dis = new DataInputStream(bais);
		try {
			startRange = dis.readInt();
			endRange = dis.readInt();
			System.out.println("Science Application: Start range is " + startRange);
			System.out.println("Science Application: End range is " + endRange);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// Check the client for any saved checkpoints.
		ArrayList<Integer> previouslyFoundPrimes = new ArrayList<Integer>();
		byte[] checkpoint = ScienceApplicationClient.instance().getCheckpoint();
		
		// We know a checkpoint was retrieved if its length is greater than zero.
		if (checkpoint.length > 0) {
			System.out.println("Science Application: Recovered a checkpoint.");
			// Interpret the checkpoint data.
			bais = new ByteArrayInputStream(checkpoint);
			dis = new DataInputStream(bais);
			
			// Extract the last start position from the checkpoint data.
			int newStartRange = -1;
			try {
				newStartRange = dis.readInt();
			} catch (IOException e) {
				
			}
			
			// Make sure the start range value is valid.
			if ((newStartRange >= 0) && (newStartRange < endRange)) {
				startRange = newStartRange;
				System.out.println("Science Application: New start range is " + newStartRange);
				
				// Extract all prime numbers found up to the last checkpoint.
				boolean doneReading = false;
				try {
					while (doneReading == false)
						previouslyFoundPrimes.add(new Integer(dis.readInt()));
				} catch (EOFException e) {
					doneReading = true;
				} catch (IOException e) {
					
				}
			}
		}
		
		// Find all primes
		ArrayList<Integer> foundPrimes = findPrimes(startRange, endRange);
		
		// Add any primes found previously to the newly found primes.
		ArrayList<Integer> primes = new ArrayList<Integer>(previouslyFoundPrimes.size() + foundPrimes.size());
		primes.addAll(previouslyFoundPrimes);
		primes.addAll(foundPrimes);
		
		// Convert the integers into a byte array to be sent as the result.
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream dos = new DataOutputStream(baos);
		try {
			// Write all primes in the correct order into the output stream.
			for (Integer prime : primes) {
				dos.writeInt(prime);
			}
			dos.flush();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// Extract the byte array from the output stream.
		byte[] resultData = baos.toByteArray();
		
		// Return the computed result.
		return resultData;
	}
	
	/**
	 * This method is the science algorithm.  It finds all prime numbers in the given interval.
	 * @param start The start of the interval (inclusive) in which to search.
	 * @param end The end of the interval (inclusive) in which to search.
	 * @return Returns an array of all primes that were found, or an array of length zero if no primes were found.
	 */
	private ArrayList<Integer> findPrimes(int start, int end) {
		// This will store all the prime numbers that were found.
		ArrayList<Integer> primes = new ArrayList<Integer>();
	
		System.out.println("Science Application: Testing Range [" + start + "," + end + "]");

		// This algorithm is not the most efficient, but it works.	
		for (int ii = start; ii <= end; ii++) {
			boolean isPrime = true;
			int maxSearch = (int)java.lang.Math.sqrt((double)ii);	// Only need to search up to the square root of the number.
			for (int jj = 2; jj <= maxSearch; jj++) {
				if (ii % jj == 0) {
					isPrime = false;
					break;
				}
			}
		
			if (isPrime == true) {
				System.out.println("Science Application: Found Prime: " + ii);
				// Add the prime to the ArrayList of primes.
				primes.add(ii);
			}
			
			// Save a checkpoint after every 1000 numbers checked
			// so that if the science application is killed before 
			// finishing a work unit, it can resume from where it 
			// left off the next time it is started.
			if ((ii % CHECKPOINT_INTERVAL == 0) && (ii + 1 < end)) {
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				DataOutputStream dos = new DataOutputStream(baos);
				try {
					// Write the current state of the prime search.
					dos.writeInt(ii + 1);
					// Write all the primes found so far. 
					for (Integer prime : primes)
						dos.writeInt(prime.intValue());
					dos.flush();
					byte[] checkpoint = baos.toByteArray();
					// Send checkpoint to project client.
					ScienceApplicationClient.instance().saveCheckpoint(checkpoint);
				} catch (IOException e) {
					
				}
			}
		}
		
		return primes;
	}
}
