package edu.wpi.prc.example.java.validator;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.Vector;

/**
 * The ResultValidator class contains project-specific code
 * used to validate work units and spot-checks, and to select
 * the canonical result for a work unit.
 * @author James Baldassari
 */
public class ResultValidator {
	private static final int SIZEOF_INT = Integer.SIZE / 8;		// Size, in bytes, of an integer.
	
	/**
	 * Validates a single result.
	 * @param r The result to validate.
	 * @return Returns true if the result is valid, false otherwise.
	 */
	public boolean validateSingleResult(Result r) {
		System.out.println("Result Validator: Validating result " + r.getResultID());
		
		// We want to verify that each integer calculated by the science application is in the 
		// range given by the work unit, so first get the work unit from which this result was computed.
		byte[] workUnitData = ValidatorClient.instance().getAssociatedWorkUnit(r.getResultID());
		
		int startRange = 0;
		int endRange = 0;
		
		// Convert the byte array into two integers representing the range.
		ByteArrayInputStream bais = new ByteArrayInputStream(workUnitData);
		DataInputStream dis = new DataInputStream(bais);
		try {
			startRange = dis.readInt();
			endRange = dis.readInt();
			System.out.println("Result Validator: Start range was " + startRange);
			System.out.println("Result Validator: End range was " + endRange);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// Examine each computed integer and verify that it is in the range.
		boolean allIntegersValid = true;
		int numChecked = 0;
		bais = new ByteArrayInputStream(r.getData());
		dis = new DataInputStream(bais);
		try {
			// Keep looping while there are more integers to read.
			while (dis.available() >= SIZEOF_INT) {
				int prime = dis.readInt();
				// Check whether the prime is in the range.
				if ((prime < startRange) || (prime > endRange)) {
					System.out.println("Result Validator: Detected invalid result data.");
					System.out.println("Result Validator: " + prime + " is not in the work unit's range.");
					allIntegersValid = false;
					break;
				}
				numChecked++;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		if (allIntegersValid == true)
			System.out.println("Result Validator: Found " + numChecked + " valid primes.");
		
		return allIntegersValid;
	}
	
	/**
	 * Selects a canonical result from a set of results.
	 * @param results The set of valid results for a work unit.
	 * @return Returns the result ID of the canonical result that was selected.
	 */
	public String selectCanonicalResult(Vector<Result> results) {
		// In this validator we always select the first result.
		System.out.println("Result Validator: Selecting canonical result " + results.get(0).getResultID());
		return results.get(0).getResultID();
	}
	
	/**
	 * Validates a spot-check.
	 * @param check The spot-check to validate.
	 * @return Returns true if the spot-check is valid, false otherwise.
	 */
	public boolean validateSpotCheck(SpotCheck check) {
		// For this project we don't have floating point values or any nondeterminism, 
		// so all results should be exactly the same.  This method does a byte-by-byte 
		// comparison of the spot-check to the accepted result.
		System.out.println("Result Validator: Validating spot-check " + check.getResultID());
		
		boolean validSpotCheck = true;
		
		byte[] spotCheckData = check.getClientData();
		byte[] acceptedData = check.getAcceptedData();
		
		// If the length of the two byte arrays is not the same, the spot-check is invalid.
		if (spotCheckData.length != acceptedData.length)
			validSpotCheck = false;
		
		// Do a byte-by-byte comparison of the two byte arrays.
		else {
			for (int ii = 0; ii < spotCheckData.length; ii++) {
				if (spotCheckData[ii] != acceptedData[ii]) {
					validSpotCheck = false;
					break;
				}	
			}
		}
		
		return validSpotCheck;
	}
}
