/**
 * This file contains the implementation of the client class.
 * The client class initializes the XML-RPC client and executes
 * XML-RPCs
 * @author James Baldassari (jdb@cs.wpi.edu)
 */

// System includes
#include <iostream>
#include <cassert>
#include <cstdlib>
#include <cstring>

// XML-RPC includes
#include <xmlrpc-c/client_simple.hpp>
#include <xmlrpc-c/girmem.hpp>

// Local includes
#include "client.hpp"

// Declare which objects from std we're using.
using std::string;
using std::cout;
using std::cerr;
using std::endl;

// Define client to be in the prc namespace
namespace prc {

// Wait 5 minutes before retrying an XML-RPC.
const int prc::Client::RETRY_DELAY = 300;

/**
 * Constructor accepting the URL of the 
 * XML-RPC server in the form: http://host:port/RPC2.
 * @param addr The URL of the XML-RPC server.
 */
Client::Client(string addr) : serverAddr(addr) {
	initializeXMLRPCClient();
}

/**
 * The destructor.  Frees memory allocated to 
 * the XML-RPC objects.
 */
Client::~Client() {
	delete xmlrpcTransport;
	delete xmlrpcClient;
	delete cParm;
}

/**
 * Initializes the XML-RPC client.
 */
void Client::initializeXMLRPCClient() {
	/*
	 * The default limit for the amount of data that can be 
	 * transmitted in the body of an XML-RPC is 512KB, which 
	 * is too small in many cases, especially when large 
	 * work units or results are used.  The validator will 
	 * have to receive multiple results to select a canonical 
	 * result, so 512KB can be much too small in some cases.  
	 * It is necessary to increase this limit.
 	 */
	// Set the limit on XML-RPC data to 32MB per call.
	xmlrpc_limit_set(XMLRPC_XML_SIZE_LIMIT_ID, 1024*1024*32);

	/*
	 * The XML transport specifies what library to use for 
	 * sending the XML over HTTP.  Some options include 
	 * libwww and curl.  The libwww library has worked 
	 * best for me.
	 */
	xmlrpcTransport = new xmlrpc_c::clientXmlTransport_libwww();

	// This is the XML-RPC client that will be used to 
	// execute the XML-RPCs.
	xmlrpcClient = new xmlrpc_c::client_xml(xmlrpcTransport);

	/*
	 * The carraigeParm supplies the URL of the XML-RPC
	 * server to the XML-RPC library.  The URL has the 
	 * form: http://host:port/RPC2
	 */
	cParm = new xmlrpc_c::carriageParm_libwww0(serverAddr);
}

/**
 * Sets the URL of the server.
 * @param serverAddr The URL to set.
 */
void Client::setServerAddress(string serverAddr) {
	this->serverAddr = serverAddr;

	// Reset the carriage parameter.
	delete cParm;
	cParm = new xmlrpc_c::carriageParm_libwww0(serverAddr);
}

/**
 * Executes the given XML-RPC.  If any XML-RPC error occurs, 
 * including low-level errors such as connection refused, this
 * method will throw an std::exception.  If the server throws an 
 * exception while executing the RPC, this method will throw an 
 * std::string containing the error message.
 * @param xmlrpc The XML-RPC to execute.
 * @return Returns the XML-RPC value returned by the server.
 */
xmlrpc_c::value Client::executeRPC(xmlrpc_c::rpcPtr& xmlrpc) {
	xmlrpc_c::value result;

	// Try the XML-RPC call.  Note: The server will throw
	// an exception if the given strings are not valid integers.
	try {
		xmlrpc->call(xmlrpcClient, cParm);
	} catch (std::exception const e) {
		/*
		 * There was some other error, usually caused by 
		 * the connection being refused (i.e. the server
		 * was not running or was listening on a different
		 * port than we connected to).
		 */
		//cerr << e.what() << endl;	// We can print the error description if it helps
		// Rethrow the exception.
		throw e;
	}

	// We can only proceed if xmlrpc->isSuccessful() returns true
	if (xmlrpc->isSuccessful() == true) {
		/*
		 * WARNING: Calling any of the methods below when 
		 * xmlrpc->isSuccessful() returns false, including
		 * xmlrpc->isFinished(), will probably crash your 
		 * program.
		 */

		// xmlrpc->isFinished() must be true before we can 
		// access the return value of the RPC.
		assert(xmlrpc->isFinished());

		/*
		 * Get the return value of the RPC, which will
		 * initially be an xmlrpc_c::value.  We will 
		 * have to cast it to the expected type.
		 */
		result = xmlrpc->getResult();
	}
	else {
		/*
		 * Some other XML-RPC error occurred.
		 * If xmlrpc->isFinished() is true, we can 
		 * get the error code and description.  If 
		 * the error code is zero, it means that the
		 * server threw some exception while executing 
		 * the RPC.
		 */
		if (xmlrpc->isFinished()) {
			xmlrpc_c::fault err = xmlrpc->getFault();
			if (err.getCode() == 0) {
				/*
				 * The server threw an exception because
				 * it could not convert one of the given
				 * strings into an int.
				 */
				string error = "The server threw an exception: " + err.getDescription();
				throw error;
			}
			else {
				std::exception e;
				throw e;
			}
		}
		else {
			std::exception e;
			throw e;
		}
	}

	return result;
}

/**
 * Keeps attempting to execute an XML-RPC until it succeeds.  If it fails due to an 
 * XML-RPC error or some lower-level connection error (like connection refused), an
 * error message will be displayed and the thread will sleep for RETRY_DELAY seconds.
 * It will keep trying to execute the XML-RPC until it is successful.  This method 
 * may throw an std::string if the project server throws an exception while executing 
 * the RPC.
 * @param xmlrpc The XML-RPC to execute.
 * @param errorMessage The message to be displayed when an error occurs.
 * @return Returns the return value of the XML-RPC.
 */
xmlrpc_c::value Client::executeRPCUntilSuccess(xmlrpc_c::rpcPtr& xmlrpc, const string& errorMessage) {
	bool rpcSuccess = false;
	xmlrpc_c::value result;

	// Keep trying until success.
	while (rpcSuccess == false) {
		try {
			result = executeRPC(xmlrpc);
			rpcSuccess = true;
		} catch (std::exception e) {
			// A connection error occurred.
			cerr << errorMessage << ":\n\tError connecting to the remote host." << endl;
			cout << "Retrying in " << RETRY_DELAY / 60 << " minutes." << endl;
			sleep(RETRY_DELAY);
			rpcSuccess = false;
		} catch (string e) {
			/*
			 * Some exceptional condition occurred while the server
			 * was executing the RPC.  This is not a connection error, 
			 * but instead carries important information about the 
			 * result of the RPC, so it will be rethrown.
			 */
			throw e;
		}
	}

	return result;
}

} // end namespace prc
