/**
 * This file contains the implementation of the Server class.
 * The Server class initializes the XML-RPC server, starts the 
 * XML-RPC server in a new thread, and passes RPC requests to 
 * the appropriate RPC handler.
 * @author James Baldassari (jdb@cs.wpi.edu)
 */

// System includes
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <pthread.h>
#include <signal.h>

// XML-RPC includes
#include <xmlrpc-c/base.hpp>
#include <xmlrpc-c/registry.hpp>
#include <xmlrpc-c/server_abyss.hpp>
#include <xmlrpc-c/girmem.hpp>

// Local includes
#include "server.hpp"

// Declare which objects from std we're using.
using std::cout;
using std::cerr;
using std::endl;
using std::string;

// Define Server to be in the prc namespace
namespace prc {

// Initialize constants...
// Initialize the singleton instance to NULL.
Server* prc::Server::serverInstance = NULL;
// Initialize the path to the XML-RPC server log.
const string prc::Server::LOG_PATH = "log/xmlrpc.log";

/**
 * Default constructor.  Initializes the running flag to false 
 * and instantiates a new xmlrpc_c::registry.
 */
Server::Server() : running(false) {
	/*
	 * The default limit for the amount of data that can be 
	 * transmitted in the body of an XML-RPC is 512KB, which 
	 * is too small in many cases, especially when large 
	 * work units or results are used.  The maximum work unit 
	 * size that will be sent to the science application is 4MB, 
	 * which is clearly over the 512KB limit. It is necessary to 
	 * increase this limit.
 	 */
	// Set the limit on XML-RPC data to 32MB per call.
	xmlrpc_limit_set(XMLRPC_XML_SIZE_LIMIT_ID, 1024*1024*32);

	/*
	 * The XML-RPC registry maps RPCs
	 * to their handler functions.
	 */
	xmlrpcRegistry = new xmlrpc_c::registry();
}

/**
 * The destructor.  Frees memory allocated to XML-RPC objects.
 */
Server::~Server() {
	delete abyssServer;
	delete xmlrpcRegistry;
}

/**
 * @return Returns the singleton instance of the Server class.
 */
Server* Server::instance() {
	if (Server::serverInstance == NULL)
		Server::serverInstance = new Server();
	return serverInstance;
}

/**
 * Sets the port on which the server will listen.
 * @param serverPort The server port.
 */
void Server::setServerPort(int serverPort) {
	this->serverPort = serverPort;
}

/**
 * @return Returns true if the server is running, false otherwise.
 */
bool Server::isRunning() const {
	return running;
}

/**
 * Adds an RPC handler to the XML-RPC server.
 * @param rpcName The name of the RPC handler.  Clients will use this name when executing the RPC.
 * @param rpcHandler Pointer to an instance of the RPC handler to add.
 */
void Server::addRPCHandler(string rpcName, xmlrpc_c::method* rpcHandler) {
	xmlrpc_c::methodPtr const handlerMethod(rpcHandler);
	xmlrpcRegistry->addMethod(rpcName, handlerMethod);
}

/**
 * Starts the XML-RPC server in its own thread.
 */
void Server::start() {
	cout << "Server: Starting XML-RPC Server." << endl;

	// Initialize the XML-RPC server.
	abyssServer = new xmlrpc_c::serverAbyss(
		*xmlrpcRegistry, serverPort, Server::LOG_PATH);

	// Start the XML-RPC server thread.
	int status = 
		pthread_create(&serverThread, NULL, Server::runServerThread, abyssServer);
	if (status != 0) {
		cerr << "Server: Error creating XML-RPC server thread.  Exiting." << endl;
		exit(EXIT_FAILURE);
	}

	// Set the server state to running.
	running = true;
}

/**
 * Kills the server thread.
 */
void Server::stop() {
	// Cancel the server thread
	pthread_cancel(serverThread);

	// Wait for the XML-RPC server thread to terminate.
	pthread_join(serverThread, NULL);
}

/**
 * This is the server thread.  It simply starts the XML-RPC server,
 * which will run in this thread until the application is terminated.
 */
void* Server::runServerThread(void* serverPtr) {
	// Cast the thread parameter to an abyss server object.
	xmlrpc_c::serverAbyss* server = static_cast<xmlrpc_c::serverAbyss*>(serverPtr);

	// Start the XML-RPC server in this thread.
	// This thread will run until the server
	// is shut down.
	server->run();

	pthread_exit(EXIT_SUCCESS);
}

/**
 * Shuts down the server by setting the running flag to false.
 */
void Server::shutdown() {
	cout << "Server: Shutting down..." << endl;

	// Set the running flag to false, which will cause the main to exit.
	this->running = false;
}

/**
 * Catches a signal and shuts down the singleton server.
 * @param signum The signal number that was caught.
 */
void Server::shutdown(int signum) {
	cout << "Server: Caught signal " << signum << "." << endl;
	Server::instance()->shutdown();
}

} // end namespace prc
