package edu.wpi.prc.templates.java.common;

import java.io.IOException;
import java.net.MalformedURLException;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import java.util.Vector;

/**
 * The client class is responsible for executing XML-RPCs.
 * @author James Baldassari
 */
public class Client {
	private static final int EXIT_ERROR = 1;
	private static final int RETRY_DELAY = 300000;	// Wait 5 minutes before retrying an RPC.
	private XmlRpcClient xmlrpcClient = null;		// The XML-RPC client used to interact with the remote XML-RPC server.
	private String serverAddress = null;				// The address (host:port) of the remote XML-RPC server. 
	
	/**
	 * Constructor specifying the address of the remote XML-RPC server.
	 * @param serverAddress The address (host:port) of the remote XML-RPC server.
	 */
	public Client(String serverAddress) {
		this.serverAddress = serverAddress;
		initXmlRpcClient();
	}
	
	/**
	 * Initializes the XML-RPC client used for communicating with the project client. 
	 */
	private void initXmlRpcClient() {
		try {
			xmlrpcClient = new XmlRpcClient("http://" + serverAddress + "/RPC2");
		} catch (MalformedURLException e) {
			System.err.println("Invalid URL used to initialize the XML-RPC client: " + e);
			System.exit(EXIT_ERROR);
		}
	}
	
	/**
	 * Sets the address of the remote XML-RPC server.
	 * @param serverAddress The address (host:port) of the remote XML-RPC server.
	 */
	public void setServerAddress(String serverAddress) {
		this.serverAddress = serverAddress;
		initXmlRpcClient();
	}
	
	/**
	 * Executes an XML-RPC.
	 * @param rpcName The name of the RPC to execute.
	 * @param params The parameters to the RPC.
	 * @return Returns the return value of the RPC.
	 * @throws XmlRpcException
	 * @throws IOException
	 */
	public Object executeRPC(String rpcName, Vector params) throws XmlRpcException, IOException {
		Object rpcReturnValue = null;

		// Execute the RPC.
		rpcReturnValue = xmlrpcClient.execute(rpcName, params);
		
		return rpcReturnValue;
	}
	
	/**
	 * Keeps trying to execute the RPC until it is successful.
	 * Note that this does not necessarily mean that no errors 
	 * occurred.  It simply means that the connection to the XML-RPC
	 * server was successful, and the XML-RPC call was attempted.
	 * An RPCHandlerException will be thrown if any XML-RPC errors occur, 
	 * such as if the given RPC was not found on the server, or if the 
	 * remote RPC handler threw an Exception while executing the RPC.
	 * @param rpcName The name of the XML-RPC to execute.
	 * @param params The parameters to pass to the XML-RPC.  This must not be null.
	 * @param errorMessage The heading text to use when printing out any error messages that occur.
	 * @return Returns the value returned by the XML-RPC call.  This must be cast to the expected type.
	 * @throws XmlRpcException
	 * @throws RPCHandlerException
	 */
	public Object executeRPCUntilSuccess(String rpcName, Vector params, String errorMessage) throws RPCHandlerException {
		boolean rpcSuccess = false;
		Object rpcReturnValue = null;
		
		// Keep trying to execute the RPC until it is successful.
		while (rpcSuccess == false) {
			try {
				rpcReturnValue = executeRPC(rpcName, params);
				rpcSuccess = true;
			} catch (XmlRpcException e) {
				// Either an XML-RPC error occured, such as 
				// the given RPC was not found on the server, or
				// the RPC handler itself threw an Exception.
				throw new RPCHandlerException(e.getMessage());
			} catch (IOException e) {
				// A connection error occurred.
				rpcSuccess = false;
				System.err.println(errorMessage + ":\n\tError connecting to the remote host.");
				System.err.println("Retrying in " + (RETRY_DELAY / 1000 / 60) + " minutes.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
				}
			}
		}
		
		return rpcReturnValue;
	}
}
