package edu.wpi.prc.templates.java.generator;

import edu.wpi.prc.templates.java.common.RPCHandlerException;

/**
 * The WorkUnitGenerator contains the project-specific 
 * code to generate work units.  It uses the WorkGeneratorClient
 * to send them to the project server.
 * @author James Baldassari
 */
public class WorkUnitGenerator {
	// TODO: Change the value of PRIORITY, POINTS, and EXPIRATION_TIME if desired.
	// Alternatively, these values can be generated dynamically for each work unit.
	private static final int PRIORITY = 0;				// The PRIORITY of the work units to generate (0 is the minimum).
														// Higher values correspond to higher priorities.
	private static final int POINTS = 1;					// The number of POINTS these work units are worth (1 is the default).
	private static final double EXPIRATION_TIME = 900;	// The number of seconds before each work unit expires (1 day is the default).
	private int workUnitID = 0;							// The ID of the next work unit to be generated.  Used in the addWorkUnit RPC.
	
	/**
	 * Default constructor.
	 * Gets last work unit generated via XML-RPC so 
	 * that it can resume generating from that point.
	 */
	public WorkUnitGenerator() {
		recoverState();
	}
	
	/**
	 * If the work unit generator were shut down, then started back up, it would need to determine the last 
	 * work unit that was generated so that it would know where to resume generating new ranges.  This 
	 * method asks the server for the last work unit that was generated.  If the server throws an Exception, 
	 * it means that no work units had ever been generated, in which case the work unit generator is starting 
	 * for the first time.
	 */
	private void recoverState() {
		WorkUnit wu = null;

		try {
			// Get the last work unit from the server.
			wu = WorkGeneratorClient.instance().getLastWorkUnit();

			// NOTE: If no Exception was thrown by this point, the server returned a work unit.
			
			// Initialize the next work unit ID by getting the last work unit ID and increasing it by one.
			workUnitID = Integer.parseInt(wu.getWorkUnitID().substring(wu.getWorkUnitID().lastIndexOf("_") + 1)) + 1;
			
			// We are interested in getting the data from the last work unit that was generated 
			// so that we can resume from there.
			byte[] data = wu.getData();
			
			// TODO: Inspect the data byte array to determine where in the science data to 
			// resume generating work units.
			
			// TODO: Set the value of some private variable so that the generateWorkUnit() 
			// method will know where to start generating the next work unit.
			
		} catch (RPCHandlerException e) {
			// The server had zero work units.
			// This means that we should start generating
			// our first work unit.
			// workUnitID will be incremented when the work unit is 
			// generated, so the first work unit will have ID 1, not 0.
			workUnitID = 0;
		}
	}
	
	/**
	 * Generates the next work unit. 
	 * @return Returns a byte array representing the generated work unit.
	 */
	@SuppressWarnings("unchecked")
	private byte[] generateWorkUnit() {
		// The work unit data.
		byte[] data = null;
		
		// TODO: Generate the work unit data and store it in the "data" array.
		// TODO: Update some private variable to indicate how the next work unit should be generated.
		
		// Increment the work unit ID.
		workUnitID++;
		
		return data;
	}
	
	/**
	 * Adds a new work unit to the server by calling the server.addWorkUnit XML-RPC.
	 * @param workUnit The work unit to send to the server.
	 */
	private void sendWorkUnit(byte[] workUnit) {
		// Build the work unit ID
		String id = "WU_" + workUnitID;
		
		// TODO: If you do not need to specify the ID, priority, point value, or 
		// expiration time of the work unit, you can choose to use a different 
		// sendWorkUnit method from WorkGeneratorClient.
		
		// NOTE: There are several variants of the "WorkGeneratorClient.sendWorkUnit" method.
		// The simplest case only takes the data byte array.  This example 
		// specifies several parameters including ID, data, priority, point value, 
		// and expiration time.
		// See WorkGeneratorClient for other versions of the RPC.
		WorkGeneratorClient.instance().sendWorkUnit(id, workUnit, PRIORITY, POINTS, EXPIRATION_TIME);
		
		// Here are some other methods that can be used:
		// WorkGeneratorClient.instance().sendWorkUnit(id, workUnit, PRIORITY, POINTS);
		// WorkGeneratorClient.instance().sendWorkUnit(workUnit, PRIORITY, POINTS, EXPIRATION_TIME);
		// WorkGeneratorClient.instance().sendWorkUnit(workUnit, PRIORITY, POINTS);
		// WorkGeneratorClient.instance().sendWorkUnit(workUnit);
	}
	
	/**
	 * Generates the specified number of work units, and sends them to the project server.
	 * @param numWorkUnits The number of work units to generate and send.
	 */
	public void generateAndSend(int numWorkUnits) {
		System.out.println("Generating " + numWorkUnits + " new work units, then sending them to the server.");
		for (int ii = 0; ii < numWorkUnits; ii++) {
			byte[] workUnit = generateWorkUnit();
			sendWorkUnit(workUnit);
		}
	}
}
