package edu.wpi.prc.templates.java.validator;

import java.util.Vector;

import edu.wpi.prc.templates.java.common.Client;
import edu.wpi.prc.templates.java.common.RPCHandlerException;

/**
 * The ValidatorClient is an XML-RPC client with 
 * added functionality that is needed by the result validator.
 */
public class ValidatorClient extends Client {
	private static final String GET_RESULT_RPC_NAME = "server.getResultForValidation";
	private static final String MARK_VALID_RPC_NAME = "server.markResultValid";
	private static final String MARK_INVALID_RPC_NAME = "server.markResultInValid";
	private static final String SELECT_CANONICAL_RPC_NAME = "server.selectCanonicalResult";
	private static final String GET_ASSOCIATED_WU_RPC_NAME = "server.getAssociatedWorkUnit";
	private static final String DEFAULT_SERVER_ADDRESS = "127.0.0.1:2080";
	private static ValidatorClient client = null;
	ResultValidator validator = null;
	
	/**
	 * Private constructor because the ValidatorClient is a singleton.
	 */
	private ValidatorClient() {
		super(DEFAULT_SERVER_ADDRESS);
		validator = new ResultValidator();
	}
	
	/**
	 * Gets the next result to be validated from the server, then 
	 * invokes the appropriate method of the ResultValidator, depending
	 * on what type of result needed to be validated.  When the ResultValidator
	 * has finished validating the result, this method will inform the project
	 * server of the outcome.
	 */
	public void validateNextResult() throws RPCHandlerException {
		String errorMessage = "Error while requesting result for validation";
		
		try {
			// Execute the RPC to get the next result to validate.
			Object rpcResult = executeRPCUntilSuccess(GET_RESULT_RPC_NAME, new Vector(), errorMessage);
			Vector rpcVector = (Vector)rpcResult;
			
			// Determine what type of valication needs to occur.
			int type = (Integer)rpcVector.get(0);

			// Validate a single result.
			if (type == 0) {
				Result r = new Result((String)rpcVector.get(1), (byte[])rpcVector.get(2));
				
				// Use the ResultValidator to determine whether the result was valid.
				if (validator.validateSingleResult(r) == true) {
					// If it was valid, mark it valid on the server.
					if (markResultValid(r.getResultID()) == false) {
						System.err.println("Error marking result " + r.getResultID() + " valid.  ");
						System.err.println("The server could not find the specified result ID.");
					}
				}
				else {
					// Otherwise, mark it invalid on the server.
					if (markResultInvalid(r.getResultID()) == false) {
						System.err.println("Error marking result " + r.getResultID() + " invalid.  ");
						System.err.println("The server could not find the specified result ID.");
					}
				}
			}
			
			// Select a canonical result.			
			else if (type == 1) {
				int numResults = (Integer)rpcVector.get(1);
				Vector<Result> results = new Vector<Result>();
				
				// Build the vector of results.
				for (int ii = 2; ii <= numResults * 2; ii += 2) {
					Result r = new Result((String)rpcVector.get(ii), (byte[])rpcVector.get(ii+1));
					results.add(r);
				}
				
				// Select the canonical result with the ResultValidator.
				String canonicalID = validator.selectCanonicalResult(results);

				// Inform the server of the canonical result choice.
				if (selectCanonicalResult(canonicalID) == false) {
					System.err.println("Error selecting canonical result " + canonicalID + " invalid.  ");
					System.err.println("The server could not find the specified result ID.");
				}
			}

			// Validate a spot-check.
			else if (type == 2) {
				SpotCheck check = new SpotCheck((String)rpcVector.get(1), (byte[])rpcVector.get(2), (byte[])rpcVector.get(3));

				// Use the ResultValidator to determine whether the spot-check was valid.
				if (validator.validateSpotCheck(check) == true) {
					// If it was valid, mark it valid on the server.
					if (markResultValid(check.getResultID()) == false) {
						System.err.println("Error marking spot-check " + check.getResultID() + " valid.  ");
						System.err.println("The server could not find the specified result ID.");
					}
					else {
						// Otherwise, mark it invalid on the server.
						if (markResultInvalid(check.getResultID()) == false) {
							System.err.println("Error marking spot-check " + check.getResultID() + " invalid.  ");
							System.err.println("The server could not find the specified result ID.");
						}
					}
				}
			}
			
		} catch (ClassCastException e) {
			//System.out.println("The server returned: " + e);
			System.out.println("The server did not have any results that needed to be validated.");
			// Rethrow the exception, to be caught by the main.
			throw new RPCHandlerException(e.getMessage());
		} catch (RPCHandlerException e) {
			//System.out.println("The server returned: " + e);
			System.out.println("The server did not have any results that needed to be validated.");
			// Rethrow the exception, to be caught by the main.
			throw e;
		}
	}
	
	/**
	 * Marks the result with the given ID as valid.
	 * @param resultID The ID of the result to mark valid.
	 * @return Returns true if the result was successfully marked valid, 
	 * false if no result with the given ID was found.
	 */
	@SuppressWarnings("unchecked")
	public boolean markResultValid(String resultID) {
		String errorMessage = "Error while marking result valid";
		Vector params = new Vector<String>(1);
		params.add(resultID);
		boolean success = false;
		
		// Execute the RPC.
		Object rpcResult = null;
		try {
			rpcResult = executeRPCUntilSuccess(MARK_VALID_RPC_NAME, params, errorMessage);
			success = (Boolean)rpcResult;
		} catch (ClassCastException e) {
			System.out.println("The server returned an unexpected error: " + e);
			success = false;
		} catch (RPCHandlerException e) {
			System.err.println("The server returned: " + e);
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Marks the result with the given ID as valid.
	 * @param resultID The ID of the result to mark valid.
	 * @return Returns true if the result was successfully marked valid, 
	 * false if no result with the given ID was found.
	 */
	@SuppressWarnings("unchecked")
	public boolean markResultInvalid(String resultID) {
		String errorMessage = "Error while marking result invalid";
		Vector params = new Vector<String>(1);
		params.add(resultID);
		boolean success = false;
		
		// Execute the RPC.
		Object rpcResult = null;
		try {
			rpcResult = executeRPCUntilSuccess(MARK_INVALID_RPC_NAME, params, errorMessage);
			success = (Boolean)rpcResult;
		} catch (ClassCastException e) {
			System.out.println("The server returned an unexpected error: " + e);
			success = false;
		} catch (RPCHandlerException e) {
			System.err.println("The server returned: " + e);
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Selects the result with the given ID as the canonical result for a work unit.
	 * @param resultID The ID of the result to select.
	 * @return Returns true if the result was successfully selected, 
	 * false if no result with the given ID was found.
	 */
	@SuppressWarnings("unchecked")
	public boolean selectCanonicalResult(String resultID) {
		String errorMessage = "Error while selecting canonical result";
		Vector params = new Vector<String>(1);
		params.add(resultID);
		boolean success = false;
		
		// Execute the RPC.
		Object rpcResult = null;
		try {
			rpcResult = executeRPCUntilSuccess(SELECT_CANONICAL_RPC_NAME, params, errorMessage);
			success = (Boolean)rpcResult;
		} catch (RPCHandlerException e) {
			System.err.println("The server returned: " + e);
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Gets the associated work unit for the given result.
	 * @param resultID The result for which to get the associated work unit.
	 * @return Returns the data contained in the associated work unit.
	 */
	@SuppressWarnings("unchecked")
	public byte[] getAssociatedWorkUnit(String resultID) {
		String errorMessage = "Error while getting associated work unit";
		Vector params = new Vector<String>(1);
		params.add(resultID);
		byte[] workUnitData = null;
		
		// Execute the RPC.
		Object rpcResult = null;
		try {
			rpcResult = executeRPCUntilSuccess(GET_ASSOCIATED_WU_RPC_NAME, params, errorMessage);
			workUnitData = (byte[])((Vector)rpcResult).get(0);
		} catch (ClassCastException e) {
			System.out.println("The server returned an unexpected error: " + e);
			workUnitData = null;
		} catch (RPCHandlerException e) {
			System.err.println("The server returned: " + e);
			workUnitData = null;
		}
		
		return workUnitData;
	}
	
	/**
	 * @return Returns the singleton instance of the ValidatorClient class.
	 */
	public static ValidatorClient instance() {
		if (client == null)
			client = new ValidatorClient();
		return client;
	}
}
