package edu.wpi.prc.persistence;

import java.io.File;
import java.util.Properties;

import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.tool.hbm2ddl.SchemaExport;
import org.hibernate.tool.hbm2ddl.SchemaUpdate;

import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;

/**
 * The HibernateController class controls access to important Hibernate functions.
 * @author James Baldassari
 */
public class HibernateController {
	private static HibernateController hibernateController = null;
	private DatabaseType dbType = DatabaseType.HSQLDB;
	private Properties hibernateProperties = null;
	private Configuration config = null;
	private SessionFactory sessions = null;
	private boolean isConfigured = false;
	
	/**
	 * Private constructor prevents direct instantiation of this class.
	 */
	private HibernateController() {
		
	}

	/**
	 * @return Returns the hibernateProperties.
	 */
	public Properties getHibernateProperties() {
		return hibernateProperties;
	}
	
	/**
	 * @param hibernateProperties The hibernateProperties to set.
	 */
	public void setHibernateProperties(Properties hibernateProperties) {
		this.hibernateProperties = hibernateProperties;
	}

	/**
	 * Reads project configuration from the project properties file, then translates it into Hibernate properties.
	 * @return Returns a Properties object with Hibernate configuration information.
	 * @throws MissingConfigurationFileException
	 */
	private Properties readConfigFromFile() throws MissingConfigurationFileException {
		Project proj = Project.readFromFile();
		return readProjectProperties(proj);
	}
	
	/**
	 * Reads project configuration from the project properties file, then translates it into Hibernate properties.
	 * @param configFilePath to the project configuration file.
	 * @return Returns a Properties object with Hibernate configuration information.
	 * @throws MissingConfigurationFileException
	 */	
	private Properties readConfigFromFile(String configFilePath) throws MissingConfigurationFileException {
		Project proj = Project.readFromFile(configFilePath);
		return readProjectProperties(proj);
	}
	
	/**
	 * Configures Hibernate by inspecting a project.
	 * @param proj The project to use to configure Hibernate.
	 * @return Returns a properties object containing Hibernate configuration information.
	 */
	private Properties readProjectProperties(Project proj) {
		Properties props = new Properties();
		
		if (proj.getDbType() != null)
			dbType = proj.getDbType();
		else {
			System.err.println("WARNING: Database type not specified.  Defaulting to HSQLDB.");
			dbType = DatabaseType.HSQLDB;
		}
		
		if (proj.getDbURI() != null) {
			String connStr = "jdbc:" + dbType.toString().toLowerCase();
			if (dbType == DatabaseType.HSQLDB)
				props.setProperty("hibernate.connection.url", connStr + ":file:" + proj.getDbURI());
			else
				props.setProperty("hibernate.connection.url", connStr + "://" + proj.getDbURI());
		}
		else {
			System.err.println("WARNING: Database URI not specified.  Defaulting to data" + File.separator + "prc.");
			props.setProperty("hibernate.connection.url", ":file:data" + File.separator + "prc");
		}
		
		props.setProperty("hibernate.connection.username", proj.getDbUserName());
		props.setProperty("hibernate.connection.password", proj.getDbPassword());
		
		return props;
	}
	
	/**
	 * Sets the type of database hibernate will use if none is specified.
	 * @param db The default database type to set.
	 */
	public void setDatabaseType(DatabaseType db) {
		dbType = db;
	}
	
	/**
	 * Initializes hibernate with the specified Hibernate configuration file.
	 * @param db The type of database to use.
	 * @throws MissingConfigurationFileException
	 * @throws HibernateException
	 */
	public void initialize(DatabaseType db) throws MissingConfigurationFileException, HibernateException {
		dbType = db;
		initialize();
	}
	
	/**
	 * Initializes hibernate using an alternate project configuration file path.
	 * @param configFilePath Path to the project configuration file.
	 * @throws MissingConfigurationFileException
	 * @throws HibernateException
	 */
	public void initialize(String configFilePath) throws MissingConfigurationFileException, HibernateException {
		if (sessions == null) {
			hibernateProperties = readConfigFromFile(configFilePath);
			setConfiguration();
		}
	}
	
	/**
	 * Initializes hibernate.
	 * @throws MissingConfigurationFileException
	 * @throws HibernateException
	 */
	public void initialize() throws MissingConfigurationFileException, HibernateException {
		if (sessions == null) {
			hibernateProperties = readConfigFromFile();
			setConfiguration();
		}
	}
	
	/**
	 * @return Returns true if Hibernate has been configured and initialized, false otherwise.
	 */
	public boolean isConfigured() {
		return isConfigured;
	}
	
	/**
	 * Sets the Hibernate configuration that was extracted from the project.
	 * @throws HibernateException
	 */
	private void setConfiguration() throws HibernateException {
		System.out.println("Configuring Hibernate for " + dbType + ".");
		config = new Configuration();
		config.setProperties(hibernateProperties);
		config.configure(dbType.getConfigurationFileName());
		sessions = config.buildSessionFactory();
		isConfigured = true;
	}
	
	/**
	 * Initializes Hibernate for use with the JUnit tests.
	 * Hibernate is configured to use the HSQLDB database in data/prc.
	 * Using this method instead of initialize() frees the unit tests from 
	 * the need to have a project configuration file.
	 */
	public void initializeForUnitTests() {
		if (sessions == null) {		
			// Set the database to HSQLDB.
			dbType = DatabaseType.HSQLDB;
			System.out.println("Configuring Hibernate for " + dbType + ".");
			
			// Store HSQLDB database in data/prc
			hibernateProperties = new Properties();
			hibernateProperties.setProperty("hibernate.connection.url", "jdbc:" + dbType.toString().toLowerCase() + ":file:data" + File.separator + "prctests");
			hibernateProperties.setProperty("hibernate.connection.username", "sa");
			hibernateProperties.setProperty("hibernate.connection.password", "");
			
			// Configure Hibernate
			config = new Configuration();
			config.setProperties(hibernateProperties);
			config.configure(dbType.getConfigurationFileName());
			sessions = config.buildSessionFactory();
			isConfigured = true;
		}
	}
	
	/**
	 * Creates a new Hibernate session to use for database access. 
	 * @return A new Hibernate session.
	 */
	public Session createSession() {
		return sessions.openSession();
	}
	
	/**
	 * Exports the schema to the database.  WARNING: This process will 
	 * first drop all tables in the database, erasing all data!
	 * @throws HibernateException
	 */
	public void exportSchema() throws HibernateException {
		assert(isConfigured == true);
		new SchemaExport(config).create(false, true);
	}
	
	/**
	 * Updates the database schema.  This process should not result 
	 * in any data loss.
	 * @throws HibernateException
	 */
	public void updateSchema() throws HibernateException {
		assert(isConfigured == true);
		new SchemaUpdate(config).execute(false, true);
	}
	
	/**
	 * Shuts down Hibernate and closes all open sessions.
	 */
	public void shutdown() {
		if (sessions != null) {
			sessions.close();
			sessions = null;
			isConfigured = false;
		}
	}
	
	/**
	 * @return Returns the singleton instance of the HibernateController class.
	 */
	public static HibernateController instance() {
		if (hibernateController == null)
			hibernateController = new HibernateController();
		return hibernateController;
	}
}
