package edu.wpi.prc.project;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Properties;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;

import edu.wpi.prc.persistence.DatabaseType;
import edu.wpi.prc.persistence.HibernateController;

/**
 * The Project class stores information specific to the PRC project.
 * @author James Baldassari
 */
public class Project {
	private static Project project = null;			// Static instance of this singleton class
	
	// Constants
	private static final String configFilePath = "cfg" + File.separator + "project.properties";
	private static final String clientDir = "client";
	private static final String serverDir = "server";
	private static final String generatorProgram = "generator";
	private static final String resultValidatorProgram = "resultvalidator";
	private static final String unixExtension = "sh";
	private static final String winExtension = "bat";
	
	private long id;									// Hibernate ID.
	private String name = null;						// The name of the project.
	private String password = null;					// The project password.
	private String serverURI = null;					// Host name and port to which clients will connect.
	private int clientPort = 0;						// Port on which the client's XML-RPC server will listen.
	private String scienceApp = "run_sci_app";		// Path to science application.
	private int scienceAppPort = 0;					// Port on which the science application's XML-RPC server is listening.
	private String transitionerURI = null;			// For remote transitioners.
	private boolean defaultValidator = true;			// Use the default validator?
	private String dbTypeStr = null;					// The type of database used by this project.
	private String dbURI = null;						// Path to database, either file or connection string.
	private String dbUserName = null;				// Database user name.
	private String dbPassword = null;				// Database password.
	private int minNumResults = 0;					// Minimum number of results for each work unit.
	private float spotCheckProbability = 0.0f;		// Probability that a user will be given a spot check work unit (between 0.0f and 1.0f).
	private int minNumSpotChecks = 0;				// Minimum number of spot checks for the PRC server to produce.
	private String failedSpotCheckActionStr = SpotCheckFailureAction.Log.toString();		// What to do when a user fails a spot check.
	
	/**
	 * Protected constructor because this class is a singleton
	 * but needs to be constructed by Hibernate.
	 */
	protected Project() {
		
	}
	
	/**
	 * @return Returns the id.
	 */
	protected long getId() {
		return id;
	}

	/**
	 * @param id The id to set.
	 */
	protected void setId(long id) {
		this.id = id;
	}
	
	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * @param name The name to set.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return Returns the password.
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * @param password The password to set.
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * @return Returns the dbTypeStr.
	 */
	protected String getDbTypeStr() {
		return dbTypeStr;
	}

	/**
	 * @param dbTypeStr The dbTypeStr to set.
	 */
	protected void setDbTypeStr(String dbTypeStr) {
		this.dbTypeStr = dbTypeStr;
	}

	/**
	 * @return Returns the type of database used by the project, or null if no database type was defined.
	 */
	public DatabaseType getDbType() {
		try {
			return Enum.valueOf(DatabaseType.class, dbTypeStr);
		} catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * @param dbType The dbType to set.
	 */
	public void setDbType(DatabaseType dbType) {
		this.dbTypeStr = dbType.toString();
	}

	/**
	 * @return Returns the serverURI.
	 */
	public String getServerURI() {
		return serverURI;
	}

	/**
	 * @param projectURI The serverURI to set.
	 */
	public void setServerURI(String projectURI) {
		this.serverURI = projectURI;
	}

	/**
	 * @return Returns the clientPort.
	 */
	public int getClientPort() {
		return clientPort;
	}

	/**
	 * @param clientPort The clientPort to set.
	 */
	public void setClientPort(int clientPort) {
		this.clientPort = clientPort;
	}

	/**
	 * @return Returns the scienceApp.
	 */
	public String getScienceApp() {
		return scienceApp;
	}
	
	/**
	 * @param scienceApp The scienceApp to set.
	 */
	public void setScienceApp(String scienceApp) {
		this.scienceApp = scienceApp;
	}

	/**
	 * @return Returns the scienceAppPort.
	 */
	public int getScienceAppPort() {
		return scienceAppPort;
	}

	/**
	 * @param scienceAppPort The scienceAppPort to set.
	 */
	public void setScienceAppPort(int scienceAppPort) {
		this.scienceAppPort = scienceAppPort;
	}

	/**
	 * @return Returns the transitionerURI.
	 */
	public String getTransitionerURI() {
		return transitionerURI;
	}

	/**
	 * @param transitionerURI The transitionerURI to set.
	 */
	public void setTransitionerURI(String transitionerURI) {
		this.transitionerURI = transitionerURI;
	}

	/**
	 * @return Returns the defaultValidator.
	 */
	public boolean getDefaultValidator() {
		return defaultValidator;
	}

	/**
	 * @param defaultValidator The defaultValidator to set.
	 */
	public void setDefaultValidator(boolean defaultValidator) {
		this.defaultValidator = defaultValidator;
	}

	/**
	 * @return Returns the dbPath.
	 */
	public String getDbURI() {
		return dbURI;
	}
	
	/**
	 * @param dbPath The dbPath to set.
	 */
	public void setDbURI(String dbPath) {
		this.dbURI = dbPath;
	}
	
	/**
	 * @return Returns the dbUserName.
	 */
	public String getDbUserName() {
		return dbUserName;
	}

	/**
	 * @param dbUserName The dbUserName to set.
	 */
	public void setDbUserName(String dbUserName) {
		this.dbUserName = dbUserName;
	}

	
	/**
	 * @return Returns the dbPassword.
	 */
	public String getDbPassword() {
		return dbPassword;
	}

	/**
	 * @param dbPassword The dbPassword to set.
	 */
	public void setDbPassword(String dbPassword) {
		this.dbPassword = dbPassword;
	}

	/**
	 * @return Returns the minNumResults.
	 */
	public int getMinNumResults() {
		return minNumResults;
	}

	/**
	 * @param minNumResults The minNumResults to set.
	 */
	public void setMinNumResults(int minNumResults) {
		this.minNumResults = minNumResults;
	}
	
	/**
	 * @return Returns the spotCheckProbability.
	 */
	public float getSpotCheckProbability() {
		return spotCheckProbability;
	}

	/**
	 * @param spotCheckProbability The spotCheckProbability to set.
	 */
	public void setSpotCheckProbability(float spotCheckProbability) {
		this.spotCheckProbability = spotCheckProbability;
	}

	/**
	 * @return Returns the minNumSpotChecks.
	 */
	public int getMinNumSpotChecks() {
		return minNumSpotChecks;
	}

	/**
	 * @param minNumSpotChecks The minNumSpotChecks to set.
	 */
	public void setMinNumSpotChecks(int minNumSpotChecks) {
		this.minNumSpotChecks = minNumSpotChecks;
	}

	/**
	 * @return Returns the failedSpotCheckActionStr.
	 */
	protected String getFailedSpotCheckActionStr() {
		return failedSpotCheckActionStr;
	}

	/**
	 * @param failedSpotCheckActionStr The failedSpotCheckActionStr to set.
	 */
	protected void setFailedSpotCheckActionStr(String failedSpotCheckActionStr) {
		this.failedSpotCheckActionStr = failedSpotCheckActionStr;
	}
	
	/**
	 * @return Returns the action to take when a user fails a spot check.
	 */
	public SpotCheckFailureAction getSpotCheckFailureAction() {
		return Enum.valueOf(SpotCheckFailureAction.class, failedSpotCheckActionStr);
	}
	
	/**
	 * Sets the action to take when a user fails a spot check.
	 * @param action The action to take.
	 */
	public void setSpotCheckFailureAction(SpotCheckFailureAction action) {
		failedSpotCheckActionStr = action.toString();
	}

	/**
	 * Reads the "os.name" system property and parses it to determine 
	 * what operating system is being used.
	 * @return The type of operating system used.
	 */
	private static OperatingSystemType getOSType() {
		String os = System.getProperty("os.name");
		if (os.indexOf("Windows") != -1)
			return OperatingSystemType.WINDOWS;
		else
			return OperatingSystemType.UNIX;
	}
	
	/**
	 * Returns the operating system specific file extension for scripts.
	 * @return The file extension for scripts.
	 */
	private static String getScriptExtension() {
		if (getOSType() == OperatingSystemType.WINDOWS)
			return winExtension;
		else
			return unixExtension;
	}
	
	/**
	 * @return Returns the absolute path to the current working directory, which is assumed to be the project directory.
	 */
	private static String getProjectPath() {
		File cwd = new File(".");
		return cwd.getAbsolutePath();
	}
	
	/**
	 * Returns the path to the client directory.
	 * @return The path to the client directory.
	 */
	public String getClientDirectoryPath() {
		return (getProjectPath() + File.separator + clientDir);
	}
	
	/**
	 * Returns the path to the server directory.
	 * @return The path to the server directory.
	 */
	public String getServerDirectoryPath() {
		return (getProjectPath() + File.separator + serverDir);
	}
	
	/**
	 * Returns the path to the project specific work unit generator program.
	 * @return The path to the work unit generator program.
	 */
	public String getGeneratorPath() {
		return (getProjectPath() + File.separator + serverDir + File.separator + generatorProgram + getScriptExtension());
	}
	
	/**
	 * Returns the path to the project specific result validator program.
	 * @return The path to the result validator program.
	 */
	public String getResultValidatorPath() {
		return (getProjectPath() + File.separator + serverDir + File.separator + resultValidatorProgram + getScriptExtension());
	}
	
	/**
	 * Returns the path to the project specific science application.
	 * @return The path to the science application.
	 */
	public String getScienceAppPath() {
		return (getProjectPath() + File.separator + clientDir + File.separator + scienceApp + "." + getScriptExtension());
	}
	
	/**
	 * Saves this Project to the database or updates it if it already exists.
	 */
	public void saveOrUpdate() throws HibernateException {
		 Session hibernateSession = HibernateController.instance().createSession();
		 Transaction tx = null;
		 try {
			 // Start new transaction, save this Project, then commit the transaction.
			 tx = hibernateSession.beginTransaction();
			 hibernateSession.saveOrUpdate(this);
			 tx.commit();
		 } finally {
			 // Clean up the session.
			 hibernateSession.close();
		 }
	}
	
	/**
	 * Looks for the project configuration file in the default location.
	 * @return Returns true if the file exists, false otherwise.
	 */
	public static boolean projectFileExists() {
		File projectConfigFile = new File(configFilePath);
		return projectConfigFile.exists();
	}
	
	/**
	 * Retrieves the project settings from the configuration file rather than the database 
	 * using the default location for the configuration file.
	 * @return Returns a new project with values read from the configuration file.
	 * @throws MissingConfigurationFileException
	 */
	public static Project readFromFile()  throws MissingConfigurationFileException {
		return readFromFile(configFilePath);
	}
	
	/**
	 * Writes all configuration information that is necessary for server components to a file. 
	 * @param filePath Path to the file that should be written.
	 * @throws IOException
	 */
	public void writeServerProjectFile(String filePath) throws IOException {
		Properties props = new Properties();
		props.setProperty("server_URI", serverURI);
		if (transitionerURI != null)
			props.setProperty("transitioner_URI", transitionerURI);
		props.setProperty("db_type", dbTypeStr);
		props.setProperty("db_URI", dbURI);
		props.setProperty("db_username", dbUserName);
		if (dbPassword != null)
			props.setProperty("db_password", dbPassword);
		
		props.store(new FileOutputStream(filePath), null);
	}
	
	/**
	 * Writes all configuration information that is necessary for client components to a file. 
	 * @param filePath Path to the file that should be written.
	 * @throws IOException
	 */
	public void writeClientProjectFile(String filePath) throws IOException {
		Properties props = new Properties();
		props.setProperty("server_URI", serverURI);
		props.setProperty("client_port", new Integer(clientPort).toString());
		props.setProperty("science_app", scienceApp);
		props.setProperty("science_app_port", new Integer(scienceAppPort).toString());
		
		props.store(new FileOutputStream(filePath), null);
	}
	
	/**
	 * Retrieves the project settings from the configuration file rather than the database 
	 * allowing the path of the configuration file to be specified.
	 * @param filePath The path to the project configuration file.
	 * @return Returns a new project with values read from the configuration file.
	 * @throws MissingConfigurationFileException
	 */
	public static Project readFromFile(String filePath) throws MissingConfigurationFileException {
		Project proj = new Project();
		
		// Open the configuration file and parse the values.
		try {
			InputStream is;
			is = new FileInputStream(filePath);
			Properties props = new Properties();
			props.load(is);
			
			// Read the properties from the file
			String serverURI = props.getProperty("server_URI", "");
			String clientPort = props.getProperty("client_port", "2081");
			String scienceApp = props.getProperty("science_app", "");
			String scienceAppPort = props.getProperty("science_app_port", "2082");
			String transitionerURI = props.getProperty("transitioner_URI", "");
			String dbType = props.getProperty("db_type", "");
			String dbURI = props.getProperty("db_URI", "");
			String dbUserName = props.getProperty("db_username", "sa");
			String dbPassword = props.getProperty("db_password", "");
			
			// Set the new Project's properties
			
			if (!serverURI.equals(""))
				proj.setServerURI(serverURI);
			
			try {
				proj.setClientPort(Integer.parseInt(clientPort));
			} catch (Exception e) {
				System.err.println("Invalid client_port value \"" + scienceAppPort + "\" specified in configuration file.");
				proj.setClientPort(2081);
			}
			
			if (!scienceApp.equals(""))
				proj.setScienceApp(scienceApp);
			
			try {
				proj.setScienceAppPort(Integer.parseInt(scienceAppPort));
			} catch (Exception e) {
				System.err.println("Invalid science_app_port value \"" + scienceAppPort + "\" specified in configuration file.");
				proj.setScienceAppPort(2082);
			}
			
			if (!transitionerURI.equals(""))
				proj.setTransitionerURI(transitionerURI);
			else
				proj.setTransitionerURI(null);
			
			if (!dbType.equals("")) {
				try {
					proj.setDbType(Enum.valueOf(DatabaseType.class, dbType));
				} catch (Exception e) {
					System.err.println("Invalid database_type value \"" + dbType + "\" specified in configuration file.");
				}
			}

			if (!dbURI.equals("")) {
				// Make sure the path to the HSQLDB database is correct for our platform.
				// Replace all slashes with File.separator.  NOTE: regular expressions in Java do not work very well.
				if (proj.getDbType() == DatabaseType.HSQLDB) {
					String[] unixParts = dbURI.split("/");
					if (unixParts.length > 1) {
						String fixedURI = new String(unixParts[0]);
						for (int ii = 1; ii < unixParts.length; ii++)
							fixedURI += (File.separator + unixParts[ii]);
						dbURI = fixedURI;
					}
					String[] winParts = dbURI.split("\\\\");
					if (winParts.length > 1) {
						String fixedURI = new String(winParts[0]);
						for (int ii = 1; ii < winParts.length; ii++)
							fixedURI += (File.separator + winParts[ii]);
						dbURI = fixedURI;
					}
				}
				
				proj.setDbURI(dbURI);
			}
			
			proj.setDbUserName(dbUserName);
			proj.setDbPassword(dbPassword);
		} catch (FileNotFoundException e) {
			throw new MissingConfigurationFileException("No project configuration file found or insufficient permissions to access the file.");
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return proj;
	}
	
	/**
	 * Attempts to read the project configuration from the database specified in the project configuration file. 
	 * @return Returns the project that was stored in the database, or null if no project was found.
	 * @throws HibernateException
	 */
	public static Project readFromDatabase() throws HibernateException {
		Project proj = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		Iterator projectIterator = null;
		try {
			Criteria query = hibernateSession.createCriteria(Project.class);
			projectIterator = query.list().iterator();
			if (projectIterator.hasNext() == true)
				proj = (Project)projectIterator.next();
		} finally {
			// Clean up the session.
			hibernateSession.close();
		}
		
		return proj;
	}

	/**
	 * There is one case in which it is necessary to create a new project without reading project 
	 * information from the database or from the project.properties file: when creating a new 
	 * project from the ProjectBuilder.  It is only for this reason that the createNewProject() method 
	 * is available.  All other code should use Project.instance() or Project.readFromFile(). 
	 * @return Returns a new Project initialized with default values.
	 */
	public static Project createProject() {
		return new Project();
	}
	
	/**
	 * Retrieves the project information from the database and 
	 * returns it as a Project object.
	 * @return Project object containing project information from the database.
	 */
	public static Project instance() {
		// If the project is null, try to get the project information from the database.
		if (project == null) {
			project = readFromDatabase();
		}
		
		// If the Project was never stored in the database, or the project configuration 
		// file does not exist, create a new Project.
		if (project == null)
			project = new Project();
		
		return project;
	}
}
