package edu.wpi.prc.server;

import java.util.Date;
import java.util.List;
import java.util.Vector;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;
import edu.wpi.prc.user.VolunteerState;

/**
 * Processes all client XML-RPC requests by delegating to the appropriate classes.
 * @author James Baldassari
 */
public class ServerRPCHandler {
	/*
	 * The numTransactions variable is a count of the number of outstanding 
	 * XML-RPC requests that will result in data or state being modified.
	 * These transactions include adding a work unit to the system, marking 
	 * a result as valid, retrieving a new work unit for a client, etc.
	 * The server should not shut down before all transactions of this type 
	 * have completed, so the numTransactions variable can be used by the 
	 * PRCServer to determine when it is safe to shut down (when 
	 * numTransactions = 0).  
	 */
	private volatile int numTransactions = 0;
	
	/**
	 * @return Returns the number of outstanding transactions that 
	 * can cause changes in data and/or state.
	 */
	public int getNumTransactions() {
		return numTransactions;
	}
	
	/**
	 * Begins a transaction.  This method will cause the calling thread to block 
	 * if the transitioner is not running. 
	 */
	private synchronized void beginTransaction() {
		if (PRCServer.instance().getTransitionerSemaphore().availablePermits() < 1) {
			SysLogger.println("ServerRPCHandler: No available permits for transitioner semaphore.");
			try {
				PRCServer.instance().getTransitionerSemaphore().acquire();
			} catch (InterruptedException e) {
				
			} finally {
				PRCServer.instance().getTransitionerSemaphore().release();
			}
		}
		
		numTransactions++;
	}
	
	/**
	 * Ends a transaction.
	 */
	private synchronized void endTransaction() {
		if (numTransactions > 0)
			numTransactions--;
	}
	
	/**
	 * Gets the next available work unit to be computed by the given user.
	 * @param userID The volunteer's user ID.
	 * @param eMailAddress The user's e-mail address.
	 * @return Returns a new WorkUnit.
	 * @throws EmptyQueueException 
	 * @throws InvalidUserException 
	 */
	@SuppressWarnings("unchecked")
	public Vector getWorkUnit(String userID, String eMailAddress) throws EmptyQueueException, InvalidUserException {
		beginTransaction();
		try {
			SysLogger.println("User " + userID + " requesting new work unit.");
			Volunteer user = findUser(userID, eMailAddress);
			WorkUnit wu = PRCServer.instance().getTransitioner().getWorkUnit(user);
			Vector result = new Vector();
			result.add(wu.getTaskID());
			result.add(wu.getData());
			SysLogger.println("Assigned work unit " + wu.getTaskID() + " to user " + userID + ".");
			return result;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Searches the database for the last work unit that was added (by creation date).
	 * @return Returns the last work unit that was added.
	 * @throws EmptyQueueException
	 */
	@SuppressWarnings("unchecked")
	public Vector getLastWorkUnit() throws EmptyQueueException {
		// Build the query.
		String queryString = "from WorkUnit as wu";					// Get WorkUnits
		queryString += " order by wu.creationDate DESC";				// in descending order by creation date
		queryString += ", wu.id DESC";								// and in descending order by Hibernate ID
	
		// Execute the query.
		Session hibernateSession = HibernateController.instance().createSession();
		Query resultQuery = hibernateSession.createQuery(queryString);
		resultQuery.setMaxResults(1);		// Only get the top result.
		List results = resultQuery.list();
		
		try {
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			
		}
		
		if (results.size() == 0) {
			throw new EmptyQueueException("There are no work units.");
		}
		else {
			WorkUnit wu = (WorkUnit) results.get(0);
			Vector wuVector = new Vector();
			wuVector.add(wu.getTaskID());
			wuVector.add(wu.getData());
			wuVector.add(wu.getCreationDate());
			wuVector.add(wu.getPoints());
			wuVector.add(wu.getPriority());
			return wuVector;
		}
	}
	
	/**
	 * Gets work unit data associated with the given result.
	 * This can be used by the result validator if it is necessary to examine
	 * the work unit from which the result was computed.
	 * @param resultID The result for which to find the associated work unit data.
	 * @return Returns the work unit data associated with the given result.
	 * @throws InvalidTaskException
	 */
	public Vector getAssociatedWorkUnit(String resultID) throws InvalidTaskException {
		Result r = findResult(resultID);
		WorkUnit wu = r.getWorkUnit();
		Vector<byte[]> wuVector = new Vector<byte[]>(1);
		wuVector.add(wu.getData());
		return wuVector;
	}
	
	/**
	 * Adds an ingress result to the transitioner.
	 * @param userID The user who computed the result.
	 * @param eMailAddress The user's e-mail address.
	 * @param resultData The data contained in the result.
	 * @return True if the result was added to the database/transitioner, false if it was not added due to null resultData.
	 * @throws InvalidUserException
	 */
	public boolean returnResult(String userID, String eMailAddress, String workUnitID, byte[] resultData) throws InvalidUserException, InvalidTaskException, UnassignedUserException {
		beginTransaction();
		try {
			SysLogger.println("User " + userID + " returning result for work unit " + workUnitID + ".");
			Volunteer user = findUser(userID, eMailAddress);
			WorkUnit wu = findWorkUnit(workUnitID);
			
			if (wu.hasUser(user) == false) {
				SysLogger.println("ERROR: Work Unit" + wu.getTaskID() + " has not been assigned to user " + user.getUserID() + ".");
				throw new UnassignedUserException("Work Unit" + wu.getTaskID() + " has not been assigned to user " + user.getUserID() + ".");
			}
			
			// Make sure there is data to set, and add the result to the transitioner.
			if (resultData != null) {
				Result r = TaskFactory.createIngressResult();
				SysLogger.println("Adding new result " + r.getTaskID() + " from user " + userID + " to the database and transitioner.");
				r.setVolunteer(user);
				r.setWorkUnit(wu);
				r.setData(resultData);
				r.setPriority(wu.getPriority());		// Each result's priority should be the same as its work unit.
				r.save();	// Add to database.
				PRCServer.instance().getTransitioner().addIngressResult(r);
				return true;
			}
			
			else {
				SysLogger.println("ERROR: User " + userID + " returned result with null data.");
				return false;
			}
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Finds the volunteer in the database with the given user ID.
	 * @param userID The user ID to search for.
	 * @param eMailAddress The user's e-mail address.
	 * @return The first volunteer in the database with the given user ID. 
	 * @throws InvalidUserException
	 */
	private Volunteer findUser(String userID, String eMailAddress) throws InvalidUserException {
		Volunteer user = new Volunteer(userID);
		user.setEMailAddress(eMailAddress);
		Volunteer[] matches = user.find();
		
		if (matches.length == 0)
			throw new InvalidUserException("User " + userID + " is not a known user.");
		else if (matches[0].getState() == VolunteerState.Banned)
			throw new InvalidUserException("User " + userID + " has been banned.");
		
		return matches[0];
	}
	
	/**
	 * Searches the database for a work unit with the given ID.
	 * @param workUnitID The ID of the work unit to find.
	 * @return Returns the work unit with the given ID.
	 * @throws InvalidTaskException
	 */
	private WorkUnit findWorkUnit(String workUnitID) throws InvalidTaskException {
		WorkUnit wu = TaskFactory.createWorkUnit(workUnitID);
		WorkUnit[] matches = wu.find();
		
		// If the work unit was not found, throw an exception.
		if (matches.length == 0)
			throw new InvalidTaskException("Work unit " + workUnitID + " is not a known work unit.");
		
		return matches[0];
	}
	
	/**
	 * Searches the database for a result with the given ID.
	 * @param resultID The ID of the result to find.
	 * @return Returns the result with the given ID.
	 * @throws InvalidTaskException
	 */
	private Result findResult(String resultID) throws InvalidTaskException {
		Result r = TaskFactory.createResult(resultID);
		Result[] matches = r.find();
		
		// If the result was not found, throw an exception.
		if (matches.length == 0)
			throw new InvalidTaskException("Result " + resultID + " is not a known result.");
		
		return matches[0];
	}
	
	/**
	 * Adds a new user to the system.
	 * @param userID The user ID for the new user.
	 * @param eMailAddress The new user's e-mail address.
	 * @return Returns true if the user was successfully added, false if the user already exists in the database.
	 */
	public boolean addUser(String userID, String eMailAddress) {
		beginTransaction();
		try {
			Volunteer v = new Volunteer(userID, eMailAddress, new Date());
			
			// If the user does not exist, return true.  Otherwise, return false.
			if (v.find().length == 0) {
				SysLogger.println("Adding user " + userID + " with e-mail address " + eMailAddress + ".");
				v.save();
				return true;
			}
			else {
				return false;
			}
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Determines whether a given user is a member of the project.
	 * @param userID The user ID to search for.
	 * @param eMailAddress The new user's e-mail address.
	 * @return Returns true if the user is a member of the project, false otherwise.
	 */
	public boolean verifyUser(String userID, String eMailAddress) {
		SysLogger.println("Verifying user " + userID + " with e-mail address " + eMailAddress + ".");
		Volunteer v = new Volunteer(userID, eMailAddress, new Date());
		
		// If the user exists, return true.  Otherwise, return false.
		if (v.find().length > 0)
			return true;
		else
			return false;
	}
	
	/**
	 * Adds an ingress work unit to the transitioner.
	 * @param taskID The ID of the new work unit.
	 * @param data The data contained in the work unit.
	 * @param priority The priority of the work unit.
	 * @param points The point value of the work unit.
	 * @param expirationTime The number of seconds until the work unit should expire.
	 * @return The number of users who are still waiting for a work unit.
	 * @throws DuplicateTaskIDException
	 */
	public int addWorkUnit(String taskID, byte[] data, int priority, int points, double expirationTime) throws DuplicateTaskIDException {
		beginTransaction();
		
		// Determine whether the requested task ID is unique.
		if (TaskFactory.createWorkUnit(taskID).find().length != 0) {
			SysLogger.println("WARNING: Attempt to add work unit with duplicate task ID: " + taskID);
			throw new DuplicateTaskIDException("The requested task ID is already in use by another task.");
		}
		
		try {
			WorkUnit wu = TaskFactory.createIngressWorkUnit(taskID);
			wu.setData(data);
			wu.setPriority(priority);
			wu.setPoints(points);
			wu.setExpirationTime((long)expirationTime);
			SysLogger.println("Adding work unit with task ID: " + wu.getTaskID());
			wu.save();
			
			PRCServer.instance().getTransitioner().addIngressWorkUnit(wu);
			int numUsersWaiting = Volunteer.getNumUsersWaiting();
			
			// Return the number of users that are blocked waiting for a work unit.
			return numUsersWaiting;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Adds an ingress work unit to the transitioner.
	 * @param taskID The ID of the new work unit.
	 * @param data The data contained in the work unit.
	 * @param priority The priority of the work unit.
	 * @param points The point value of the work unit.
	 * @return The number of users who are still waiting for a work unit.
	 * @throws DuplicateTaskIDException
	 */
	public int addWorkUnit(String taskID, byte[] data, int priority, int points) throws DuplicateTaskIDException {
		beginTransaction();
		
		// Determine whether the requested task ID is unique.
		if (TaskFactory.createWorkUnit(taskID).find().length != 0) {
			SysLogger.println("WARNING: Attempt to add work unit with duplicate task ID: " + taskID);
			throw new DuplicateTaskIDException("The requested task ID is already in use by another task.");
		}
		
		try {
			WorkUnit wu = TaskFactory.createIngressWorkUnit(taskID);
			wu.setData(data);
			wu.setPriority(priority);
			wu.setPoints(points);
			SysLogger.println("Adding work unit with task ID: " + wu.getTaskID());
			wu.save();
			
			PRCServer.instance().getTransitioner().addIngressWorkUnit(wu);
			int numUsersWaiting = Volunteer.getNumUsersWaiting();
			
			// Return the number of users that are blocked waiting for a work unit.
			return numUsersWaiting;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Adds an ingress work unit to the transitioner.
	 * @param data The data contained in the work unit.
	 * @param priority The priority of the work unit.
	 * @param points The point value of the work unit.
	 * @param expirationTime The number of seconds until the work unit should expire.
	 * @return The number of users who are still waiting for a work unit.
	 */
	public int addWorkUnit(byte[] data, int priority, int points, double expirationTime) {
		beginTransaction();
		try {
			WorkUnit wu = TaskFactory.createIngressWorkUnit();
			wu.setData(data);
			wu.setPriority(priority);
			wu.setPoints(points);
			wu.setExpirationTime((long)expirationTime);
			SysLogger.println("Adding work unit with task ID: " + wu.getTaskID());
			wu.save();
			
			PRCServer.instance().getTransitioner().addIngressWorkUnit(wu);
			int numUsersWaiting = Volunteer.getNumUsersWaiting();
			
			// Return the number of users that are blocked waiting for a work unit.
			return numUsersWaiting;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Adds an ingress work unit to the transitioner.
	 * @param data The data contained in the work unit.
	 * @param priority The priority of the work unit.
	 * @param points The point value of the work unit.
	 * @return The number of users who are still waiting for a work unit.
	 */
	public int addWorkUnit(byte[] data, int priority, int points) {
		beginTransaction();
		try {
			WorkUnit wu = TaskFactory.createIngressWorkUnit();
			wu.setData(data);
			wu.setPriority(priority);
			wu.setPoints(points);
			SysLogger.println("Adding work unit with task ID: " + wu.getTaskID());
			wu.save();
			
			PRCServer.instance().getTransitioner().addIngressWorkUnit(wu);
			int numUsersWaiting = Volunteer.getNumUsersWaiting();
			
			// Return the number of users that are blocked waiting for a work unit.
			return numUsersWaiting;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Adds an ingress work unit to the transitioner using all default values.
	 * @param data The data contained in the work unit.
	 * @return The number of users who are still waiting for a work unit.
	 */
	public int addWorkUnit(byte[] data) {
		beginTransaction();
		try {
			WorkUnit wu = TaskFactory.createIngressWorkUnit();
			wu.setData(data);
			SysLogger.println("Adding work unit with task ID: " + wu.getTaskID());
			wu.save();
			
			PRCServer.instance().getTransitioner().addIngressWorkUnit(wu);
			int numUsersWaiting = Volunteer.getNumUsersWaiting();
			
			// Return the number of users that are blocked waiting for a work unit.
			return numUsersWaiting;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * @return Returns the number of ingress work units in the system.
	 */
	public int getNumWorkUnits() {
		// Build the query.
		String queryString = "select count(wu) from WorkUnit as wu";	// Get number of work units
		queryString += " where wu.stateStr='Ingress'";					// that are ingress work units.
		
		// Execute the query.
		Session hibernateSession = HibernateController.instance().createSession();		
		Query resultQuery = hibernateSession.createQuery(queryString);
		List results = resultQuery.list();
		
		try {
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			
		}
		
		//return WorkUnit.findAllIngress().length;
		return ((Integer)results.get(0)).intValue();
	}
	
	/**
	 * Gets the next result to be validated, or a set of results for which 
	 * a canonical result must be selected.
	 * @return Vector containing one or more results.
	 * @throws EmptyQueueException
	 */
	@SuppressWarnings("unchecked")
	public Vector getResultForValidation() throws EmptyQueueException {
		beginTransaction();
		try {
			Vector result = new Vector();
			boolean isCanonical;
			WorkUnit canonicalWU = null;
			// First look for work units for which a canonical result must be selected. 
			try {
				canonicalWU = PRCServer.instance().getTransitioner().getCanonicalWorkUnit();
				isCanonical = true;
			} catch (EmptyQueueException e) {
				isCanonical = false;
			}
			
			if (isCanonical) {
				result.add(ResultType.Canonical.toInteger());
				Result searchResult = TaskFactory.createResult(null);	// Create result with null task ID so that 
				searchResult.setWorkUnit(canonicalWU);					// the find() call below does not search for that ID.
				Result[] results = searchResult.find();
				result.add(new Integer(results.length));
				for (Result nextResult : results) {
					result.add(nextResult.getTaskID());
					result.add(nextResult.getData());
				}
			}
			
			// Otherwise, return either a single normal result or a spot check result.
			// Throws EmptyQueueException if no ingress results were found.
			else {
				Result r = PRCServer.instance().getTransitioner().getResult();
				
				// Check if it's a spot check result.
				if (r.getWorkUnit().getState() == TaskState.SpotCheck) {
					// Build query to find the accepted result.
					String queryString = "from Result as result";											// Get Results
					queryString += " where result.workUnit.taskID='" + r.getWorkUnit().getTaskID() + "'";	// with workUnit = wu
					queryString += " and result.stateStr='" + TaskState.SpotCheck + "'";					// and task state = spot check
					
					// Execute the query.
					Session hibernateSession = HibernateController.instance().createSession();		
					Query resultQuery = hibernateSession.createQuery(queryString);
					List results = resultQuery.list();
					
					try {
						// Clean up the session.
						hibernateSession.close();
					} catch (HibernateException e) {
						
					}
					
					// Get the accepted result.
					Result acceptedResult = (Result)results.get(0);
					
					result.add(ResultType.SpotCheck.toInteger());
					result.add(r.getTaskID());
					result.add(r.getData());
					result.add(acceptedResult.getData());
				}
				
				// It is a single, normal result.
				else {
					result.add(ResultType.Single.toInteger());
					result.add(r.getTaskID());
					result.add(r.getData());	
				}
			}
			return result;
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Informs the transitioner that the specified ingress result is valid.
	 * @param resultID The ID of the ingress result to mark as valid.
	 * @return Returns true if the the result was successfully marked valid, false otherwise.
	 */
	public boolean markResultValid(String resultID) {
		beginTransaction();
		try {
			Result search = TaskFactory.createResult(resultID);
			Result[] matches = search.find();
			
			if (matches.length > 0) {
				SysLogger.println("Validator is marking valid result " + resultID + ".");
				PRCServer.instance().getTransitioner().markResultValid(matches[0]);
				return true;
			}
			
			else {
				SysLogger.println("Validator tried to mark valid unknown result " + resultID + ".");
				return false;
			}
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Informs the transitioner that the specified ingress result is NOT valid.
	 * @param resultID The ID of the ingress result to mark as invalid.
	 * @return Returns true if the the result was successfully marked invalid, false otherwise.
	 */
	public boolean markResultInvalid(String resultID) {
		beginTransaction();
		try {
			Result search = TaskFactory.createResult(resultID);
			Result[] matches = search.find();
			
			if (matches.length > 0) {
				SysLogger.println("Validator is marking invalid result " + resultID + ".");
				PRCServer.instance().getTransitioner().markResultInvalid(matches[0]);
				return true;
			}
			
			else {
				SysLogger.println("Validator tried to mark invalid unknown result " + resultID + ".");
				return false;
			}
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Designates a result as the canonical result for a work unit.
	 * @param resultID The task ID of the canonical result.
	 * @return Returns true if the canonical result was successfully selected, false if the result did not exist.
	 */
	public boolean selectCanonicalResult(String resultID) {
		beginTransaction();
		try {
			Result r = TaskFactory.createResult(resultID);
			Result[] matches = r.find();
			if (matches.length > 0) {
				SysLogger.println("Validator is selecting canonical result " + resultID + ".");
				PRCServer.instance().getTransitioner().selectCanonicalResult(matches[0]);
				return true;
			}
			else {
				SysLogger.println("Validator tried to select unknown canonical result " + resultID + ".");
				return false;
			}
		}
		finally {
			// Always end the transaction.
			endTransaction();
		}
	}
	
	/**
	 * Returns the valid result queue.
	 * Builds a vector that has twice as many elements as the number of valid 
	 * results.  The first element for each result is the task ID of that result's 
	 * associated work unit.  The second element is the result data.
	 * @param password The project password.
	 * @return Returns a Vector containing all valid results in the format described above. 
	 */
	@SuppressWarnings("unchecked")
	public Vector getValidResultQueue(String password) throws InvalidPasswordException {
		if (!password.equals(Project.instance().getPassword())) {
			SysLogger.println("WARNING: Attempt to retrieve all valid results failed due to the use of an incorrect password.");
			throw new InvalidPasswordException();
		}
		
		Vector results = new Vector();
		TransitionerQueue<Result> queue = PRCServer.instance().getTransitioner().getValidResultQueue();
		for (Result r : queue) {
			results.add(r.getWorkUnit().getTaskID());
			results.add(r.getData());
		}

		return results;
	}
	
	/**
	 * Shuts down the server if given the correct password.
	 * This RPC will cause the server to wait for all transitioner transactions to complete before shutting down.
	 * @param password The project password.
	 * @return Returns true.
	 * @throws InvalidPasswordException
	 */
	public boolean shutdown(String password) throws InvalidPasswordException {
		if (!password.equals(Project.instance().getPassword())) {
			SysLogger.println("WARNING: Attempt to shut down the project server failed due to the use of an incorrect password.");
			throw new InvalidPasswordException();
		}
		
		// Wait for all transactions to complete.
		PRCServer.instance().setWaitForTransactions(true);
		PRCServer.instance().initiateShutdown();
		return true;
	}
	
	/**
	 * Shuts down the server without waiting for all transitioner transactions to complete.
	 * This RPC should be used only if one or more RPCs are "stuck," and the server is waiting
	 * for an indefinite time for them to complete. 
	 * @param password The project password.
	 * @return Returns true.
	 * @throws InvalidPasswordException
	 */
	public boolean kill(String password) throws InvalidPasswordException {
		if (!password.equals(Project.instance().getPassword())) {
			SysLogger.println("WARNING: Attempt to kill the project server failed due to the use of an incorrect password.");
			throw new InvalidPasswordException();
		}
		
		// Don't wait for all transactions to complete.
		PRCServer.instance().setWaitForTransactions(false);
		PRCServer.instance().initiateShutdown();
		return true;		
	}
}
