package edu.wpi.prc.server.spotcheck;

import java.io.File;
import java.io.IOException;
import java.util.Vector;
import java.util.concurrent.Callable;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.persistence.SpotCheckLogger;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.WorkUnit;

/**
 * Invokes the project's science application to compute the result of a work unit.
 * @author James Baldassari
 */
public class SpotCheckGenerator implements Callable<Result> {
	private static final int SCIAPP_RETRY_DELAY = 10000;		// Wait 10 seconds before retrying an RPC call to the science app.
	private static final int STARTUP_DELAY = 10000;			// Wait 10 seconds for science application to start.
	private WorkUnit wu;
	private XmlRpcClient xmlrpcClient = null;
	
	/**
	 * Default constructor.
	 * @param wu The work unit for which to compute a result.
	 */
	public SpotCheckGenerator(WorkUnit wu, XmlRpcClient xmlrpcClient) {
		this.wu = wu;
		this.xmlrpcClient = xmlrpcClient;
	}
	
	/**
	 * Sends the work unit to the science application, receives the result for that work unit, 
	 * and makes it available to the SpotCheckThread through the Callable interface.
	 * @return The result computed for wu.
	 */
	public Result call() {
		SpotCheckLogger.println("SpotCheckGenerator.call: invoking science app " + Project.instance().getScienceAppPath());
		return invokeScienceApp();
	}
	
	/**
	 * Runs the science application, then passes it a work unit to compute 
	 * and waits for the science application to return a result.
	 * @return Returns the result computed by the science application. 
	 */
	private Result invokeScienceApp() {
		SpotCheckLogger.println("PRCClient.invokeScienceApp: Sending work unit " + wu.getTaskID() + " to science application for processing.");
		
		Vector<byte[]> args = new Vector<byte[]>();
		args.add(wu.getData());
		Vector resultVector = null;
		boolean receivedResult = false;
		
		// Execute the XML-RPC call.
		while (receivedResult == false) {
			try {
				Object result = xmlrpcClient.execute("sciapp.computeSpotCheckResult", args);
				try {
					resultVector = (Vector)result;
					receivedResult = true;
				} catch (Exception e) {
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Waiting for " + (SCIAPP_RETRY_DELAY / 1000) + " seconds before retry attempt.");
					try {
						Thread.sleep(SCIAPP_RETRY_DELAY);
					} catch (InterruptedException e1) {
						SpotCheckLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
					}
				}
			} catch (XmlRpcException e) {
				SpotCheckLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
				try {
					Thread.sleep(SCIAPP_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			} catch (IOException e) {
				// If control reaches this block, the science application is not running.
				SpotCheckLogger.println("PRCClient.invokeScienceApp: Science Application is not running.");
				SpotCheckLogger.println("PRCClient.invokeScienceApp: Starting Science Application.");
				try {
					Runtime.getRuntime().exec(Project.instance().getScienceAppPath(), null, new File(Project.instance().getClientDirectoryPath()));
				} catch (IOException e1) {
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Error starting Science Application: " + e1);
					e1.printStackTrace();
				}
				
				// Wait for the science application to start up before sending any work units.
				try {
					Thread.sleep(STARTUP_DELAY);
				} catch (InterruptedException e1) {
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			} catch (Exception e) {
				SpotCheckLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
				try {
					Thread.sleep(SCIAPP_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SpotCheckLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			}
		}
		
		// Build the result.
		Result r = TaskFactory.createResult();
		r.setData((byte[])resultVector.get(0));
		
		SpotCheckLogger.println("PRCClient.invokeScienceApp: Received result from science application for work unit " + wu.getTaskID() + ".");
		
		return r;
	}

}
