/**
 * 
 */
package edu.wpi.prc.server.spotcheck;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.persistence.InvalidLogPathException;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.persistence.SpotCheckLogger;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;

/**
 * Creates a specified number of spot checks by taking work units from the 
 * ingress work queue, computing their results, then adding the work units 
 * and results to the spot check queues.
 * @author James Baldassari
 */
public class SpotCheckThread extends Thread {
	private TransitionerQueue<WorkUnit> ingressWork;
	private TransitionerQueue<WorkUnit> spotCheckWork;
	private TransitionerQueue<Result> spotCheckResults;
	private int minNumSpotChecks;
	private String scienceAppURI = null;
	private XmlRpcClient xmlrpcClient = null;

	/**
	 * Initializes the spot-check thread.
	 * @param ingressWork The transitioner's ingress work queue.
	 * @param spotCheckWork The transitioner's spot check work unit queue.
	 * @param spotCheckResult The transitioner's spot check result queue.
	 * @param minNumSpotChecks The number of required spot checks.
	 */
	public SpotCheckThread(
			TransitionerQueue<WorkUnit> ingressWork, 
			TransitionerQueue<WorkUnit> spotCheckWork, 
			TransitionerQueue<Result> spotCheckResult,
			int minNumSpotChecks) {
		this.ingressWork = ingressWork;
		this.spotCheckWork = spotCheckWork;
		this.spotCheckResults = spotCheckResult;
		this.minNumSpotChecks = minNumSpotChecks;
		
		// Set the name of this thread.
		this.setName("Spot-Check WorkUnitGenerator Thread");
		
		// Set minimum priority
		this.setPriority(Thread.MIN_PRIORITY);
		
		// Initialize the spot check log file.
		try {
			SpotCheckLogger.initialize();
		} catch (InvalidLogPathException e) {
			e.printStackTrace();
		}
		
		// Redirect stderr to spot-check log file.
		try {
			System.setErr(new PrintStream(new FileOutputStream(SpotCheckLogger.DEFAULT_LOG_PATH)));
		} catch (FileNotFoundException e) {

		}
		
		this.scienceAppURI = "http://localhost:" + Project.instance().getScienceAppPort() + "/RPC2";
		initXmlRpcClient();
	}
	
	/**
	 * Initializes the XML-RPC clients used for communicating with the project server 
	 * and science application.
	 */
	private void initXmlRpcClient() {
		try {
			xmlrpcClient = new XmlRpcClient(scienceAppURI);
		} catch (MalformedURLException e) {
			SpotCheckLogger.println("SpotCheckGenerator: Invalid URL used to initialize the XML-RPC client: " + scienceAppURI);
		}
	}
	
	/**
	 * Creates as many spot checks as needed by invoking the science application via XML-RPC.
	 */
	public void run() {
		SysLogger.println("SpotCheckThread running.");
		SpotCheckLogger.println("SpotCheckThread running.");
		while (spotCheckResults.size() < minNumSpotChecks) {
			WorkUnit wu = null;
			Result r = null;
			
			// Keep trying to get a WorkUnit.
			while (wu == null) {
				SpotCheckLogger.println("SpotCheckThread: Attempting to resume spot check.");
				
				/*
				 * If there are fewer spot check results than spot check work units, 
				 * it means that the SpotCheckThread started computing a spot check 
				 * result, but did not finish.  If this happened, we'll just resume 
				 * the computation.  NOTE: There should never be more than one 
				 * unfinished spot check.
				 */
				if (spotCheckResults.size() < spotCheckWork.size()) {
					boolean[] workFinished = new boolean[spotCheckWork.size()];
					for (int ii = 0; ii < spotCheckWork.size(); ii++)
						workFinished[ii] = false;
					
					for (Result spotResult : spotCheckResults) {
						int workIndex = 0;
						for (WorkUnit spotWU : spotCheckWork) {
							if (spotResult.getWorkUnit().equals(spotWU)) {
								workFinished[workIndex] = true;
								break;
							}
							workIndex++;
						}
					}
					
					// Now just search the workFinished array to determine which work unit has not been completed.
					int workIndex = 0;
					for (WorkUnit spotWU : spotCheckWork) {
						if (workFinished[workIndex] == false) {
							wu = spotWU;
							break;
						}
					}
				}
				
				// If there were no uncompleted spot check work units, get an ingress work unit.
				else {
					SpotCheckLogger.println("SpotCheckThread: Attempting to acquire an ingress work unit.");
					try {
						// Try to steal a work unit from the ingress work queue.
						wu = ingressWork.removeFirst();
						
						// Move the work unit to spot check queue.
						wu.setState(TaskState.SpotCheck);
						wu.update();
						spotCheckWork.add(wu);
					} catch (NoSuchElementException e) {
						try {
							SpotCheckLogger.println("SpotCheckThread: Ingress work queue empty.  Sleeping for 10 seconds.");
							// Sleep for 10 seconds, then try to get a WorkUnit again.
							Thread.sleep(10000);
						} catch (InterruptedException e2) {
							// Silent failure.
						}	
					}
					SpotCheckLogger.println("SpotCheckThread: Got an ingress work unit!");
				}
			}
			
			// Compute the result for this work unit.
			SpotCheckGenerator generator = new SpotCheckGenerator(wu, xmlrpcClient);
			FutureTask<Result> task = new FutureTask<Result>(generator);
			task.run();
			
			// Block until the Result has been computed.
			try {
				r = task.get();
			} catch (InterruptedException e) {
				r = null;
			} catch (ExecutionException e2) {
				r = null;
			}
			
			// If the Result was generated correctly, add it as a spot check.
			if (r != null) {
				// Add the Result to the spot check queue.
				r.setWorkUnit(wu);
				r.setState(TaskState.SpotCheck);
				r.save();
				spotCheckResults.add(r);
			}
		}
		
		// Now done generating spot check results, so shut down the science application.
		shutdownScienceApplication();
	}
	
	/**
	 * Shuts down the science application via XML-RPC.
	 */
	private void shutdownScienceApplication() {
		try {
			xmlrpcClient.execute("sciapp.shutdown", new Vector());
		} catch (XmlRpcException e) {
			SpotCheckLogger.println("SpotCheckThread.shutdown: Error shutting down science application: " + e);
		} catch (IOException e) {
			SpotCheckLogger.println("SpotCheckThread.shutdown: Error shutting down science application: " + e);
		}
	}
	
	/**
	 * Shuts down the SpotCheckThread.
	 * First, it attempts to shut down the science application via XML-RPC, 
	 * then it interrupts itself to stop the thread.
	 */
	public void shutdown() {
		// Shut down the science application.
		SysLogger.println("Spot Check Thread: Shutting down Science Application.");
		SpotCheckLogger.println("Spot Check Thread: Shutting down Science Application.");
		System.out.println("Spot Check Thread: Shutting down Science Application.");
		shutdownScienceApplication();
		SysLogger.println("Spot Check Thread: Science Application Halted.");
		SpotCheckLogger.println("Spot Check Thread: Science Application Halted.");
		System.out.println("Spot Check Thread: Science Application Halted.");
		// Interrupt the SpotCheckThread.
		this.interrupt();
	}
}
