package edu.wpi.prc.server.test;

import java.util.Date;

import org.hibernate.HibernateException;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.project.SpotCheckFailureAction;
import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.server.transitioner.LocalTransitioner;
import edu.wpi.prc.server.transitioner.Transitioner;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;
import junit.framework.TestCase;

/**
 * Tests the LocalTransitioner.
 * @author James Baldassari
 */
public class LocalTransitionerTest extends TestCase {
	private LocalTransitioner transitioner = null;
	private Volunteer user = new Volunteer();
	private Project p = Project.instance();

	public LocalTransitionerTest(String name) {
		super(name);
		
		HibernateController.instance().initializeForUnitTests();
		HibernateController.instance().exportSchema();
		
		transitioner = new LocalTransitioner();
	}

	protected void setUp() throws Exception {
		super.setUp();
		
		HibernateController.instance().exportSchema();
		
		// Configure the project.
		p.setServerURI("localhost:2080");
		p.setTransitionerURI(null);
		p.setSpotCheckFailureAction(SpotCheckFailureAction.Log);
		p.setMinNumSpotChecks(0);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {

		}
	}
	
	/**
	 * Tests restoring the local transitioner from the database.
	 */
	public void testRestoreFromDatabase() {
		// Add a WorkUnit to the database.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		WorkUnit ingressWU = TaskFactory.createIngressWorkUnit();
		ingressWU.assignUser(v1);
		ingressWU.save();
		
		// shut down and re-initialize
		HibernateController.instance().shutdown();
		HibernateController.instance().initializeForUnitTests();
		
		// Will restore tasks from the database
		Transitioner transitioner = new LocalTransitioner();
		
		// Get all ingress work units
		TransitionerQueue<WorkUnit> ingressWork = transitioner.getIngressWorkQueue();
		
		assertNotNull(ingressWork);
		assertTrue(ingressWork.size() == 1);
		assertTrue(ingressWork.toArray(new WorkUnit[0])[0].getTaskID().equals(ingressWU.getTaskID()));
	}
	
	/**
	 * Tests the getIngressWorkQueue method.
	 */
	public void testGetIngressWorkQueue() {
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.save();
		transitioner.addIngressWorkUnit(wu1);
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.save();
		transitioner.addIngressWorkUnit(wu2);
		WorkUnit wu3 = TaskFactory.createWorkUnit();
		wu3.save();
		transitioner.addPendingWorkUnit(wu3);
		WorkUnit wu4 = TaskFactory.createWorkUnit();
		wu4.save();
		transitioner.addEgressWorkUnit(wu4);
		
		TransitionerQueue<WorkUnit> queue = transitioner.getIngressWorkQueue();
		assertTrue(queue.size() == 2);
		for (WorkUnit wu : queue)
			assertTrue(wu.getState() == TaskState.Ingress);
		assertTrue(queue.toArray(new WorkUnit[0])[0].getTaskID().equals(wu1.getTaskID()));
		assertTrue(queue.toArray(new WorkUnit[0])[1].getTaskID().equals(wu2.getTaskID()));
	}
	
	/**
	 * Tests the getPendingWorkQueue method.
	 */
	public void testGetPendingWorkQueue() {
		WorkUnit wu1 = TaskFactory.createWorkUnit();
		wu1.save();
		transitioner.addPendingWorkUnit(wu1);
		WorkUnit wu2 = TaskFactory.createWorkUnit();
		wu2.save();
		transitioner.addPendingWorkUnit(wu2);
		WorkUnit wu3 = TaskFactory.createIngressWorkUnit();
		wu3.save();
		transitioner.addIngressWorkUnit(wu3);
		WorkUnit wu4 = TaskFactory.createWorkUnit();
		wu4.save();
		transitioner.addEgressWorkUnit(wu4);
		
		TransitionerQueue<WorkUnit> queue = transitioner.getPendingWorkQueue();
		assertTrue(queue.size() == 2);
		for (WorkUnit wu : queue)
			assertTrue(wu.getState() == TaskState.Pending);
		assertTrue(queue.toArray(new WorkUnit[0])[0].getTaskID().equals(wu1.getTaskID()));
		assertTrue(queue.toArray(new WorkUnit[0])[1].getTaskID().equals(wu2.getTaskID()));
	}
	
	/**
	 * Tests the getRetiredWorkQueue method.
	 */
	public void testGetRetiredWorkQueue() {
		WorkUnit wu1 = TaskFactory.createWorkUnit();
		wu1.save();
		transitioner.addEgressWorkUnit(wu1);
		WorkUnit wu2 = TaskFactory.createWorkUnit();
		wu2.save();
		transitioner.addEgressWorkUnit(wu2);
		WorkUnit wu3 = TaskFactory.createIngressWorkUnit();
		wu3.save();
		transitioner.addIngressWorkUnit(wu3);
		WorkUnit wu4 = TaskFactory.createWorkUnit();
		wu4.save();
		transitioner.addPendingWorkUnit(wu4);
		
		TransitionerQueue<WorkUnit> queue = transitioner.getRetiredWorkQueue();
		assertTrue(queue.size() == 2);
		for (WorkUnit wu : queue)
			assertTrue(wu.getState() == TaskState.Egress);
		assertTrue(queue.toArray(new WorkUnit[0])[0].getTaskID().equals(wu1.getTaskID()));
		assertTrue(queue.toArray(new WorkUnit[0])[1].getTaskID().equals(wu2.getTaskID()));
	}
	
	/**
	 * Tests the getIngressResultQueue method.
	 */
	public void testGetIngressResultQueue() {
		user.save();
		WorkUnit wu = TaskFactory.createWorkUnit();
		wu.setState(TaskState.Pending);
		wu.save();
		Result r1 = TaskFactory.createIngressResult();
		r1.setVolunteer(user);
		r1.setWorkUnit(wu);
		r1.save();
		transitioner.addIngressResult(r1);
		Result r2 = TaskFactory.createIngressResult();
		r2.setVolunteer(user);
		r2.setWorkUnit(wu);
		r2.save();
		transitioner.addIngressResult(r2);
		Result r3 = TaskFactory.createResult();
		r3.setVolunteer(user);
		r3.setWorkUnit(wu);
		r3.save();
		transitioner.addPendingResult(r3);
		Result r4 = TaskFactory.createResult();
		r4.setVolunteer(user);
		r4.setWorkUnit(wu);
		r4.save();
		transitioner.addEgressResult(r4);
		
		TransitionerQueue<Result> queue = transitioner.getIngressResultQueue();
		assertTrue(queue.size() == 2);
		for (Result r : queue)
			assertTrue(r.getState() == TaskState.Ingress);
		assertTrue(queue.toArray(new Result[0])[0].getTaskID().equals(r1.getTaskID()));
		assertTrue(queue.toArray(new Result[0])[1].getTaskID().equals(r2.getTaskID()));
	}
	
	/**
	 * Tests the getPendingResultQueue method.
	 */
	public void testGetPendingResultQueue() {
		user.save();
		WorkUnit wu = TaskFactory.createWorkUnit();
		wu.setState(TaskState.Pending);
		wu.save();
		Result r1 = TaskFactory.createResult();
		r1.setVolunteer(user);
		r1.setWorkUnit(wu);
		r1.save();
		transitioner.addPendingResult(r1);
		Result r2 = TaskFactory.createResult();
		r2.setVolunteer(user);
		r2.setWorkUnit(wu);
		r2.save();
		transitioner.addPendingResult(r2);
		Result r3 = TaskFactory.createIngressResult();
		r3.setVolunteer(user);
		r3.setWorkUnit(wu);
		r3.save();
		transitioner.addIngressResult(r3);
		Result r4 = TaskFactory.createResult();
		r4.setVolunteer(user);
		r4.setWorkUnit(wu);
		r4.save();
		transitioner.addEgressResult(r4);
		
		TransitionerQueue<Result> queue = transitioner.getPendingResultQueue();
		assertTrue(queue.size() == 2);
		for (Result r : queue)
			assertTrue(r.getState() == TaskState.Pending);
		assertTrue(queue.toArray(new Result[0])[0].getTaskID().equals(r1.getTaskID()));
		assertTrue(queue.toArray(new Result[0])[1].getTaskID().equals(r2.getTaskID()));
	}
	
	/**
	 * Tests the getValidResultQueue method.
	 */
	public void testGetValidResultQueue() {
		user.save();
		WorkUnit wu = TaskFactory.createWorkUnit();
		wu.setState(TaskState.Pending);
		wu.save();
		Result r1 = TaskFactory.createResult();
		r1.setVolunteer(user);
		r1.setWorkUnit(wu);
		r1.save();
		transitioner.addEgressResult(r1);
		Result r2 = TaskFactory.createResult();
		r2.setVolunteer(user);
		r2.setWorkUnit(wu);
		r2.save();
		transitioner.addEgressResult(r2);
		Result r3 = TaskFactory.createIngressResult();
		r3.setVolunteer(user);
		r3.setWorkUnit(wu);
		r3.save();
		transitioner.addIngressResult(r3);
		Result r4 = TaskFactory.createResult();
		r4.setVolunteer(user);
		r4.setWorkUnit(wu);
		r4.save();
		transitioner.addPendingResult(r4);
		
		TransitionerQueue<Result> queue = transitioner.getValidResultQueue();
		assertTrue(queue.size() == 2);
		for (Result r : queue)
			assertTrue(r.getState() == TaskState.Egress);
		assertTrue(queue.toArray(new Result[0])[0].getTaskID().equals(r1.getTaskID()));
		assertTrue(queue.toArray(new Result[0])[1].getTaskID().equals(r2.getTaskID()));
	}
	
	/**
	 * Tests the addIngressWorkUnit method.
	 */
	public void testAddIngressWorkUnit() {
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.save();
		transitioner.addIngressWorkUnit(wu1);
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.save();
		transitioner.addIngressWorkUnit(wu2);
		
		WorkUnit[] found = WorkUnit.findAll();
		assertTrue(found.length == 2);
		
		TransitionerQueue<WorkUnit> queue = transitioner.getIngressWorkQueue();
		assertTrue(queue.size() == 2);
		for (WorkUnit wu : queue)
			assertTrue(wu.getState() == TaskState.Ingress);
		assertTrue(queue.toArray(new WorkUnit[0])[0].getTaskID().equals(wu1.getTaskID()));
		assertTrue(queue.toArray(new WorkUnit[0])[1].getTaskID().equals(wu2.getTaskID()));
	}
	
	/**
	 * Tests the getAssociatedIngressResults(String hqlQuery) method.
	 */
	public void testGetAssociatedIngressResults() {
		// Introduce a new WorkUnit into the transitioner.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		WorkUnit ingressWU = TaskFactory.createIngressWorkUnit();
		ingressWU.assignUser(v1);
		ingressWU.save();
		transitioner.addIngressWorkUnit(ingressWU);
		
		// Get the WorkUnit from the database (with the hibernate ID set).
		WorkUnit wu = ingressWU.find()[0];
		
		// Introduce new Results for ingressWU into the transitioner.
		Volunteer v2 = new Volunteer();
		v2.setUserID("jdb");
		v2.setEMailAddress("jdb@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Result ingressResult1 = TaskFactory.createIngressResult();
		ingressResult1.setWorkUnit(wu);
		ingressResult1.setVolunteer(v2);
		ingressResult1.save();
		transitioner.addIngressResult(ingressResult1);
		Result ingressResult2 = TaskFactory.createIngressResult();
		ingressResult2.setWorkUnit(wu);
		ingressResult2.setVolunteer(v2);
		ingressResult2.save();
		transitioner.addIngressResult(ingressResult2);
		
		// Add some "decoy" results that should not be returned by getAssociatedIngressResults.
		WorkUnit decoyWorkUnit = TaskFactory.createIngressWorkUnit();
		decoyWorkUnit.save();
		transitioner.addIngressWorkUnit(decoyWorkUnit);
		Result decoyResult1 = TaskFactory.createIngressResult();
		decoyResult1.setWorkUnit(decoyWorkUnit);
		decoyResult1.setVolunteer(v2);
		decoyResult1.save();
		transitioner.addIngressResult(decoyResult1);
		Result decoyResult2 = TaskFactory.createIngressResult();
		decoyResult2.setWorkUnit(decoyWorkUnit);
		decoyResult2.setVolunteer(v2);
		decoyResult2.save();
		transitioner.addIngressResult(decoyResult2);
		Result decoyResult3 = TaskFactory.createIngressResult();
		decoyResult3.setWorkUnit(decoyWorkUnit);
		decoyResult3.setVolunteer(v2);
		decoyResult3.save();
		transitioner.addIngressResult(decoyResult3);
		Result decoyResult4 = TaskFactory.createResult();
		decoyResult4.setWorkUnit(decoyWorkUnit);
		decoyResult4.setState(TaskState.Pending);
		decoyResult4.setVolunteer(v2);
		decoyResult4.save();
		transitioner.addPendingResult(decoyResult4);
		
		
		// We should be able to find ingressResult[1,2] as associated results of ingressWU.
		Result[] associatedResults = transitioner.getAssociatedIngressResults(ingressWU);
		assertNotNull(associatedResults);
		assertTrue(associatedResults.length == 2);
		for (Result r : associatedResults)
			assertTrue(r.getState() == TaskState.Ingress);
		assertTrue(associatedResults[0].getTaskID().equals(ingressResult1.getTaskID()) || 
				associatedResults[0].getTaskID().equals(ingressResult2.getTaskID()));
		assertTrue(associatedResults[0].getWorkUnit().getTaskID().equals(ingressWU.getTaskID()));
		assertTrue(associatedResults[1].getWorkUnit().getTaskID().equals(ingressWU.getTaskID()));
	}
	
	/**
	 * Tests the addIngressResult method.
	 * This tests the normal case in which the result's work unit
	 * has not been retired and the result is not a spot check 
	 * result.
	 */
	public void testAddIngressResultNormal() {
		// Add a Volunteer to the database.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		
		// Get the Volunteer from the database.
		Volunteer storedUser = Volunteer.findAll()[0];
		assertTrue(storedUser.getScore() == 0);		// Score should be zero initially.
		
		// Add a WorkUnit to the database.		
		WorkUnit ingressWU = TaskFactory.createIngressWorkUnit();
		ingressWU.setPoints(2);			// This WU is worth 2 points.
		ingressWU.assignUser(storedUser);
		ingressWU.save();
		transitioner.addIngressWorkUnit(ingressWU);
		
		// Get the WorkUnit from the database.
		WorkUnit storedWU = WorkUnit.findAll()[0];
		assertTrue(storedWU.getTaskID().equals(ingressWU.getTaskID()));
		
		// Add a Result for the WorkUnit.
		Result ingressResult = TaskFactory.createIngressResult();
		ingressResult.setWorkUnit(storedWU);
		ingressResult.setVolunteer(storedUser);
		ingressResult.save();
		
		// Add the result to the transitioner.
		transitioner.addIngressResult(ingressResult);
		
		// The result should be returned in the ingress result queue.
		TransitionerQueue<Result> ingressResults = transitioner.getIngressResultQueue();
		assertTrue(ingressResults.size() == 1);
		for (Result r : ingressResults) {
			assertTrue(r.getTaskID().equals(ingressResult.getTaskID()));
			assertTrue(r.getState() == TaskState.Ingress);
			assertTrue(r.getVolunteer().getUserID().equals("jdb"));
			assertTrue(r.getVolunteer().getScore() == 2);
			//System.out.println("User's last result date = " + r.getUser().getLastResultDate());
		}
	}
	
	/**
	 * Tests the addIngressResult method.
	 * This tests the exceptional case when the result's work unit 
	 * has been retired and the result is not a spot check result.
	 */
	public void testAddIngressResultRetiredWorkUnit() {
		// Add a Volunteer to the database.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		
		// Get the Volunteer from the database.
		Volunteer storedUser = Volunteer.findAll()[0];
		assertTrue(storedUser.getScore() == 0);		// Score should be zero initially.
		
		// Add a WorkUnit to the database.		
		WorkUnit ingressWU = TaskFactory.createIngressWorkUnit();
		ingressWU.setPoints(2);			// This WU is worth 2 points.
		ingressWU.assignUser(storedUser);
		ingressWU.save();
		transitioner.addIngressWorkUnit(ingressWU);
		
		// Get the WorkUnit from the database and set its state to Egress.
		WorkUnit storedWU = WorkUnit.findAll()[0];
		assertTrue(storedWU.getTaskID().equals(ingressWU.getTaskID()));
		storedWU.setState(TaskState.Egress);
		storedWU.update();
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// Add a Result for the WorkUnit.
		Result ingressResult = TaskFactory.createIngressResult();
		ingressResult.setWorkUnit(storedWU);
		ingressResult.setVolunteer(storedUser);
		ingressResult.save();
		
		// Add the result to the transitioner.
		transitioner.addIngressResult(ingressResult);
		
		// The result should be deleted by the transitioner because the work unit has been retired.
		TransitionerQueue<Result> ingressResults = transitioner.getIngressResultQueue();
		assertTrue(ingressResults.size() == 0);
	
		// The user's score should still be increased.
		Volunteer v = Volunteer.findAll()[0];
		assertTrue(v.getUserID().equals("jdb"));
		assertTrue(v.getScore() == 2);
	}
	
	/**
	 * Tests the addIngressResult method.
	 * This tests the exceptional case when the result is a 
	 * spot check result.
	 */
	public void testAddIngressResultSpotCheck() {
		// Add a Volunteer to the database.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		
		// Get the Volunteer from the database.
		Volunteer storedUser = Volunteer.findAll()[0];
		assertTrue(storedUser.getScore() == 0);		// Score should be zero initially.
		
		// Add a WorkUnit to the database.		
		WorkUnit spotCheckWU = TaskFactory.createWorkUnit();
		spotCheckWU.setState(TaskState.SpotCheck);
		spotCheckWU.save();
		
		// Get the WorkUnit from the database and set its state to Egress.
		WorkUnit storedWU = WorkUnit.findAll()[0];
		assertTrue(storedWU.getTaskID().equals(spotCheckWU.getTaskID()));
		assertTrue(storedWU.getState() == TaskState.SpotCheck);
		
		// Add an accepted spot check result to the database.
		Result spotCheckResult = TaskFactory.createResult();
		spotCheckResult.setState(TaskState.SpotCheck);
		spotCheckResult.setWorkUnit(storedWU);
		spotCheckResult.setData(new byte[] { 4, 3, 2, 1, 0 });
		spotCheckResult.save();
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// Add a Result for the WorkUnit.
		Result ingressResult = TaskFactory.createIngressResult();
		ingressResult.setWorkUnit(storedWU);
		ingressResult.setVolunteer(storedUser);
		ingressResult.setData(new byte[] { 4, 3, 2, 1, 0 });
		ingressResult.save();
		
		// Add the result to the transitioner.
		transitioner.addIngressResult(ingressResult);
		transitioner.markResultValid(ingressResult);
		
		// The spot check result should be deleted by the transitioner.
		TransitionerQueue<Result> ingressResults = transitioner.getIngressResultQueue();
		assertTrue(ingressResults.size() == 0);
	
		// The user's score should still be increased.
		Volunteer v = Volunteer.findAll()[0];
		assertTrue(v.getUserID().equals("jdb"));
		assertTrue(v.getScore() == 1);
	}
	
	/**
	 * Tests the getAssociatedPendingResults(String hqlQuery) method.
	 */
	public void testGetAssociatedPendingResults() {
		// Introduce a new WorkUnit into the transitioner.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		WorkUnit ingressWU = TaskFactory.createIngressWorkUnit();
		ingressWU.assignUser(v1);
		ingressWU.save();
		transitioner.addIngressWorkUnit(ingressWU);
		
		// Get the WorkUnit from the database (with the hibernate ID set).
		WorkUnit wu = ingressWU.find()[0];
		
		// Introduce new Results for ingressWU into the transitioner.
		Volunteer v2 = new Volunteer();
		v2.setUserID("jdb");
		v2.setEMailAddress("jdb@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Result pendingResult1 = TaskFactory.createResult();
		pendingResult1.setWorkUnit(wu);
		pendingResult1.setState(TaskState.Pending);
		pendingResult1.setVolunteer(v2);
		pendingResult1.save();
		transitioner.addPendingResult(pendingResult1);
		Result pendingResult2 = TaskFactory.createResult();
		pendingResult2.setWorkUnit(wu);
		pendingResult2.setState(TaskState.Pending);
		pendingResult2.setVolunteer(v2);
		pendingResult2.save();
		transitioner.addPendingResult(pendingResult2);
		
		// Add some "decoy" results that should not be returned by getAssociatedIngressResults.
		WorkUnit decoyWorkUnit = TaskFactory.createWorkUnit();
		decoyWorkUnit.setState(TaskState.Pending);
		decoyWorkUnit.save();
		transitioner.addPendingWorkUnit(decoyWorkUnit);
		Result decoyResult1 = TaskFactory.createResult();
		decoyResult1.setWorkUnit(decoyWorkUnit);
		decoyResult1.setState(TaskState.Pending);
		decoyResult1.setVolunteer(v2);
		decoyResult1.save();
		transitioner.addPendingResult(decoyResult1);
		Result decoyResult2 = TaskFactory.createResult();
		decoyResult2.setWorkUnit(decoyWorkUnit);
		decoyResult2.setState(TaskState.Pending);
		decoyResult2.setVolunteer(v2);
		decoyResult2.save();
		transitioner.addPendingResult(decoyResult2);
		Result decoyResult3 = TaskFactory.createResult();
		decoyResult3.setWorkUnit(decoyWorkUnit);
		decoyResult3.setState(TaskState.Ingress);
		decoyResult3.setVolunteer(v2);
		decoyResult3.save();
		transitioner.addPendingResult(decoyResult3);
		
		
		// We should be able to find pendingResult[1,2] as associated results of ingressWU.
		Result[] associatedResults = transitioner.getAssociatedPendingResults(ingressWU);
		assertNotNull(associatedResults);
		assertTrue(associatedResults.length == 2);
		for (Result r : associatedResults)
			assertTrue(r.getState() == TaskState.Pending);
		assertTrue(associatedResults[0].getTaskID().equals(pendingResult1.getTaskID()) || 
				associatedResults[0].getTaskID().equals(pendingResult2.getTaskID()));
		assertTrue(associatedResults[0].getWorkUnit().getTaskID().equals(ingressWU.getTaskID()));
		assertTrue(associatedResults[1].getWorkUnit().getTaskID().equals(ingressWU.getTaskID()));
	}
	
	/**
	 * Tests the getWorkUnit method.
	 */
	public void testGetWorkUnit() {
		// v1 is working on wu1.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		Volunteer firstUser = v1.find()[0];
		
		// wu1 is a pending work unit.
		WorkUnit wu1 = TaskFactory.createWorkUnit();
		wu1.setData(new byte[] { 1, 2, 3, 4 });
		wu1.setState(TaskState.Pending);
		wu1.assignUser(firstUser);
		wu1.save();
		
		// wu2 is an ingress work unit.
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.setData(new byte[] { 1, 2, 3, 4 });
		wu2.save();
		transitioner.addIngressWorkUnit(wu2);
		
		// v2 will request a new work unit.
		Volunteer v2 = new Volunteer();
		v2.setUserID("toth");
		v2.setEMailAddress("toth@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Volunteer secondUser = v2.find()[0];
		
		// Make sure that the project is expecting 2 results for every work unit.
		p.setMinNumResults(2);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {
			
		}
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// First work unit returned to the second user should be wu1.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(secondUser);
			assertTrue(firstWU.getTaskID().equals(wu1.getTaskID()));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// Second work unit returned to the second user should be wu2.
		try {
			WorkUnit secondWU = transitioner.getWorkUnit(secondUser);
			assertTrue(secondWU.getTaskID().equals(wu2.getTaskID()));
			// Also, wu2's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu2.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find()[0].getState() == TaskState.Pending);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// Second work unit returned to the first user should be wu2.
		try {
			WorkUnit secondWU = transitioner.getWorkUnit(firstUser);
			assertTrue(secondWU.getTaskID().equals(wu2.getTaskID()));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// There shouldn't be any available work units left, so we should get an exception.
		try {
			WorkUnit thirdWU = transitioner.getWorkUnit(firstUser);
			System.out.println(thirdWU);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
		
		// There shouldn't be any avilable work units left, so we should get an exception.
		try {
			WorkUnit thirdWU = transitioner.getWorkUnit(secondUser);
			System.out.println(thirdWU);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
	}
	
	/**
	 * Tests the getWorkUnit method using different work unit priorities.
	 */
	public void testGetWorkUnitPriority() {
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		Volunteer firstUser = v1.find()[0];
		
		// wu1 is an ingress work unit with default priority (0).
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.setData(new byte[] { 1, 2, 3, 4 });
		wu1.save();
		transitioner.addIngressWorkUnit(wu1);
		
		// wu2 is an ingress work unit with priority 1.
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.setData(new byte[] { 1, 2, 3, 4 });
		wu2.setPriority(1);
		wu2.save();
		transitioner.addIngressWorkUnit(wu2);
		
		Volunteer v2 = new Volunteer();
		v2.setUserID("toth");
		v2.setEMailAddress("toth@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Volunteer secondUser = v2.find()[0];
		
		// Make sure that the project is expecting 2 results for every work unit.
		p.setMinNumResults(2);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {
			
		}
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// First work unit returned to the first user should be wu2 because it has a higher priority than wu1.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(firstUser);
			assertTrue(firstWU.getTaskID().equals(wu2.getTaskID()));
			// Also, wu2's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu2.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// First work unit returned to the second user should be wu2.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(secondUser);
			assertTrue(firstWU.getTaskID().equals(wu2.getTaskID()));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// Second work unit returned to the first user should be wu1.
		try {
			WorkUnit secondWU = transitioner.getWorkUnit(firstUser);
			assertTrue(secondWU.getTaskID().equals(wu1.getTaskID()));
			// Also, wu1's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu1.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// Second work unit returned to the second user should be wu1.
		try {
			WorkUnit secondWU = transitioner.getWorkUnit(secondUser);
			assertTrue(secondWU.getTaskID().equals(wu1.getTaskID()));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// There shouldn't be any available work units left, so we should get an exception.
		try {
			WorkUnit thirdWU = transitioner.getWorkUnit(firstUser);
			System.out.println(thirdWU);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
		
		// There shouldn't be any avilable work units left, so we should get an exception.
		try {
			WorkUnit thirdWU = transitioner.getWorkUnit(secondUser);
			System.out.println(thirdWU);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
	}
	
	/**
	 * Tests the LocalTransitioner's optimistic work unit schedulilng policy.
	 */
	public void testOptimisticWorkUnitSchedulingPolicy() {
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		Volunteer firstUser = v1.find()[0];
		
		// wu1 is an ingress work unit with default priority (0).
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.setData(new byte[] { 1, 2, 3, 4 });
		wu1.save();
		transitioner.addIngressWorkUnit(wu1);
		
		// wu2 is an ingress work unit with priority 1.
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.setData(new byte[] { 1, 2, 3, 4 });
		wu2.save();
		transitioner.addIngressWorkUnit(wu2);
		
		Volunteer v2 = new Volunteer();
		v2.setUserID("toth");
		v2.setEMailAddress("toth@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Volunteer secondUser = v2.find()[0];
		
		// Make sure that the project is expecting 1 result for every work unit.
		p.setMinNumResults(1);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {
			
		}
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// First work unit returned to the first user should be wu1.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(firstUser);
			assertTrue(firstWU.getTaskID().equals(wu1.getTaskID()));
			// Also, wu1's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu1.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// First work unit returned to the second user should be wu2 
		// because wu1 already has been assigned to enough users.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(secondUser);
			assertTrue(firstWU.getTaskID().equals(wu2.getTaskID()));
			// Also, wu2's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu2.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// Now first user is done with wu1 and is returning a result.
		Result r1a = TaskFactory.createIngressResult();
		r1a.setData(new byte[] { 1, 2, 3, 4 });
		r1a.setVolunteer(firstUser);
		wu1.setState(TaskState.Pending);
		r1a.setWorkUnit(wu1.find()[0]);
		r1a.save();
		assertTrue(Result.findAll().length == 1);
		transitioner.addIngressResult(r1a);
		firstUser = v1.find()[0];
		assertTrue(firstUser.getScore() == 1);
		
		// Mark this result as valid.
		transitioner.markResultValid(r1a);
		
		// Select the canonical result.
		transitioner.selectCanonicalResult(r1a);
		
		// The first work unit should now be retired.
		wu1.setState(TaskState.Egress);
		assertTrue(wu1.find()[0].getState() == TaskState.Egress);
		
		// Now second user is done with wu2 and is returning a result.
		Result r1b = TaskFactory.createIngressResult();
		r1b.setData(new byte[] { 1, 2, 3, 4 });
		r1b.setVolunteer(secondUser);
		wu2.setState(TaskState.Pending);
		r1b.setWorkUnit(wu2.find()[0]);
		r1b.save();
		assertTrue(Result.findAll().length == 2);
		transitioner.addIngressResult(r1b);
		secondUser = v2.find()[0];
		assertTrue(secondUser.getScore() == 1);
		
		transitioner.markResultValid(r1b);
		transitioner.selectCanonicalResult(r1b);
		
		// The second work unit should now be retired.
		wu2.setState(TaskState.Egress);
		assertTrue(wu2.find()[0].getState() == TaskState.Egress);
	}
	
	public void testWorkUnitExpiration() {
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		Volunteer firstUser = v1.find()[0];
		
		// wu1 is an ingress work unit with default priority (0).
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.setData(new byte[] { 1, 2, 3, 4 });
		wu1.setExpirationTime(5);	// This work unit will self destruct in 5 seconds.
		wu1.save();
		transitioner.addIngressWorkUnit(wu1);
		
		Volunteer v2 = new Volunteer();
		v2.setUserID("toth");
		v2.setEMailAddress("toth@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Volunteer secondUser = v2.find()[0];
		
		// Make sure that the project is expecting 1 result for every work unit.
		p.setMinNumResults(1);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {
			
		}
		
		// Re-initialize the transitioner so that it re-reads the work units from the database.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		// First work unit returned to the first user should be wu1, although this is all time sensitive.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(firstUser);
			assertTrue(firstWU.getTaskID().equals(wu1.getTaskID()));
			// Also, wu1's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu1.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		// The second user shouldn't get a work unit due to the transitioner's 
		// optimistic scheduling policy.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(secondUser);
			System.out.println(firstWU.getTaskID());
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
		
		// After sleeping for 5 seconds the work unit should time out.
		try {
			Thread.sleep(5000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		// Now the second user should get the first work unit.
		try {
			WorkUnit firstWU = transitioner.getWorkUnit(secondUser);
			assertTrue(firstWU.getTaskID().equals(wu1.getTaskID()));
			// Also, wu1's state should now be pending.
			WorkUnit search = TaskFactory.createWorkUnit();
			search.setTaskID(wu1.getTaskID());
			search.setState(TaskState.Pending);
			assertTrue(search.find().length == 1);
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Tests the entire life cycle of a work unit and its associated results.
	 * Both begin as ingress tasks, then make state transitions until the work 
	 * unit is retired, a canonical result is selected, and the other results 
	 * are deleted.
	 */
	public void testEntireTaskLifeCycle() {
		// Add a couple users to the system.
		Volunteer v1 = new Volunteer();
		v1.setUserID("jdb");
		v1.setEMailAddress("jdb@cs.wpi.edu");
		v1.setJoinDate(new Date());
		v1.save();
		Volunteer firstUser = v1.find()[0];
		
		Volunteer v2 = new Volunteer();
		v2.setUserID("toth");
		v2.setEMailAddress("toth@cs.wpi.edu");
		v2.setJoinDate(new Date());
		v2.save();
		Volunteer secondUser = v2.find()[0];
		
		// Add a couple of work units to the transitioner.
		WorkUnit wu1 = TaskFactory.createIngressWorkUnit();
		wu1.setData(new byte[] { 1, 2, 3, 4 });
		wu1.save();
		WorkUnit firstWU = wu1.find()[0];
		transitioner.addIngressWorkUnit(wu1);
		
		WorkUnit wu2 = TaskFactory.createIngressWorkUnit();
		wu2.setData(new byte[] { 1, 2, 3, 4 });
		wu2.save();
		WorkUnit secondWU = wu2.find()[0];
		transitioner.addIngressWorkUnit(wu2);
		
		// Make sure that the project is expecting 2 results for every work unit.
		p.setMinNumResults(2);
		try {
			p.saveOrUpdate();
		} catch (HibernateException e) {
			
		}
		
		// Re-initialize the transitioner so that it gets the new project settings.
		transitioner = null;
		transitioner = new LocalTransitioner();
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 2);
		assertTrue(transitioner.getPendingWorkQueue().size() == 0);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 0);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// First user should get wu1, and wu1 should now be in the pending queue.
		try {
			WorkUnit wu = transitioner.getWorkUnit(firstUser);
			assertTrue(wu.getTaskID().equals(wu1.getTaskID()));
			assertTrue(wu.getUsers().contains(firstUser));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 1);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 0);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// Second user should also get wu1, and wu1 should still be in the pending queue.
		try {
			WorkUnit wu = transitioner.getWorkUnit(secondUser);
			assertTrue(wu.getTaskID().equals(wu1.getTaskID()));
			assertTrue(wu.getUsers().contains(secondUser));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 1);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 0);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// Now first user is done with wu1 and is returning a result.
		firstWU.setState(TaskState.Pending);
		Result r1a = TaskFactory.createIngressResult();
		r1a.setData(new byte[] { 1, 2, 3, 4 });
		r1a.setVolunteer(firstUser);
		r1a.setWorkUnit(firstWU.find()[0]);
		r1a.save();
		assertTrue(Result.findAll().length == 1);
		transitioner.addIngressResult(r1a);
		firstUser = v1.find()[0];
		assertTrue(firstUser.getScore() == 1);
		
		assertTrue(transitioner.getIngressResultQueue().size() == 1);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 0);

		// The validator marks firstUser's result as valid.
		transitioner.markResultValid(r1a);
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 1);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 0);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// First user requests a new work unit, and gets wu2.
		try {
			WorkUnit wu = transitioner.getWorkUnit(firstUser);
			assertTrue(wu.getTaskID().equals(wu2.getTaskID()));
			assertTrue(wu.getUsers().contains(firstUser));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 2);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 0);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// Now second user is done with wu1 and is returning a result.
		Result r1b = TaskFactory.createIngressResult();
		r1b.setData(new byte[] { 1, 2, 3, 4 });
		r1b.setVolunteer(secondUser);
		r1b.setWorkUnit(firstWU.find()[0]);
		r1b.save();
		assertTrue(Result.findAll().length == 2);
		transitioner.addIngressResult(r1b);
		secondUser = v2.find()[0];
		assertTrue(secondUser.getScore() == 1);
		
		assertTrue(transitioner.getIngressResultQueue().size() == 1);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 0);

		// The validator marks secondUser's result as valid.
		transitioner.markResultValid(r1b);
		
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 2);
		assertTrue(transitioner.getValidResultQueue().size() == 0);
		
		// There should now be one "canonical" work unit.
		assertTrue(transitioner.getCanonicalWorkQueue().size() == 1);
		
		// Select the canonical result.
		transitioner.selectCanonicalResult(r1b);
		
		// Canonical work queue should now be empty.
		assertTrue(transitioner.getCanonicalWorkQueue().size() == 0);
		
		// One of the two results should have been deleted.
		assertTrue(Result.findAll().length == 1);
		
		// wu1 should now be in the retired queue.
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 1);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 1);
		
		// Second user should get wu2
		try {
			WorkUnit wu = transitioner.getWorkUnit(secondUser);
			assertTrue(wu.getTaskID().equals(wu2.getTaskID()));
			assertTrue(wu.getUsers().contains(secondUser));
		} catch (EmptyQueueException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 1);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 1);
		
		// Now first user is done with wu2 and is returning a result.
		secondWU.setState(TaskState.Pending);
		Result r2a = TaskFactory.createIngressResult();
		r2a.setData(new byte[] { 1, 2, 3, 4 });
		r2a.setVolunteer(firstUser);
		r2a.setWorkUnit(secondWU.find()[0]);
		r2a.save();
		transitioner.addIngressResult(r2a);
		firstUser = v1.find()[0];
		assertTrue(firstUser.getScore() == 2);
		
		assertTrue(transitioner.getIngressResultQueue().size() == 1);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 1);

		// The validator marks firstUser's result as valid.
		transitioner.markResultValid(r2a);
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 1);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 1);
		
		// First user requests a new work unit, but there are none left.
		try {
			WorkUnit wu = transitioner.getWorkUnit(firstUser);
			System.out.println(wu);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 1);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 1);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 1);
		
		// Now second user is done with wu2 and is returning a result.
		Result r2b = TaskFactory.createIngressResult();
		r2b.setData(new byte[] { 1, 2, 3, 4 });
		r2b.setVolunteer(secondUser);
		r2b.setWorkUnit(secondWU.find()[0]);
		r2b.save();
		transitioner.addIngressResult(r2b);
		secondUser = v2.find()[0];
		assertTrue(secondUser.getScore() == 2);
		
		assertTrue(transitioner.getIngressResultQueue().size() == 1);
		assertTrue(transitioner.getPendingResultQueue().size() == 1);
		assertTrue(transitioner.getValidResultQueue().size() == 1);

		// The validator marks firstUser's result as valid.
		transitioner.markResultValid(r2b);
		
		// There should now be one "canonical" work unit.
		assertTrue(transitioner.getCanonicalWorkQueue().size() == 1);
		
		// Select this result as the canonical result.
		transitioner.selectCanonicalResult(r2b);
		
		// Canonical work queue should now be empty.
		assertTrue(transitioner.getCanonicalWorkQueue().size() == 0);
		
		// wu2 should now be in the retired queue.
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 0);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 2);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 2);
		
		// There should be no work units left.
		try {
			WorkUnit wu = transitioner.getWorkUnit(secondUser);
			System.out.println(wu);
			assertTrue(false);
		} catch (EmptyQueueException e) {
			assertTrue(true);
		}
		
		assertTrue(transitioner.getIngressWorkQueue().size() == 0);
		assertTrue(transitioner.getPendingWorkQueue().size() == 0);
		assertTrue(transitioner.getRetiredWorkQueue().size() == 2);
		assertTrue(transitioner.getIngressResultQueue().size() == 0);
		assertTrue(transitioner.getPendingResultQueue().size() == 0);
		assertTrue(transitioner.getValidResultQueue().size() == 2);
	}
	
	// TODO Test spot check failure handling.
}
