package edu.wpi.prc.server.transitioner;

import java.io.File;

import org.apache.xmlrpc.WebServer;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.persistence.InvalidLogPathException;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;

/**
 * The RemoteTransitoner is intended to run on a different machine than the 
 * PRCServer to reduce the load on that machine.  The RemoteTransitioner 
 * maintains its own connections to the database, performs all localTransitioner 
 * functions, and communicates with the PRCServer via XML-RPC calls.  Since HSQLDB 
 * only allows a single process to access the database at any given time, projects 
 * that use the HSQLDB database cannot use a RemoteTransitioner.  These projects should 
 * use the LocalTransitioner instead.
 * @author James Baldassari
 */
public class RemoteTransitioner implements Transitioner {
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private static RemoteTransitioner remoteTransitioner = null;
	private Transitioner localTransitioner = null;
	private Project project = null;
	private WebServer xmlrpcServer = null;
	private boolean running = false;
	
	/**
	 * Initializes the remote transitioner.
	 * Hibernate must be configured before this constructor is called.
	 * @throws MissingConfigurationFileException
	 */
	private RemoteTransitioner() throws MissingConfigurationFileException {
		assert(HibernateController.instance().isConfigured() == true);
		project = Project.instance();
		localTransitioner = new LocalTransitioner();
		initXmlRpcServer();
	}

	/**
	 * Initializes the XML-RPC server.
	 */
	private void initXmlRpcServer() {
		// Instantiate the web server.
		try {
			int serverPort = Integer.parseInt(project.getTransitionerURI().split(":")[1]);
			xmlrpcServer = new WebServer(serverPort);
			xmlrpcServer.addHandler("transitioner", new RemoteTransitionerRPCHandler(this));
		} catch (Exception e) {
			SysLogger.println("Error initializing the XML-RPC server: " + e);
			System.err.println("Error initializing the XML-RPC server: " + e);
			System.exit(EXIT_ERROR);
		}
	}
	
	public TransitionerQueue<WorkUnit> getIngressWorkQueue() {
		SysLogger.println("RemoteTransitioner.getIngressWorkQueue()");
		return localTransitioner.getIngressWorkQueue();
	}

	public TransitionerQueue<WorkUnit> getPendingWorkQueue() {
		SysLogger.println("RemoteTransitioner.getPendingWorkQueue()");
		return localTransitioner.getPendingWorkQueue();
	}

	public TransitionerQueue<WorkUnit> getCanonicalWorkQueue() {
		SysLogger.println("RemoteTransitioner.getCanonicalWorkQueue()");
		return localTransitioner.getCanonicalWorkQueue();
	}

	public TransitionerQueue<WorkUnit> getRetiredWorkQueue() {
		SysLogger.println("RemoteTransitioner.getRetiredWorkQueue()");
		return localTransitioner.getRetiredWorkQueue();
	}

	public TransitionerQueue<Result> getIngressResultQueue() {
		SysLogger.println("RemoteTransitioner.getIngressResultQueue()");
		return localTransitioner.getIngressResultQueue();
	}

	public TransitionerQueue<Result> getPendingResultQueue() {
		SysLogger.println("RemoteTransitioner.getPendingResultQueue()");
		return localTransitioner.getPendingResultQueue();
	}

	public TransitionerQueue<Result> getValidResultQueue() {
		SysLogger.println("RemoteTransitioner.getValidResultQueue()");
		return localTransitioner.getValidResultQueue();
	}

	public void addIngressWorkUnit(WorkUnit wu) {
		SysLogger.println("RemoteTransitioner.addIngressWorkUnit()");
		localTransitioner.addIngressWorkUnit(wu);	
	}

	public void addIngressResult(Result r) {
		SysLogger.println("RemoteTransitioner.addIngressResult()");
		localTransitioner.addIngressResult(r);	
	}

	public Result[] getAssociatedIngressResults(WorkUnit wu) {
		SysLogger.println("RemoteTransitioner.getAssociatedIngressResults()");
		return localTransitioner.getAssociatedIngressResults(wu);
	}

	public Result[] getAssociatedPendingResults(WorkUnit wu) {
		SysLogger.println("RemoteTransitioner.getAssociatedPendingResults()");
		return localTransitioner.getAssociatedPendingResults(wu);
	}

	public WorkUnit getWorkUnit(Volunteer v) throws EmptyQueueException {
		SysLogger.println("RemoteTransitioner.getWorkUnit()");
		return localTransitioner.getWorkUnit(v);
	}

	public WorkUnit getCanonicalWorkUnit() throws EmptyQueueException {
		SysLogger.println("RemoteTransitioner.getCanonicalWorkUnit()");
		return localTransitioner.getCanonicalWorkUnit();
	}

	public Result getResult() throws EmptyQueueException {
		SysLogger.println("RemoteTransitioner.getResult()");
		return localTransitioner.getResult();
	}

	public void markResultValid(Result validResult) {
		SysLogger.println("RemoteTransitioner.markResultValid()");
		localTransitioner.markResultValid(validResult);	
	}

	public void markResultInvalid(Result invalidResult) {
		SysLogger.println("RemoteTransitioner.markResultInvalid()");
		localTransitioner.markResultInvalid(invalidResult);
	}

	public void selectCanonicalResult(Result canonicalResult) {
		SysLogger.println("RemoteTransitioner.selectCanonicalResult()");
		localTransitioner.selectCanonicalResult(canonicalResult);	
	}
	
	/**
	 * Starts the remote transitioner.
	 */
	public void run() {
		if (running == false) {
			// Start the XML-RPC server thread.
			SysLogger.println("Starting XML-RPC Server.");
			System.out.println("Starting XML-RPC Server.");
			xmlrpcServer.start();
			running = true;
		}
	}
	
	/**
	 * @return Returns true if the remote transitioner is running, false otherwise.
	 */
	public boolean isRunning() {
		return running;
	}
	
	/**
	 * Performs any tasks necessary to clearnly shut down the remote transitioner.
	 */
	public void shutdown() {
		// Shut down the XML-RPC server.
		SysLogger.println("Shutting down Remote Transitioner.");
		System.out.println("Shutting down Remote Transitioner.");
		running = false;
	}
	
	/**
	 * Stops the XML-RPC server.
	 */
	public void stopServer() {
		xmlrpcServer.shutdown();
	}
	
	/**
	 * Hibernate must be configured before getting the singleton instance of this class.
	 * @return Returns an instance of the RemoteTransitioner class.
	 * @throws MissingConfigurationFileException
	 */
	public static RemoteTransitioner instance() throws MissingConfigurationFileException {
		if (remoteTransitioner == null)
			remoteTransitioner = new RemoteTransitioner();
		return remoteTransitioner;
	}
	
	/**
	 * Main control for the remote transitioner.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		RemoteTransitioner remoteTransitioner = null;
		
		// Hibernate MUST be initialized before calling RemoteTransitioner.instance().
		try {
			HibernateController.instance().initialize();
		} catch (MissingConfigurationFileException e) {
			System.out.println(e);
			System.err.println(e);
			System.exit(EXIT_ERROR);
		}
		// Now we can use Project.instance() to get the rest of the project configuration.
		
		try { 
			SysLogger.initialize("log" + File.separator + "transitioner.log");
		} catch (InvalidLogPathException e) {
			e.printStackTrace();
		}
		
		try {
			remoteTransitioner = RemoteTransitioner.instance();
		} catch (MissingConfigurationFileException e) {
			System.out.println(e);
			System.err.println(e);
			System.exit(EXIT_ERROR);
		}
		remoteTransitioner.run();
		
		SysLogger.println("Remote Transitioner Ready.");
		System.out.println("Remote Transitioner Ready.");
		
		// XML-RPC server is started in a new thread, so just put the main thread 
		// to sleep while the remote transitioner is running.
		while (remoteTransitioner.isRunning()) {
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		// Shut down the XML-RPC server.
		remoteTransitioner.stopServer();
		
		// Shut down Hibernate.
		HibernateController.instance().shutdown();
		
		SysLogger.println("Remote Transitioner Halted.");
		System.out.println("Remote Transitioner Halted.");
		System.exit(EXIT_SUCCESS);
	}
}
