package edu.wpi.prc.server.transitioner;

import java.util.Vector;

import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.server.InvalidPasswordException;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;

/**
 * Handles XML-RPC calls for the RemoteTransitioner.
 * @author James Baldassari
 */
public class RemoteTransitionerRPCHandler {
	private RemoteTransitioner transitioner = null;
	
	/**
	 * Contructor that takes a reference to the remote transition that created this RPC handler.
	 * @param transitioner The RemoteTransitioner object that instantiated the instance of this class.
	 */
	public RemoteTransitionerRPCHandler(RemoteTransitioner transitioner) {
		this.transitioner = transitioner;
	}
	
	/**
	 * @return Returns the ingress work queue. 
	 */
	public Vector getIngressWorkQueue() {
		TransitionerQueue<WorkUnit> ingressQueue = transitioner.getIngressWorkQueue();
		Vector<String> ingressQueueVector = new Vector<String>(ingressQueue.size());
		for (WorkUnit wu : ingressQueue)
			ingressQueueVector.add(wu.getTaskID());
		return ingressQueueVector;
	}
	
	/**
	 * @return Returns the pending work queue.
	 */
	public Vector getPendingWorkQueue() {
		TransitionerQueue<WorkUnit> pendingQueue = transitioner.getPendingWorkQueue();
		Vector<String> pendingQueueVector = new Vector<String>(pendingQueue.size());
		for (WorkUnit wu : pendingQueue)
			pendingQueueVector.add(wu.getTaskID());
		return pendingQueueVector;
	}
	
	/**
	 * @return Returns the work units that have enough results to select a canonical result.
	 */
	public Vector getCanonicalWorkQueue() {
		TransitionerQueue<WorkUnit> canonicalQueue = transitioner.getCanonicalWorkQueue();
		Vector<String> canonicalQueueVector = new Vector<String>(canonicalQueue.size());
		for (WorkUnit wu : canonicalQueue)
			canonicalQueueVector.add(wu.getTaskID());
		return canonicalQueueVector;
	}
	
	/**
	 * @return Returns the retired work queue.
	 */
	public Vector getRetiredWorkQueue() {
		TransitionerQueue<WorkUnit> retiredQueue = transitioner.getRetiredWorkQueue();
		Vector<String> retiredQueueVector = new Vector<String>(retiredQueue.size());
		for (WorkUnit wu : retiredQueue)
			retiredQueueVector.add(wu.getTaskID());
		return retiredQueueVector;
	}
	
	/**
	 * @return Returns the ingress result queue.
	 */
	public Vector getIngressResultQueue() {
		TransitionerQueue<Result> ingressQueue = transitioner.getIngressResultQueue();
		Vector<String> ingressQueueVector = new Vector<String>(ingressQueue.size());
		for (Result r : ingressQueue)
			ingressQueueVector.add(r.getTaskID());
		return ingressQueueVector;
	}
	
	/**
	 * @return Returns the pending result queue.
	 */
	public Vector getPendingResultQueue() {
		TransitionerQueue<Result> pendingQueue = transitioner.getPendingResultQueue();
		Vector<String> pendingQueueVector = new Vector<String>(pendingQueue.size());
		for (Result r : pendingQueue)
			pendingQueueVector.add(r.getTaskID());
		return pendingQueueVector;
	}
	
	/**
	 * @return Returns the valid result queue.
	 */
	public Vector getValidResultQueue() {
		TransitionerQueue<Result> validQueue = transitioner.getValidResultQueue();
		Vector<String> validQueueVector = new Vector<String>(validQueue.size());
		for (Result r : validQueue)
			validQueueVector.add(r.getTaskID());
		return validQueueVector;
	}
	
	/**
	 * Enqueues a work unit in the ingress work queue.
	 * NOTE: work unit state MUST be set to TaskState.Ingress before calling this method.
	 * @param taskID The task ID of the work unit to add.
	 */
	public boolean addIngressWorkUnit(String taskID) {
		boolean success = false;
		WorkUnit search = TaskFactory.createWorkUnit(taskID);
		try {
			WorkUnit ingressWU = search.find()[0];
			transitioner.addIngressWorkUnit(ingressWU);
			success = true;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.addIngressWorkUnit: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.addIngressWorkUnit: " + e + ".");
			e.printStackTrace();
		}
		return success;
	}
	
	/**
	 * Enqueues a result in the ingress result queue.
	 * NOTE: result state MUST be set to TaskState.Ingress before calling this method.
	 * @param taskID The task ID of the result to add.
	 */
	public boolean addIngressResult(String taskID) {
		boolean success = false;
		Result search = TaskFactory.createResult(taskID);
		try {
			Result ingressResult = search.find()[0];
			transitioner.addIngressResult(ingressResult);
			success = true;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.addIngressResult: Could not find result " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.addIngressResult: " + e + ".");
			e.printStackTrace();
		}
		return success;
	}
	
	/**
	 * Finds all ingress Results associated with a given WorkUnit.
	 * @param taskID The task ID of the WorkUnit for which to find all associated ingress Results.
	 * @return Vector of ingress Results associated with the given WorkUnit or a Vector of length 0 if none were found.
	 */
	public Vector getAssociatedIngressResults(String taskID) {
		WorkUnit search = TaskFactory.createWorkUnit(taskID);
		Vector<String> resultsVector = new Vector<String>();
		try {
			WorkUnit ingressWU = search.find()[0];
			Result[] associatedResults = transitioner.getAssociatedIngressResults(ingressWU);
			for (Result r : associatedResults)
				resultsVector.add(r.getTaskID());
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getAssociatedIngressResults: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getAssociatedIngressResults: " + e + ".");
			e.printStackTrace();
		}
		return resultsVector;
	}
	
	/**
	 * Finds all pending Results associated with a given WorkUnit.
	 * @param taskID The task ID of the WorkUnit for which to find all associated pending Results.
	 * @return Vector of pending Results associated with the given WorkUnit or an Vector of length 0 if none were found.
	 */
	public Vector getAssociatedPendingResults(String taskID) {
		WorkUnit search = TaskFactory.createWorkUnit(taskID);
		Vector<String> resultsVector = new Vector<String>();
		try {
			WorkUnit pendingWU = search.find()[0];
			Result[] associatedResults = transitioner.getAssociatedPendingResults(pendingWU);
			for (Result r : associatedResults)
				resultsVector.add(r.getTaskID());
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getAssociatedPendingResults: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getAssociatedPendingResults: " + e + ".");
			e.printStackTrace();
		}
		return resultsVector;
	}
	
	/**
	 * Gets the next work unit to be processed.  First, it will try to find  
	 * a work unit in the pending work queue that has not already been sent 
	 * to the given volunteer.  If there are no satisfiable work units in 
	 * the pending work queue, it will try to return a work unit from the 
	 * ingress work queue.  If there are no work units in the ingress work 
	 * queue, an EmptyQueueException will be thrown.
	 * @param userID The user ID of the volunteer who will receive it.
	 * @param eMailAddr The e-mail address of the volunteer who will receive it.
	 * @return The task ID of the next work unit to be processed.
	 * @throws EmptyQueueException
	 */
	public String getWorkUnit(String userID, String eMailAddr) throws EmptyQueueException {
		Volunteer search = new Volunteer();
		search.setUserID(userID);
		search.setEMailAddress(eMailAddr);
		String taskID = "";
		try {
			Volunteer v = search.find()[0];
			WorkUnit wu = transitioner.getWorkUnit(v);
			taskID = wu.getTaskID();
		} catch (EmptyQueueException e) {
			throw e;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getWorkUnit: Could not find user " + userID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.getWorkUnit: " + e + ".");
			e.printStackTrace();
		}
		return taskID;
	}
	
	/**
	 * Gets the first work unit that has enough Results.
	 * @return The first work unit in the canonical work queue.
	 * @throws EmptyQueueException
	 */
	public String getCanonicalWorkUnit() throws EmptyQueueException {
		WorkUnit wu = transitioner.getCanonicalWorkUnit();
		return wu.getTaskID();
	}
	
	/**
	 * Gets the first result that needs to be validated.
	 * @return The first result in the ingress queue.
	 * @throws EmptyQueueException
	 */
	public String getResult() throws EmptyQueueException {
		Result r = transitioner.getResult();
		return r.getTaskID();
	}
	
	/**
	 * Marks a Result as valid.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit, 
	 * the Result is moved from the ingress result queue to 
	 * the pending result queue.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit has 
	 * been reached, all Results for that WorkUnit are removed 
	 * from the pending result queue, a canonical result is chosen, 
	 * and the canonical result is moved to the valid result queue. 
	 * @param taskID The task ID of the result from the ingress result queue to be marked valid.
	 */
	public boolean markResultValid(String taskID) {
		boolean success = false;
		Result search = TaskFactory.createResult(taskID);
		try {
			Result r = search.find()[0];
			transitioner.markResultValid(r);
			success = true;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.markResultValid: Could not result " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.markResultValid: " + e + ".");
		}
		return success;
	}
	
	/**
	 * Marks a result as invalid.  This will remove the result 
	 * from the ingress queue and delete it from the database.  
	 * If the result is a spot check result, additional steps 
	 * may be taken depending on the project's spot check 
	 * failure action.
	 * @param taskID The task ID of the result to be marked invalid.
	 */
	public boolean markResultInvalid(String taskID) {
		boolean success = false;
		Result search = TaskFactory.createResult(taskID);
		try {
			Result r = search.find()[0];
			transitioner.markResultInvalid(r);
			success = true;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.markResultInvalid: Could not result " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.markResultInvalid: " + e + ".");
		}
		return success;
	}
	
	/**
	 * Designates a canonical Result by removing all related results from 
	 * the ingress and pending queues, adding the canonical result to 
	 * the valid queue, and retiring the canonical result's work unit.
	 * @param taskID The task ID of the canonical result to select.
	 */
	public boolean selectCanonicalResult(String taskID) {
		boolean success = false;
		Result search = TaskFactory.createResult(taskID);
		try {
			Result r = search.find()[0];
			transitioner.selectCanonicalResult(r);
			success = true;
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.selectCanonicalResult: Could not result " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerRPCHandler.selectCanonicalResult: " + e + ".");
		}
		return success;
	}
	
	/**
	 * Performs any necessary tasks to safely shut down the Transitioner.
	 */
	public boolean shutdown(String password) throws InvalidPasswordException {
		if (!password.equals(Project.instance().getPassword())) {
			SysLogger.println("WARNING: Attempt to shut down the project server failed due to the use of an incorrect password.");
			throw new InvalidPasswordException();
		}
		
		transitioner.shutdown();
		return true;
	}
}
