package edu.wpi.prc.server.transitioner;

import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;

/**
 * State machine for tasks.
 * It controls transitions of tasks between work and result queues.
 * @author James Baldassari
 */
public interface Transitioner {
	/**
	 * @return Returns the ingressWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getIngressWorkQueue();
	
	/**
	 * @return Returns the pendingWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getPendingWorkQueue();
	
	/**
	 * @return Returns the work units that have enough results to select a canonical result.
	 */
	public TransitionerQueue<WorkUnit> getCanonicalWorkQueue();
	
	/**
	 * @return Returns the retiredWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getRetiredWorkQueue();
	
	/**
	 * @return Returns the ingressResultQueue.
	 */
	public TransitionerQueue<Result> getIngressResultQueue();
	
	/**
	 * @return Returns the pendingResultQueue.
	 */
	public TransitionerQueue<Result> getPendingResultQueue();
	
	/**
	 * @return Returns the validResultQueue.
	 */
	public TransitionerQueue<Result> getValidResultQueue();
	
	/**
	 * Enqueues a work unit in the ingress work queue.
	 * @param wu The work unit to add.
	 */
	public void addIngressWorkUnit(WorkUnit wu);
	
	/**
	 * Enqueues a result in the ingress result queue.
	 * @param r The result to add.
	 */
	public void addIngressResult(Result r);
	
	/**
	 * Finds all ingress Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated ingress Results.
	 * @return Array of ingress Results associated with the given WorkUnit or null if none were found.
	 */
	public Result[] getAssociatedIngressResults(WorkUnit wu);
	
	/**
	 * Finds all pending Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated pending Results.
	 * @return Array of pending Results associated with the given WorkUnit or null if none were found.
	 */
	public Result[] getAssociatedPendingResults(WorkUnit wu);
	
	/**
	 * Gets the next work unit to be processed.  First, it will try to find  
	 * a work unit in the pending work queue that has not already been sent 
	 * to the given volunteer.  If there are no satisfiable work units in 
	 * the pending work queue, it will try to return a work unit from the 
	 * ingress work queue.  If there are no work units in the ingress work 
	 * queue, an EmptyQueueException will be thrown.
	 * @param v The volunteer who will receive t
	 * @return The next work unit to be processed.
	 * @throws EmptyQueueException
	 */
	public WorkUnit getWorkUnit(Volunteer v) throws EmptyQueueException;
	
	/**
	 * Gets the first WorkUnit that has enough Results.
	 * @return The first WorkUnit in the canonical work queue.
	 * @throws EmptyQueueException
	 */
	public WorkUnit getCanonicalWorkUnit() throws EmptyQueueException;
	
	/**
	 * Gets the first result that needs to be validated.
	 * @return The first result in the ingress queue.
	 * @throws EmptyQueueException
	 */
	public Result getResult() throws EmptyQueueException;
	
	/**
	 * Marks a Result as valid.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit, 
	 * the Result is moved from the ingress result queue to 
	 * the pending result queue.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit has 
	 * been reached, all Results for that WorkUnit are removed 
	 * from the pending result queue, a canonical result is chosen, 
	 * and the canonical result is moved to the valid result queue. 
	 * @param validResult The result from the ingress result queue to be marked valid.
	 */
	public void markResultValid(Result validResult);
	
	/**
	 * Marks a result as invalid.  This will remove the result 
	 * from the ingress queue and delete it from the database.  
	 * If the result is a spot check result, additional steps 
	 * may be taken depending on the project's spot check 
	 * failure action.
	 * @param invalidResult The result to be marked invalid.
	 */
	public void markResultInvalid(Result invalidResult);
	
	/**
	 * Designates a canonical Result by removing all related results from 
	 * the ingress and pending queues, adding the canonical result to 
	 * the valid queue, and retiring the canonical result's work unit.
	 * @param canonicalResult The canonical result to select.
	 */
	public void selectCanonicalResult(Result canonicalResult);
	
	/**
	 * Performs any necessary tasks to safely shut down the Transitioner.
	 */
	public void shutdown();
}
