package edu.wpi.prc.task;

import java.util.Iterator;
import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Expression;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.user.Volunteer;

/**
 * A Result is a type of Task that contains data that was  
 * analyzed by the science application and returned by a
 * client.  Results are unique, and can only have one 
 * associated volunteer.  However, WorkUnits can be sent to 
 * more than one client, so there can be several Results 
 * for the same WorkUnit.  In that case, the data contained 
 * in all of the Results should be the same.
 * @author James Baldassari
 */
public class Result extends Task {
	private WorkUnit workUnit = null;
	private Volunteer volunteer = null;

	/**
	 * Default constructor is protected so that the Result will either be created
	 * by the TaskFactory or the public constructor that takes a task ID will be 
	 * called.  This restriction ensures that the Result will always have a valid 
	 * task ID.
	 */
	protected Result() {
		
	}
	
	/**
	 * Constructor that takes a task ID.  The task ID CANNOT be null.
	 * @param taskID The task ID to set.
	 */
	protected Result(String taskID) {
		assert(taskID != null);
		super.setTaskID(taskID);
	}
	
	/**
	 * @return Returns the workUnit.
	 */
	public WorkUnit getWorkUnit() {
		return workUnit;
	}

	/**
	 * @param wu The workUnit to set.
	 */
	public void setWorkUnit(WorkUnit wu) {
		this.workUnit = wu;
	}
	
	/**
	 * @return Returns the volunteer.
	 */
	public Volunteer getVolunteer() {
		return volunteer;
	}

	/**
	 * @param user The volunteer to set.
	 */
	public void setVolunteer(Volunteer user) {
		this.volunteer = user;
	}
	
	/**
	 * Removes the associated volunteer from this result.
	 */
	public void removeUser() {
		this.volunteer = null;
	}
	
	/**
	 * Inserts this Result into the database as a new record.
	 */
	public void save() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.save(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Updates this Result in the database.
	 */
	public void update() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.update(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Finds all Results in the database that match the attributes set in this Result.
	 * @return An array of all matching Results.
	 */
	public Result[] find()  {
		Result[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			String tID = getTaskID();
			String stateStr = getStateStr();
			WorkUnit wu = getWorkUnit();
			Volunteer v = getVolunteer();
			Criteria query = hibernateSession.createCriteria(Result.class);
			
			// Dynamically build the query.
			if (tID != null)
				query.add(Expression.eq("taskID", tID));
			if (stateStr != null)
				query.add(Expression.eq("stateStr", stateStr));
			if (wu != null)
				query.add(Expression.eq("workUnit", wu));
			if (v != null)
				query.add(Expression.eq("volunteer", v));
			
			// Execute the query.
			List results = query.list();
			
			// Build array of WorkUnits
			matches = new Result[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (Result)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
			e.printStackTrace();
		} finally {
			try {
				if (hibernateSession != null)
					hibernateSession.close();
			} catch (Exception e) {
				
			}
		}
		
		return matches;
	}
	
	/**
	 * Finds all Results in the database.
	 * @return An array of all Results in the database.
	 */
	public static Result[] findAll() {
		Result[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			Criteria query = hibernateSession.createCriteria(Result.class);
			
			// Execute the query.
			List results = query.list();
			
			// Build array of WorkUnits
			matches = new Result[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (Result)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
			e.printStackTrace();
		} finally {
			try {
				if (hibernateSession != null)
					hibernateSession.close();
			} catch (Exception e) {
				
			}
		}
		
		return matches;
	}
	
	/**
	 * @return Returns an array of all Results in the Ingress state.
	 */
	public static Result[] findAllIngress() {
		Result r = new Result();
		r.setState(TaskState.Ingress);
		return r.find();
	}
	
	/**
	 * @return Returns an array of all Results in the Pending state.
	 */
	public static Result[] findAllPending() {
		Result r = new Result();
		r.setState(TaskState.Pending);
		return r.find();
	}
	
	/**
	 * @return Returns an array of all Results in the Egress state.
	 */
	public static Result[] findAllEgress() {
		Result r = new Result();
		r.setState(TaskState.Egress);
		return r.find();
	}
	
	/**
	 * @return Returns an array of all Results in the SpotCheck state.
	 */
	public static Result[] findAllSpotCheck() {
		Result r = new Result();
		r.setState(TaskState.SpotCheck);
		return r.find();
	}
	
	/**
	 * Deletes a Result from the database.
	 */
	public void delete() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.delete(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}

	/**
	 * Tests whether this Result is equal to another.
	 * @param o The object to compare Result to.
	 */
	public boolean equals(Object o) {
		try {
			Result r = (Result)o;
			return r.getTaskID().equals(this.getTaskID());
		} catch (Exception e) {
			return false;
		}
	}
}
