package edu.wpi.prc.task.queue;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;

import edu.wpi.prc.task.Prioritizable;

/**
 * Stores queues of Prioritizable objects.  Although it appears as a single
 * queue when iterating over it, the MultiPriorityQueue actually stores many 
 * different queues internally.  Each internal queue is associated with a 
 * different priority.  When iterating over the queue it is guaranteed that 
 * higher-priority objects will be returned before lower-priority ones.  If 
 * two objects in the queue have the same priority, normal FIFO queue 
 * semantics are obeyed.  When an iterator is requested, a copy of the 
 * MultiPriorityQueue is made, and the iterator uses the copy.  Therefore, 
 * if the MultiPriorityQueue is modified after an iterator is returned, a 
 * new iterator will need to be requested for those changes to be returned 
 * by the iterator.
 * @author James Baldassari
 * @param <T> The type of Prioritizable to store in this MultiPriorityQueue.
 */
public class MultiPriorityQueue<T extends Prioritizable> implements Serializable, TransitionerQueue<T> {
	private static final long serialVersionUID = 1L;
	private SortedSet<Integer> priorities = new TreeSet<Integer>(new PriorityComparator());
	private Map<Integer, LinkedList<T>> queues = new ConcurrentHashMap<Integer, LinkedList<T>>();
	int size = 0;
	
	/**
	 * Returns an iterator for this MultiPriorityQueue.  When iterating over the 
	 * queue it is guaranteed that higher-priority objects will be returned before 
	 * lower-priority ones.  If two objects in the queue have the same priority, 
	 * normal FIFO queue semantics are obeyed.  When an iterator is requested, a 
	 * copy of the MultiPriorityQueue is made, and the iterator uses the copy.  
	 * Therefore, if the MultiPriorityQueue is modified after an iterator is 
	 * returned, a new iterator will need to be requested for those changes to be 
	 * returned by the iterator.
	 */
	public Iterator<T> iterator() {
		return new MultiPriorityQueueIterator<T>(priorities, queues, size);
	}
	
	/**
	 * @return Returns the number of Prioritizable objects in the MultiPriorityQueue.
	 */
	public int size() {
		return size;
	}
	
	/**
	 * @return Returns true if there are no objects in the MultiPriorityQueue, false if 
	 * there is at least one.
	 */
	public boolean isEmpty() {
		return size == 0;
	}
	
	/**
	 * Searches through the MultiPriorityQueue for the given object.
	 * @param obj The object to search for.
	 * @return Returns true if the object was found in the MultiPriorityQueue,
	 * false otherwise.
	 */
	public boolean contains(Object obj) {
		boolean isContained = false;
		try {
			Prioritizable t = (Prioritizable) obj;
			int priority = t.getPriority();
			if (priorities.contains(priority)) {
				// Find the queue where this Prioritizable should be located.
				LinkedList<T> queue = queues.get(t.getPriority());
				isContained = queue.contains(t);			
			}
		} catch (Exception e) {
			isContained = false;
		}
		return isContained;
	}
	
	/**
	 * Checks whether all of the objects in the given collection appear in the MultiPriorityQueue.
	 * @param c The collection of objects to check for.
	 * @return Returns true if all of the objects were found in the queue, false otherwise.
	 */
	public boolean containsAll(Collection<?> c) {
		boolean allFound = true;
		for (Object element : c) {
			if (contains(element) == false) {
				allFound = false;
				break;
			}
		}
		return allFound;
	}
	
	/**
	 * Adds the given object to the MultiPriorityQueue.
	 * @param t The object to add.
	 * @return Returns true if the MultiPriorityQueue was modified as a result
	 * of adding the object, false otherwise.
	 */
	public synchronized boolean add(T t) {
		boolean modified = false;
		int priority = t.getPriority();
		// If no queue exists for this Prioritizable's priority, create one. 
		if (priorities.contains(priority) == false) {
			priorities.add(priority);
			queues.put(priority, new LinkedList<T>());
		}
		
		// Find the appropriate queue for this Prioritizable.
		LinkedList<T> queue = queues.get(priority);
		
		// Only add the Prioritizable if it does not already exist in the queue.
		if (queue.contains(t) == false) {
			// Add the Prioritizable to the appropriate queue.
			queues.get(priority).offer(t);
			modified = true;
			
			// Increase the size.
			size++;
		}
		
		return modified;
	}
	
	/**
	 * Adds all objects in the given collection to the MultiPriorityQueue.
	 * @param t The collection containing objects to add.
	 * @return Returns true if the MultiPriorityQueue was modified as a result
	 * of adding the objects, false otherwise.
	 */
	public boolean addAll(Collection<? extends T> t) {
		boolean modified = false;
		for (T p : t) {
			if (add(p) == true)
				modified = true;
		}
		return modified;
	}
	
	/**
	 * Adds all objects in the given array to the MultiPriorityQueue.
	 * @param tArray The array containing objects to add.
	 * @return Returns true if the MultiPriorityQueue was modified as a result
	 * of adding the objects, false otherwise. 
	 */
	public boolean addAll(T[] tArray) {
		boolean modified = false;
		for (T p : tArray) {
			if (add(p) == true)
				modified = true;
		}
		return modified;
	}
	
	/**
	 * Removes the given object from the MultiPriorityQueue.
	 * @param obj The object to remove.
	 * @return Returns true if the MultiPriorityQueue was modified
	 * as a result of removing the object, false otherwise.
	 */
	public boolean remove(Object obj) {
		boolean modified = false;
		try {
			Prioritizable t = (Prioritizable)obj;
			int priority = t.getPriority();
			if (priorities.contains(priority)) {
				// Find the queue where this Prioritizable should be located.
				LinkedList<T> queue = queues.get(priority);
				modified = queue.remove(t);
				
				if (modified == true) {
					// If this was the last element in the queue, remove the queue as well.
					if (queue.isEmpty() == true) {
						queues.remove(priority);
						priorities.remove(priority);
					}
					// Decrease the size.
					size--;
				}
			}
		} catch (Exception e) {
			modified = false;
		}
		return modified;
	}
	
	/**
	 * Removes the first (highest-priority) object in the MultiPriorityQueue
	 * and returns it.
	 * @return Returns the object that was removed.
	 */
	public synchronized T removeFirst() {
		T t = first();
		remove(t);
		return t;
	}
	
	/**
	 * Removes all objects contained in the given collection from the MultiPriorityQueue.
	 * @param c The collection of objects to remove.
	 * @return Returns true if the MultiPriorityQueue was modified as a result of 
	 * removing the given objects, false otherwise.
	 */
	public synchronized boolean removeAll(Collection<?> c) {
		boolean modified = false;
			for (Object element : c) {
				if (remove(element) == true)
					modified = true;
			}
		return modified;
	}
	
	/**
	 * Removes all objects from the MultiPriorityQueue except the objects
	 * contained in the given collection.
	 * @param c The collection of objects to retain.
	 * @return Returns true if the MultiPriorityQueue was modified as a result
	 * of retaining the given objects, false otherwise.
	 */
	public synchronized boolean retainAll(Collection<?> c) {	
		boolean modified = false;	
		Collection<LinkedList<T>> allQueues = queues.values();
		for (LinkedList<T> list : allQueues) {
			for (T p : list) {
				boolean retain = false;
				for (Object element : c) {
					if (p.equals(element))
						retain = true;
				}
				if (retain == false) {
					list.remove(p);
					size--;
					modified = true;
				}
			}
		}
		return modified;
	}
	
	/**
	 * @return Returns the first (highest-priority) object in the MultiPriorityQueue, 
	 * but does not remove it from the queue.
	 */
	public T peek() {
		T peeked = null;
		Iterator<T> iter = iterator();
		if (iter.hasNext()) {
			peeked = iter.next();
		}
		return peeked;
	}

	/**
	 * @return Returns the first (highest-priority) object in the MultiPriorityQueue, 
	 * but does not remove it from the queue.
	 */
	public T first() {
		if (isEmpty() == true)
			throw new NoSuchElementException("The queue is empty.");
		else
			return peek();
	}
	
	/**
	 * Removes all objects from the MultiPriorityQueue.
	 */
	public synchronized void clear() {
		Collection<LinkedList<T>> allQueues = queues.values();
		for (LinkedList<T> list : allQueues)
			list.clear();
		
		queues.clear();
		size = 0;
	}
	
	/**
	 * @return Returns an array containing all objects from the 
	 * MultiPriorityQueue, sorted according to the MultiPriorityQueue's iterator
	 * semantics.
	 */
	public Object[] toArray() {
		ArrayList<Object> array = new ArrayList<Object>(size);
		for (Integer priority : priorities) {
			LinkedList<T> queue = queues.get(priority);
			for (T p : queue)
				array.add(p);
		}
		return array.toArray();
	}

	/**
	 * @return Returns an array containing all objects from the 
	 * MultiPriorityQueue, sorted according to the MultiPriorityQueue's iterator
	 * semantics.  The runtime type of the returned array is that of the specified
	 * array.
	 */
	public <P> P[] toArray(P[] arg0) {
		ArrayList<T> array = new ArrayList<T>(size);
		for (Integer priority : priorities) {
			LinkedList<T> queue = queues.get(priority);
			for (T p : queue)
				array.add(p);
		}
		return array.toArray(arg0);
	}
	
	/**
	 * This is the iterator for the MultiPriorityQueue.  It returns all objects 
	 * in FIFO order from the highest-priority queue, then moves to the next- 
	 * highest-priority queue, returns the objects in that queue in FIFO order, 
	 * and so on.
	 * @author James Baldassari
	 * @param <P>
	 */
	private class MultiPriorityQueueIterator<P extends T> implements Iterator<P> {
		private SortedSet<Integer> priorities;
		private Map<Integer, LinkedList<P>> queues;
		private ArrayList<P> elements;
		private int index = 0;
		
		/**
		 * The constructor stores a copy of the MultiPriorityQueue.
		 * @param priorities The priorities of the queues.
		 * @param queues The queues of objects.
		 * @param size The number of objects in the MultiPriorityQueue.
		 */
		public MultiPriorityQueueIterator(SortedSet<Integer> priorities, Map<Integer, LinkedList<P>> queues, int size) {
			this.priorities = priorities;
			this.queues = queues;
			this.elements = new ArrayList<P>(size);
			populateElements();
		}
		
		/**
		 * Creates a simple array of the elements from the MultiPriorityQueue
		 * in the correct order (by priority, then FIFO).
		 */
		private void populateElements() {
			for (Integer priority : priorities) {
				LinkedList<P> list = queues.get(priority);
				for (P p : list) {
					elements.add(p);
				}
			}
		}
		
		/**
		 * @return Returns true if the iterator has another element to return,
		 * false otherwise.
		 */
		public boolean hasNext() {
			return index < elements.size();
		}

		/**
		 * @return Returns the next element.
		 */
		public P next() {
			if (hasNext()) {
				P p = elements.get(index);
				index++;
				return p;
			}
			else
				throw new NoSuchElementException("The iterator has reached the end of the queue");
		}

		/**
		 * Removes the current element.
		 */
		public void remove() {
			elements.remove(index);
		}
	}
}
