package edu.wpi.prc.task.test;

import java.util.Date;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.user.Volunteer;
import junit.framework.TestCase;

/**
 * Tests WorkUnit persistence.
 * @author James Baldassari
 */
public class WorkUnitTest extends TestCase {
	private WorkUnit wu = null;
	private Volunteer v = null;

	public WorkUnitTest(String name) {
		super(name);
		
		HibernateController.instance().initializeForUnitTests();
		wu = TaskFactory.createWorkUnit();
		v = new Volunteer();
		
		v.setUserID("jdb");
		v.setEMailAddress("jdb@cs.wpi.edu");
		v.setScore(1);
		v.setJoinDate(new Date());
		v.setLastResultDate(new Date());
		
		byte[] data = {0, 1, 2, 3};
		wu.setTaskID("T1000");
		wu.setState(TaskState.Egress);
		wu.setPoints(10);
		wu.setData(data);
		wu.incrementNumValidResults();
	}

	protected void setUp() throws Exception {
		super.setUp();
		HibernateController.instance().exportSchema();
		v.save();
	}
	
	/**
	 * Tests saving a WorkUnit to the database.
	 */
	public void testSave() {
		// Make sure this work unit has never been assigned.
		assertTrue(wu.getUsers().size() == 0);
		assertTrue(wu.getAssignmentCount() == 0);
		wu.assignUser(v);
		
		// Save the WorkUnit to the database.
		wu.save();
		
		assertTrue(wu.getUsers().size() == 1);
		assertTrue(wu.getAssignmentCount() == 1);
		
		// Find the WorkUnit in the database.
		WorkUnit wu2 = TaskFactory.createWorkUnit("T1000");
		wu2.setState(TaskState.Egress);
		WorkUnit[] matches = wu2.find();
		
		// Verify results.
		assertTrue(matches.length == 1);
		assertTrue(matches[0].getTaskID().equals("T1000"));
		assertTrue(matches[0].getState() == TaskState.Egress);
		assertTrue(matches[0].getPoints() == 10);
		assertTrue(matches[0].getNumValidResults() == 1);
		for (int ii = 0; ii < 4; ii++)
			assertTrue(matches[0].getData()[ii] == (byte)ii);
	}
	
	/**
	 * Tests work unit expiration dates.
	 */
	public void testExpiration() {
		wu.assignUser(v);
		
		// Set the work unit to expire in 3 seconds.
		wu.setExpirationTime(3);
		
		// Save the WorkUnit to the database.
		wu.save();
		
		WorkUnit[] matches = WorkUnit.findAll();
		assertFalse(matches[0].hasExpired());
		
		try {
			Thread.sleep(3000);
		} catch (InterruptedException e) {
			assertTrue(false);
		}
		
		// Work unit should have expired by now.
		assertTrue(matches[0].hasExpired());
		
		matches[0].resetExpirationDate();
		matches[0].update();
		assertFalse(matches[0].hasExpired());
		
		try {
			Thread.sleep(3000);
		} catch (InterruptedException e) {
			assertTrue(false);
		}
		
		// Work unit should have expired by now.
		assertTrue(matches[0].hasExpired());
	}

	/**
	 * Tests updating a WorkUnit in the database.
	 */
	public void testUpdate() {
		// Save the WorkUnit to the database.
		wu.save();
		
		// Find all WorkUnits.
		WorkUnit[] matches = WorkUnit.findAll();
		assertTrue(matches.length == 1);
		
		// Modify the WorkUnit.
		matches[0].setTaskID("T2000");
		matches[0].setState(TaskState.Pending);
		matches[0].incrementNumValidResults();
		
		// Save the modified WorkUnit.
		matches[0].update();
		
		// Find all WorkUnits.
		matches = null;
		matches = WorkUnit.findAll();
		
		// Verify results.
		assertTrue(matches.length == 1);
		assertTrue(matches[0].getTaskID().equals("T2000"));
		assertTrue(matches[0].getState() == TaskState.Pending);
		assertTrue(matches[0].getNumValidResults() == 2);
	}
	
	/**
	 * Tests deleting a WorkUnit from the database.
	 */
	public void testDelete() {
		// Save the WorkUnit to the database.
		wu.save();
		
		// Find all WorkUnits.
		WorkUnit[] matches = WorkUnit.findAll();
		assertTrue(matches.length == 1);
		
		// Delete the WorkUnit.
		wu.delete();
		
		// Find all WorkUnits.
		matches = null;
		matches = WorkUnit.findAll();
		
		// Verify results.
		assertTrue(matches.length == 0);
	}

}
