package edu.wpi.prc.tools.builder;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.InputStream;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.tools.builder.gui.SelectActionForm;

/**
 * This class initializes and starts the GUI used by project developers to configure their projects.
 * @author James Baldassari
 */
public class ProjectBuilder {
	public static final String DEFAULT_CONFIG_DIRECTORY = "cfg";
	public static final String SERVER_CONFIG_FILE_PATH = DEFAULT_CONFIG_DIRECTORY + File.separator + "project_server.properties";
	public static final String CLIENT_CONFIG_FILE_PATH = DEFAULT_CONFIG_DIRECTORY + File.separator + "project_client.properties";
	private static final String ICON_DIRECTORY = "edu/wpi/prc/tools/builder/gui/forms/images/";
	private static final String ERROR_ICON_FILE = "dialog-error.png";
	private static final int ERROR_ICON_FILE_SIZE = 925;
	private static final String WARNING_ICON_FILE = "dialog-warning.png";
	private static final int WARNING_ICON_FILE_SIZE = 961;
	private static final String INFORMATION_ICON_FILE = "dialog-information.png";
	private static final int INFORMATION_ICON_FILE_SIZE = 1184;
	private static final String ERROR_ICON_PATH = ICON_DIRECTORY + ERROR_ICON_FILE;
	private static final String WARNING_ICON_PATH = ICON_DIRECTORY + WARNING_ICON_FILE;
	private static final String INFORMATION_ICON_PATH = ICON_DIRECTORY + INFORMATION_ICON_FILE;
	private static ProjectBuilder builderGUI = null;		// Singleton instance.
	private Project project = null;
	private boolean newProject = true;
	private Icon errorIcon = null;
	private Icon warningIcon = null;
	private Icon informationIcon = null;

	/**
	 * Default constructor.
	 * Reads in images and builds icons from them.
	 */
	public ProjectBuilder() {
		initializeIcons();
	}
	
	/**
	 * Attempts to read in image files to create icons.
	 * If there is a problem reading the files, the default Java icons are used.
	 */
	private void initializeIcons() {
		try {			
			InputStream is = ProjectBuilder.class.getClassLoader().getResourceAsStream(ERROR_ICON_PATH);
			byte[] errorIconBytes = new byte[ERROR_ICON_FILE_SIZE];
			is.read(errorIconBytes);
			errorIcon = new ImageIcon(errorIconBytes);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		try {
			InputStream is = ProjectBuilder.class.getClassLoader().getResourceAsStream(WARNING_ICON_PATH);
			byte[] warningIconBytes = new byte[WARNING_ICON_FILE_SIZE];
			is.read(warningIconBytes);
			warningIcon = new ImageIcon(warningIconBytes);
		} catch (Exception e) {

		}
		
		try {
			InputStream is = ProjectBuilder.class.getClassLoader().getResourceAsStream(INFORMATION_ICON_PATH);
			byte[] infoIconBytes = new byte[INFORMATION_ICON_FILE_SIZE];
			is.read(infoIconBytes);
			informationIcon = new ImageIcon(infoIconBytes);
		} catch (Exception e) {

		}
	}
	
	/**
	 * @return Returns the project.
	 */
	public Project getProject() {
		return project;
	}

	/**
	 * @param project The project to set.
	 */
	public void setProject(Project project) {
		this.project = project;
	}
	
	/**
	 * @return Returns true if the user is creating a new project, false if modifying an existing project.
	 */
	public boolean isNewProject() {
		return newProject;
	}
	
	/**
	 * @return Returns the error icon to be used in error dialogs.
	 */
	public Icon getErrorIcon() {
		return errorIcon;
	}
	
	/**
	 * Returns the information icon to be used in notification dialogs.
	 */
	public Icon getInformationIcon() {
		return informationIcon;
	}
	
	/**
	 * Sets whether the user is creating a new project or modifying an existing project.
	 * @param newProject True if the user is creating a new project, false if modifying an existing project.
	 */
	public void setNewProject(boolean newProject) { 
		this.newProject = newProject;
	}

	/**
	 * Determines how to center a GUI component on the screen, then displays it.
	 * @param component The component to center.
	 */
	public static void centerComponent(JFrame component) {
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Point location = new Point((screenSize.width / 2) - (component.getPreferredSize().width / 2), (screenSize.height / 2) - (component.getPreferredSize().height / 2));
		component.setLocation(location);
	}
	
	/**
	 * Centers a child frame relative to its parent frame.
	 * @param child The child frame.
	 * @param parent The parent frame.
	 */
	public static void setRelativePosition(JFrame child, JFrame parent) {
		Point parentLocation = parent.getLocation();
		int xCoord = (parentLocation.x + (parent.getWidth() / 2)) - (child.getWidth() / 2);
		int yCoord = (parentLocation.y + (parent.getHeight() / 2)) - (child.getHeight() / 2);
		child.setLocation(new Point(xCoord, yCoord));
	}
	
	/**
	 * Adds a window close listener for a GUI component that causes the program to exit.
	 * @param component The component to which to add the listener.
	 */
	public static void addWindowCloseListener(Window component) {
		component.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent evt) {
				shutdown();
			}
		});
	}
	
	/**
	 * Displays a modal error message dialog with the specified message.
	 * @param parentComponent The component that created this dialog.
	 * @param message The message to display.
	 */
	public void showErrorDialog(Component parentComponent, String message) {
		JOptionPane.showMessageDialog(parentComponent, message, "Project Builder Error", JOptionPane.ERROR_MESSAGE, errorIcon);
	}
	
	/**
	 * Displays a modal warning message dialog with the specified message.
	 * @param parentComponent The component that created this dialog.
	 * @param message The message to display.
	 */
	public void showWarningDialog(Component parentComponent, String message) {
		JOptionPane.showMessageDialog(parentComponent, message, "Project Builder Warning", JOptionPane.WARNING_MESSAGE, warningIcon);
	}
	
	/**
	 * Displays a modal information message dialog with the specified message.
	 * @param parentComponent The component that created this dialog.
	 * @param message The message to display.
	 */
	public void showInformationDialog(Component parentComponent, String message) {
		JOptionPane.showMessageDialog(parentComponent, message, "Project Builder Notification", JOptionPane.INFORMATION_MESSAGE, informationIcon);
	}
	
	/**
	 * Cleanly shuts down the Builder GUI.
	 */
	public static void shutdown() {
		HibernateController.instance().shutdown();
		System.exit(0);
	}
	
	/**
	 * @return Returns the singleton instance of this class.
	 */
	public static ProjectBuilder instance() {
		if (builderGUI == null)
			builderGUI = new ProjectBuilder();
		return builderGUI;
	}
	
	/**
	 * Main control of the project builder GUI.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		// Force the singleton instance to be instantiated immediately to prevent delays later.
		ProjectBuilder.instance();
		
		// Create the first form.
		JFrame builderForm = new SelectActionForm();
		builderForm.setVisible(true);
		
		// Just go to sleep.
		while (true) {
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
			}
		}
	}
}
