package edu.wpi.prc.tools.builder.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import org.hibernate.HibernateException;

import com.jeta.forms.components.panel.FormPanel;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.project.SpotCheckFailureAction;
import edu.wpi.prc.tools.builder.ProjectBuilder;

/**
 * Writes project configuration files to disk and performs any necessary database commits.
 * @author James Baldassari
 */
public class ConfirmationForm extends DatabaseInitializerForm {
	private static final long serialVersionUID = 1L;
	private GUIForm previousForm = null;
	private FormPanel currentForm = null;
	private ConnectionFormThread connectionFormThread = null;
	private Project project = null;
	private boolean projectError = false;
	private boolean dbError = false;
	
	/**
	 * Constructor that takes a reference to the this form's parent form.
	 */
	public ConfirmationForm(GUIForm previousForm) {
		this.previousForm = previousForm;
		currentForm = new FormPanel("edu/wpi/prc/tools/builder/gui/forms/ConfirmationForm.jfrm");
		setLayout(new BorderLayout());
		add(currentForm, BorderLayout.CENTER);
		this.setTitle("PRC Project Builder: Step 7 of 7");
		
		currentForm.getButton(ConfirmationComponents.ID_PREVIOUS_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionPreviousClick();
			}
		});
		
		currentForm.getButton(ConfirmationComponents.ID_FINISH_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionFinishClick();
			}
		});
		
		this.setSize(430, 696);
		ProjectBuilder.addWindowCloseListener(this);
	}
	
	/**
	 * Populates all fields using the project data when this form is set to visible.
	 */
	public void setVisible(boolean isVisible) {
		if (isVisible == true) {
			populateFields();
		}
		super.setVisible(isVisible);
	}
	
	/**
	 * Reads information from the project and sets the fields in the form accordingly.
	 */
	private void populateFields() {
		project = ProjectBuilder.instance().getProject();
		
		currentForm.getTextField(ConfirmationComponents.ID_PROJECT_NAME_FIELD).setText(project.getName());
		currentForm.getTextField(ConfirmationComponents.ID_PROJECT_NAME_FIELD).setCaretPosition(0);
		currentForm.getTextField(ConfirmationComponents.ID_DATABASE_TYPE_FIELD).setText(project.getDbType().toString());
		currentForm.getTextField(ConfirmationComponents.ID_DATABASE_ADDRESS_FIELD).setText(project.getDbURI());
		currentForm.getTextField(ConfirmationComponents.ID_DATABASE_ADDRESS_FIELD).setCaretPosition(0);
		currentForm.getTextField(ConfirmationComponents.ID_DATABASE_USER_NAME_FIELD).setText(project.getDbUserName());
		currentForm.getTextField(ConfirmationComponents.ID_PROJECT_SERVER_ADDRESS_FIELD).setText(project.getServerURI().split(":")[0]);
		currentForm.getTextField(ConfirmationComponents.ID_PROJECT_SERVER_PORT_FIELD).setText(project.getServerURI().split(":")[1]);
		currentForm.getTextField(ConfirmationComponents.ID_PROJECT_CLIENT_PORT_FIELD).setText(new Integer(project.getClientPort()).toString());
		currentForm.getTextField(ConfirmationComponents.ID_SCI_APP_PORT_FIELD).setText(new Integer(project.getScienceAppPort()).toString());
		
		if (project.getTransitionerURI() == null) {
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_TYPE_FIELD).setText("Local");
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_ADDRESS_FIELD).setText("N/A");
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_PORT_FIELD).setText("N/A");
		}
		else {
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_TYPE_FIELD).setText("Remote");
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_ADDRESS_FIELD).setText(project.getTransitionerURI().split(":")[0]);
			currentForm.getTextField(ConfirmationComponents.ID_TRANSITIONER_PORT_FIELD).setText(project.getTransitionerURI().split(":")[1]);
		}
		
		if (project.getDefaultValidator() == true)
			currentForm.getTextField(ConfirmationComponents.ID_VALIDATOR_TYPE_FIELD).setText("Default");
		else
			currentForm.getTextField(ConfirmationComponents.ID_VALIDATOR_TYPE_FIELD).setText("Custom");
		
		currentForm.getTextField(ConfirmationComponents.ID_MIN_NUM_RESULTS_FIELD).setText(new Integer(project.getMinNumResults()).toString());
		currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_PROBABILITY_FIELD).setText(new Float(project.getSpotCheckProbability()).toString());
		currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_PROBABILITY_FIELD).setCaretPosition(0);
		currentForm.getTextField(ConfirmationComponents.ID_NUM_SPOT_CHECKS_FIELD).setText(new Integer(project.getMinNumSpotChecks()).toString());
		
		if (project.getSpotCheckFailureAction() == SpotCheckFailureAction.Log)
			currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_FAILURE_ACTION_FIELD).setText("Log Only");
		else if (project.getSpotCheckFailureAction() == SpotCheckFailureAction.BanUser)
			currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_FAILURE_ACTION_FIELD).setText("Ban User");
		else if (project.getSpotCheckFailureAction() == SpotCheckFailureAction.PurgeResults)
			currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_FAILURE_ACTION_FIELD).setText("Ban and Purge Results");
		
		currentForm.getTextField(ConfirmationComponents.ID_SPOT_CHECK_FAILURE_ACTION_FIELD).setCaretPosition(0);
	}
	
	/**
	 * This method is called after the database connection has been successfully established.
	 */
	protected void completeDatabaseConnection() {
		if (ProjectBuilder.instance().isNewProject() == true)
			commitDatabase();
		commitProject();
		
		if (projectError == false) {
			if (connectionFormThread != null)
				connectionFormThread.cleanup();
			ProjectBuilder.instance().showInformationDialog(this, "The project has been successfully configured!");
			ProjectBuilder.shutdown();
		}
		else {
			this.setVisible(true);
			connectionFormThread.getForm().setVisible(false);
			projectError = false;
			dbError = false;
		}
	}
	
	/**
	 * Writes project configuration files to the cfg directory.
	 */
	private void commitFiles() {
		try {
			project.writeServerProjectFile(ProjectBuilder.SERVER_CONFIG_FILE_PATH);
			project.writeClientProjectFile(ProjectBuilder.CLIENT_CONFIG_FILE_PATH);
		} catch (IOException e) {
			String message = "The project configuration files could not be saved to disk.  Please\n";
			message += "verify that you have write permissions on the \"cfg\" directory.";
			ProjectBuilder.instance().showErrorDialog(this, message);
			projectError = true;
		} 
	}
	
	/**
	 * Exports the database schema for new projects.
	 */
	private void commitDatabase() {
		try {
			HibernateController.instance().exportSchema();
		} catch (HibernateException e) {
			dbError = true;
			projectError = true;
			String message = "The Database initialization failed.  If you are using an HSQLDB\n";
			message += "database, verify that the database files exist in the location\n";
			message += "specified in your project configuration file.  If you are not\n";
			message += "using an HSQLDB database, verify that your database is\n";
			message += "configured correctly, that it is accepting connections, and\n";
			message += "that this computer can connect to the computer hosting\n";
			message += "the database.";
			ProjectBuilder.instance().showErrorDialog(this, message);			
		}
	}
	
	/**
	 * Saves the project data to the database.
	 */
	private void commitProject() {
		if (dbError == false) {
			try {
				project.saveOrUpdate();
			} catch (HibernateException e) {
				dbError = true;
				projectError = true;
				String message = "The Database initialization failed.  If you are using an HSQLDB\n";
				message += "database, verify that the database files exist in the location\n";
				message += "specified in your project configuration file.  If you are not\n";
				message += "using an HSQLDB database, verify that your database is\n";
				message += "configured correctly, that it is accepting connections, and\n";
				message += "that this computer can connect to the computer hosting\n";
				message += "the database.";
				ProjectBuilder.instance().showErrorDialog(this, message);			
			}
		}
	}
	
	/**
	 * Controls the actions that occur when the previous button is clicked.
	 */
	private void selectActionPreviousClick() {
		this.setVisible(false);
		ProjectBuilder.setRelativePosition(previousForm, this);
		previousForm.setVisible(true);
	}
	
	/**
	 * Controls the actions that occur when the finish button is clicked.
	 */
	private void selectActionFinishClick() {
		commitFiles();
		if (ProjectBuilder.instance().isNewProject() == true) {
			connectionFormThread = new ConnectionFormThread(this);
			ProjectBuilder.setRelativePosition(connectionFormThread.getForm(), this);
			connectionFormThread.start();
			// When this thread has finished establishing the database connection, it will call completeDatabaseConnection().
			
			this.setVisible(false);
		}
		else
			completeDatabaseConnection();
	}
}
