package edu.wpi.prc.tools.builder.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;

import com.jeta.forms.components.panel.FormPanel;

import edu.wpi.prc.project.SpotCheckFailureAction;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.tools.builder.ProjectBuilder;

/**
 * Allows the user to configure the spot-check settings.
 * @author James Baldassari
 */
public class SpotCheckConfigurationForm extends GUIForm {
	private static final long serialVersionUID = 1L;
	private GUIForm previousForm = null;
	private FormPanel currentForm = null;
	private GUIForm nextForm = null;
	private Project project = null;
	private SpinnerNumberModel numSpotChecksSpinnerModel;
	private SpinnerNumberModel probabilitySpinnerModel;
	private JSpinner numSpotChecksSpinner = null;
	private JSpinner probabilitySpinner = null;
	private static final String logAction = "Log Only";
	private static final String banAction = "Ban User";
	private static final String purgeAction = "Ban and Purge Results";
	
	/**
	 * Constructor that takes a reference to the this form's parent form.
	 */
	public SpotCheckConfigurationForm(GUIForm previousForm) {
		this.previousForm = previousForm;
		currentForm = new FormPanel("edu/wpi/prc/tools/builder/gui/forms/SpotCheckConfigurationForm.jfrm");
		setLayout(new BorderLayout());
		add(currentForm, BorderLayout.CENTER);
		this.setTitle("PRC Project Builder: Step 6 of 7");
		
		probabilitySpinnerModel = new SpinnerNumberModel(new Float(0.0f), new Float(0.0f), new Float(1.0f), new Float(0.01f));
		numSpotChecksSpinnerModel = new SpinnerNumberModel();
		numSpotChecksSpinnerModel.setMinimum(0);
		numSpotChecksSpinnerModel.setStepSize(1);
		numSpotChecksSpinner = currentForm.getSpinner(SpotCheckConfigurationComponents.ID_NUM_SPOT_CHECKS_SPINNER);
		probabilitySpinner = currentForm.getSpinner(SpotCheckConfigurationComponents.ID_PROBABILITY_SPINNER);
		
		numSpotChecksSpinner.setModel(numSpotChecksSpinnerModel);
		((JSpinner.DefaultEditor)numSpotChecksSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		probabilitySpinner.setModel(probabilitySpinnerModel);
		((JSpinner.DefaultEditor)probabilitySpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		
		currentForm.getButton(ServerConfigurationComponents.ID_PREVIOUS_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionPreviousClick();
			}
		});
		
		currentForm.getButton(ServerConfigurationComponents.ID_NEXT_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionNextClick();
			}
		});
		
		((JTextArea)currentForm.getComponentByName(ServerConfigurationComponents.ID_HELP_TEXT_AREA)).setCaretPosition(0);
		this.setSize(430, 372);
		ProjectBuilder.addWindowCloseListener(this);
	}
	
	/**
	 * Populates all fields using the project data when this form is set to visible.
	 */
	public void setVisible(boolean isVisible) {
		if (isVisible == true) {
			populateFields();
		}
		super.setVisible(isVisible);
	}
	
	/**
	 * Reads information from the project and sets the fields in the form accordingly.
	 */
	private void populateFields() {
		project = ProjectBuilder.instance().getProject();
		
		numSpotChecksSpinner.setValue(project.getMinNumSpotChecks());
		probabilitySpinner.setValue(project.getSpotCheckProbability());
		
		Vector<String> failureActionItems = new Vector<String>(3);
		failureActionItems.add(logAction);
		failureActionItems.add(banAction);
		failureActionItems.add(purgeAction);
		currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).setModel(new DefaultComboBoxModel(failureActionItems));
		
		SpotCheckFailureAction failureAction = project.getSpotCheckFailureAction();
		if (failureAction == null)
			currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).setSelectedItem(logAction);
		else if (failureAction == SpotCheckFailureAction.Log)
			currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).setSelectedItem(logAction);
		else if (failureAction == SpotCheckFailureAction.BanUser)
			currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).setSelectedItem(banAction);
		else if (failureAction == SpotCheckFailureAction.PurgeResults)
			currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).setSelectedItem(purgeAction);
	}
	
	/**
	 * Updates the project properties using the values in the form fields. 
	 */
	private void updateProject() {
		project.setMinNumSpotChecks((Integer)numSpotChecksSpinner.getValue());
		project.setSpotCheckProbability((Float)probabilitySpinner.getValue());
		
		String actionStr = (String)currentForm.getComboBox(SpotCheckConfigurationComponents.ID_FAILURE_ACTION_COMBO_BOX).getSelectedItem(); 
		if (actionStr.equals(logAction))
			project.setSpotCheckFailureAction(SpotCheckFailureAction.Log);
		else if (actionStr.equals(banAction))
			project.setSpotCheckFailureAction(SpotCheckFailureAction.BanUser);
		else if (actionStr.equals(purgeAction))
			project.setSpotCheckFailureAction(SpotCheckFailureAction.PurgeResults);
	}
	
	/**
	 * Checks whether all input fields are correct and displays a dialog message if there are any errors.
	 * @return Returns true if all input fields are valid, false otherwise.
	 */
	private boolean validateInput() {
		boolean isValid = true;
		if ((project.getSpotCheckProbability() > 0.0) && (project.getMinNumSpotChecks() == 0)) {
			isValid = false;
			ProjectBuilder.instance().showErrorDialog(this, "You must use at least one spot-check if the spot-check probability is greater than 0.0.");
		} else if ((project.getSpotCheckProbability() == 0.0) && (project.getMinNumSpotChecks() > 0)) {
			ProjectBuilder.instance().showWarningDialog(this, "The number of spot-checks is greater than 0, but the spot-check probability is 0.0");
		}
		return isValid;
	}
	
	/**
	 * Controls the actions that occur when the previous button is clicked.
	 */
	private void selectActionPreviousClick() {
		updateProject();
		ProjectBuilder.instance().setProject(project);
		this.setVisible(false);
		ProjectBuilder.setRelativePosition(previousForm, this);
		previousForm.setVisible(true);
	}
	
	/**
	 * Controls the actions that occur when the next button is clicked.
	 */
	private void selectActionNextClick() {
		updateProject();
		if (validateInput() == true) {
			ProjectBuilder.instance().setProject(project);
			if (nextForm == null)
				nextForm = new ConfirmationForm(this);
			ProjectBuilder.setRelativePosition(nextForm, this);
			nextForm.setVisible(true);
			this.setVisible(false);
		}
	}
}
