package edu.wpi.prc.tools.extractor;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.util.password.PasswordReader;

/**
 * Retrieves all result data from the project server and writes it to disk.
 * @author James Baldassari
 */
public class ResultExtractor {
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private static final String RESULTS_DIR = "results";
	private static final String RESULT_FILE = "result.dat";
	private static Project project = null;
	private static Vector validResults = null;
	
	/**
	 * Gets the vector containing all valid result data from the project server.
	 * @param serverAddress The address of the project server.
	 * @param password The project password.
	 * @return Returns true if the RPC was successful, false otherwise.
	 */
	private static boolean getResults(String serverAddress, String password) {
		boolean success = false;
		String serverURI = "http://" + serverAddress + "/RPC2";
		try {
			// Instantiate an XML-RPC client.
			XmlRpcClient xmlrpcClient = new XmlRpcClient(serverURI);
			// Build Vector of parameters.
			Vector<String> params = new Vector<String>();
			params.add(password);
			System.out.println("Retrieving result data from the server.  This may take a few minutes...");
			Object result = xmlrpcClient.execute("server.getValidResultQueue", params);
			try {
				validResults = (Vector)result;
				success = true;
			} catch (Exception e) {
				success = false;
				validResults = new Vector();
			}
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + serverURI + "\" from the project configuration file is invalid.");
			System.exit(EXIT_ERROR);
		} catch (XmlRpcException e) {
			System.out.println("Error executing the server.getValidResults RPC: " + e);
			System.exit(EXIT_ERROR);
		} catch (IOException e) {
			System.out.println("Could not connect to the project server.  It may not be running.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Writes all result data to disk.  This method creates a new directory 
	 * in RESULTS_DIR for each work unit.  The result data is written to the 
	 * file RESULT_FILE inside the directory of its associated work unit.
	 * @return Returns false if any of the work unit directories could not be created, true otherwise.
	 * @throws IOException
	 */
	private static boolean writeResultsToDisk() throws IOException {
		for (int ii = 0; ii < validResults.size(); ii+=2) {
			File resultDir = new File(RESULTS_DIR + File.separator + validResults.get(ii));
			// If the directory for this work unit does not exist, create it, and fail if the directory could not be created.
			if ((resultDir.exists() == false) && (resultDir.mkdirs() == false))
				return false;
			// Initialize the result file inside this work unit's directory.
			System.out.println("Saving result for work unit: " + validResults.get(ii));
			File resultFile = new File(RESULTS_DIR + File.separator + validResults.get(ii) + File.separator + RESULT_FILE);
			// Write the contents of the result to the result file.
			new FileOutputStream(resultFile).write((byte[])validResults.get(ii+1));
		}
		return true;
	}
	
	/**
	 * Gets all valid results from the server and writes them out to disk.
	 * Each work unit will have its own folder in the RESULTS_DIR directory,
	 * whose name is the task ID of the work unit.  Inside each work unit 
	 * directory there will be a file called "result.dat" which contains 
	 * the result data for that work unit.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		try {
			project = Project.readFromFile();
		} catch (MissingConfigurationFileException e) {
			String message = "Project configuration file not found.  Please verify that ";
			message += "the project_server.properties file is in the \"cfg\" directory.";
			System.out.println(message);
			System.exit(EXIT_ERROR);
		}
		
		String serverURI = project.getServerURI();				
		if (serverURI == null) {
			System.out.println("Error: Server address not found in project configuration file.");
			System.exit(EXIT_ERROR);
		}
		
		else {
			PasswordReader pr = new PasswordReader();
			
			try {
				// Read in the project password.
				String password = pr.readLine();
				// Try to get results from server.
				while (getResults(serverURI, password) == false) {
					System.out.println("The password you entered was incorrect.  Please enter the password again.");
					password = pr.readLine();
				}
			} catch (IOException e) {
				System.out.println("Error reading user input.");
				System.exit(EXIT_ERROR);
			}
			
			try {
				// Try to write all results to disk.
				if (writeResultsToDisk() == false) {
					System.out.println("One or more directories could not be created.  Please verify that you have write permissions in the current directory.");
					System.exit(EXIT_ERROR);
				}
			} catch (IOException e) {
				System.out.println("An error occurred while writing result data to disk: " + e);
				System.exit(EXIT_ERROR);
			}
			
			System.out.println("\nThe results have been extracted to the \"" + RESULTS_DIR + "\" directory.");
			System.out.println("Each work unit has its own folder in the " + RESULTS_DIR + " directory with the name of the work unit's task ID.");
			System.out.println("Inside each work unit's folder is the result data for that work unit, contained in the \"results.dat\" file.");
			System.exit(EXIT_SUCCESS);
		}
	}
}
