package edu.wpi.prc.user;

import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Expression;

import edu.wpi.prc.persistence.HibernateController;

/**
 * The Volunteer class stores information about each user.
 * @author James Baldassari
 */
public class Volunteer {
	private long id;
	private String userID = null;
	private String eMailAddress = null;
	private long score = 0;
	private Date joinDate = null;
	private Date lastResultDate = null;
	private String stateStr = null;
	private boolean waiting = false;	// Is this user waiting for a work unit?
	
	/**
	 * Default constructor.
	 */
	public Volunteer() {
		
	}
	
	/**
	 * Constructor specifying the user ID
	 * @param userID The user ID.
	 */
	public Volunteer(String userID) {
		this.userID = userID;
	}
	
	/**
	 * Constructor for new users.
	 * @param userID The user ID.
	 * @param eMailAddress The user's e-mail address.
	 * @param joinDate The date the user joined the PRC project.
	 */
	public Volunteer(String userID, String eMailAddress, Date joinDate) {
		this.userID = userID;
		this.eMailAddress = eMailAddress;
		this.joinDate = joinDate;
	}
	
	/**
	 * Constructor allowing the specification of all Volunteer attributes.
	 * @param userID The user ID.
	 * @param eMailAddress The user's e-mail address.
	 * @param score The user's score.
	 * @param joinDate The date the user joined the PRC project.
	 * @param lastResultDate The date the user last returned a result.
	 */
	public Volunteer(String userID, String eMailAddress, long score, Date joinDate, Date lastResultDate) {
		this.userID = userID;
		this.eMailAddress = eMailAddress;
		this.score = score;
		this.joinDate = joinDate;
		this.lastResultDate = lastResultDate;
	}

	/**
	 * @return Returns the id.
	 */
	protected long getId() {
		return id;
	}

	/**
	 * @param id The id to set.
	 */
	protected void setId(long id) {
		this.id = id;
	}

	/**
	 * @return Returns the eMailAddress.
	 */
	public String getEMailAddress() {
		return eMailAddress;
	}

	/**
	 * @param mailAddress The eMailAddress to set.
	 */
	public void setEMailAddress(String mailAddress) {
		eMailAddress = mailAddress;
	}

	/**
	 * @return Returns the score.
	 */
	public long getScore() {
		return score;
	}

	/**
	 * @param score The score to set.
	 */
	public void setScore(long score) {
		this.score = score;
	}
	
	/**
	 * Increase this Volunteer's score by delta points.
	 * @param delta The amount by which to increase the score.
	 */
	public void increaseScore(int delta) {
		score += delta;
	}
	
	/**
	 * Decreases this Volunteer's score by delta points.
	 * @param delta The amount by which to decrease the score.
	 */
	public void decreaseScore(int delta) {
		score -= delta;
	}

	/**
	 * @return Returns the userID.
	 */
	public String getUserID() {
		return userID;
	}

	/**
	 * @param userID The userID to set.
	 */
	public void setUserID(String userID) {
		this.userID = userID;
	}

	/**
	 * @return Returns the joinDate.
	 */
	public Date getJoinDate() {
		return joinDate;
	}

	/**
	 * @param joinDate The joinDate to set.
	 */
	public void setJoinDate(Date joinDate) {
		this.joinDate = joinDate;
	}

	/**
	 * @return Returns the lastResultDate.
	 */
	public Date getLastResultDate() {
		return lastResultDate;
	}

	/**
	 * @param lastResultDate The lastResultDate to set.
	 */
	public void setLastResultDate(Date lastResultDate) {
		this.lastResultDate = lastResultDate;
	}
	
	/**
	 * @return Returns the stateStr.
	 */
	protected String getStateStr() {
		return stateStr;
	}

	/**
	 * @param stateStr The stateStr to set.
	 */
	protected void setStateStr(String stateStr) {
		this.stateStr = stateStr;
	}
	
	/**
	 * @return Returns the state of this volunteer.
	 */
	public VolunteerState getState() {
		if (stateStr ==  null)
			return VolunteerState.Valid;
		else
			return Enum.valueOf(VolunteerState.class, stateStr);
	}
	
	/**
	 * Sets the state of this volunteer.
	 * @param state The state to set.
	 */
	public void setState(VolunteerState state) {
		this.stateStr = state.toString();
	}

	/**
	 * @return Returns the isWaiting.
	 */
	public boolean getWaiting() {
		return waiting;
	}

	/**
	 * @param isWaiting The isWaiting to set.
	 */
	public void setWaiting(boolean isWaiting) {
		this.waiting = isWaiting;
	}

	/**
	 * Saves this Volunteer to the database.
	 */
	public void save() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.save(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Updates this Volunteer in the database.
	 */
	public void update() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.update(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Finds all Volunteers in the database that match the attributes set in the this Volunteer.
	 * NOTE: user names and e-mail addresses are case-insensitive.
	 * @return An array of all matching Volunteers in the database.
	 */
	public Volunteer[] find() {
		Volunteer[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			String uID = getUserID();
			String eMail = getEMailAddress();
			String state = getStateStr();
			Criteria query = hibernateSession.createCriteria(Volunteer.class);
			
			// Dynamically build the query.
			if (uID != null)
				query.add(Expression.eq("userID", uID).ignoreCase());
			if (eMail != null)
				query.add(Expression.eq("EMailAddress", eMail).ignoreCase());
			if (state != null) {
				query.add(Expression.eq("stateStr", state));
			}
			
			// Execute the query.
			List results = query.list();
			
			// Build array of matching Volunteers.
			matches = new Volunteer[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (Volunteer)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
		} finally {
			try {
				if (hibernateSession != null)
					hibernateSession.close();
			} catch (Exception e) {
				
			}
		}
		
		return matches;
	}
	
	/**
	 * Finds all Volunteers in the database.
	 * @return An array of all Volunteers in the database.
	 */
	public static Volunteer[] findAll() {
		Volunteer[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			Criteria query = hibernateSession.createCriteria(Volunteer.class);
			
			// Execute the query.
			List results = query.list();
			
			// Build array of matching Volunteers.
			matches = new Volunteer[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (Volunteer)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
            e.printStackTrace();
        } finally {
                try {
                        if (hibernateSession != null)
                                hibernateSession.close();
                } catch (Exception e) {

                }
        }
		
		return matches;
	}
	
	/**
	 * Deletes this Volunteer from the database.
	 */
	public void delete() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.delete(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * @return Returns the number of users who are waiting for work units.
	 */
	public static int getNumUsersWaiting() {
		// Build query to find accepted spot check result.
		String queryString = "from Volunteer as user";				// Get Volunteers
		queryString += " where user.waiting='true'";				// who are waiting
		
		// Execute the query.
		Session hibernateSession = HibernateController.instance().createSession();		
		Query resultQuery = hibernateSession.createQuery(queryString);
		List results = resultQuery.list();

		try {
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			
		}
		
		return results.size();
	}

	/**
	 * Tests whether this Volunteer is the same as another.
	 * Delegates to the Hibernate EqualsHelper.
	 * @param o The other Volunteer.
	 * @return True if the the two Volunteers are the same, false otherwise.
	 */
	public boolean equals(Object o) {
		try {
			Volunteer v = (Volunteer)o;
			return v.getUserID().equals(this.getUserID());
		} catch (Exception e) {
			return false;
		}
	}
	
	/**
	 * Returns a copy of this Volunteer.
	 * @return A copy of this Volunteer.
	 */
	public Object clone() {
		return new Volunteer(new String(userID), new String(eMailAddress), 
				score, new Date(joinDate.getTime()), new Date(lastResultDate.getTime()));
	}
	
	/**
	 * Converts this Volunteer to a String.
	 * @return The String representation of this Volunteer.
	 */
	public String toString() {
		String volStr = new String("");
		
		if (userID != null)
			volStr += "User ID: " + userID;
		if (eMailAddress != null)
			volStr += "\nE-Mail Address: " + eMailAddress;
		if (score >= 0)
			volStr += "\nScore: " + score;		
		if (joinDate != null)
			volStr += "\nJoin Date: " + joinDate;
		if (lastResultDate != null)
			volStr += "\nLast Result Received: " + lastResultDate;
		
		return volStr;
	}
}
